(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015, 2016)
License: LGPL (see file COPYING.LESSER)
*)

chapter \<open>AC-Dependency Pairs\<close>


theory AC_Dependency_Pairs
imports
  FuncSet
  "../Framework/QDP_Framework"
  "../AC_Rewriting/AC_Rewriting"
  DP_Transformation
begin

lemma chain_reshuffling: fixes P Q R E :: "'a rel"
  and M :: "'a \<Rightarrow> bool"
  defines "M' \<equiv> {(s,t). M t} :: 'a rel"
  defines "S \<equiv> ((Q \<union> E) \<inter> M')^* O ((P \<union> R) \<inter> M')"
  assumes f: "\<And> i. (f i, f (Suc i)) \<in> S"  
  shows "\<not> SN_rel_ext P Q R E M \<or> (\<exists> t. M t \<and> \<not> SN_on (relto (R \<inter> M') (E \<inter> M')) {t})"
proof -
  def QE \<equiv> "(Q \<union> E) \<inter> M'"
  def Q' \<equiv> "relto (Q \<inter> M') QE"
  def PR \<equiv> "(P \<union> R) \<inter> M'"
  def P' \<equiv> "P \<inter> M'"
  def A \<equiv> "(P \<union> Q \<union> R \<union> E) \<inter> M'"
  def g \<equiv> "\<lambda> i. f (Suc i)"
  have g: "\<And> i. (g i, g (Suc i)) \<in> S" unfolding g_def using f by auto
  from f[unfolded S_def] 
  have M: "\<And> i. M (g i)" unfolding M'_def g_def by auto
  from g have "\<forall> i. \<exists> hi. (g i, hi) \<in> QE^* \<and> (hi, g (Suc i)) \<in> PR" unfolding S_def QE_def PR_def by blast
  from choice[OF this] obtain h where gh: "\<And> i. (g i, h i) \<in> QE^*" and hg: "\<And> i. (h i, g (Suc i)) \<in> PR" by blast
  def j \<equiv> "\<lambda> i :: nat. i div 2"
  def f \<equiv> "\<lambda> i. (if even i then g else h) (j i)"
  def t \<equiv> "\<lambda> i. (if even i then 
     if (f i, f (Suc i)) \<in> Q' then top_ns else normal_ns else
     if (f i, f (Suc i)) \<in> P' then top_s else normal_s)"
  have "QE \<subseteq> A" unfolding QE_def A_def by auto
  hence QE: "QE^* \<subseteq> A^*" by (rule rtrancl_mono)
  show ?thesis
  proof (cases "INFM i. t i \<in> {top_s,top_ns}")
    case True
    have "\<not> SN_rel_ext (A^* O (P \<inter> M') O A^*) (A^* O ((P \<union> Q) \<inter> M') O A^*) (A^* O ((P \<union> R) \<inter> M') O A^*) (A^*) M"
      unfolding SN_rel_ext_def not_not
    proof (rule exI[of _ f], rule exI[of _ t], intro conjI allI)
      fix i
      have M: "M (g (j i))" by fact
      show "M (f i)"
      proof (cases "even i")
        case True
        thus ?thesis using M unfolding f_def by auto
      next
        case False
        hence id: "f i = h (j i)" unfolding f_def by auto
        from gh[of "j i"] M have "M (h (j i))" unfolding QE_def M'_def
          by (induct, auto)
        thus ?thesis using id by simp
      qed
      show "(f i, f (Suc i)) \<in> SN_rel_ext_step (A\<^sup>* O (P \<inter> M') O A\<^sup>*) (A\<^sup>* O ((P \<union> Q) \<inter> M') O A\<^sup>*)
          (A\<^sup>* O ((P \<union> R) \<inter> M') O A\<^sup>*) (A\<^sup>*) (t i)"
      proof (cases "even i")
        case True
        hence fi: "f i = g (j i)" and fsi: "f (Suc i) = h (j i)" and jsi: "j (Suc i) = j i"
          unfolding f_def j_def by auto
        show ?thesis
        proof (cases "(f i, f (Suc i)) \<in> Q'")
          case True
          hence t: "t i = top_ns" unfolding t_def using `even i` by auto
          with True QE fi fsi show ?thesis by (auto simp: Q'_def)
        next
          case False
          hence t: "t i = normal_ns" unfolding t_def using `even i` by auto
          with gh[of "j i"] QE fi fsi jsi show ?thesis by auto
        qed
      next
        case False
        hence fi: "f i = h (j i)" and fsi: "f (Suc i) = g (Suc (j i))" and jsi: "j (Suc i) = Suc (j i)"
          unfolding f_def j_def by auto
        show ?thesis
        proof (cases "(f i, f (Suc i)) \<in> P'")
          case True
          hence t: "t i = top_s" unfolding t_def using `odd i` by auto
          with True QE fi fsi show ?thesis by (auto simp: P'_def)
        next
          case False
          hence t: "t i = normal_s" unfolding t_def using `odd i` by auto
          with hg[of "j i"] QE fi fsi jsi show ?thesis by (auto simp: PR_def)
        qed
      qed
      show "\<exists>\<^sub>\<infinity>i. t i \<in> {top_s, normal_s}"
        unfolding INFM_nat_le 
      proof
        fix i
        show "\<exists> j \<ge> i. t j \<in> {top_s, normal_s}"
          by (rule exI[of _ "2 * i + 1"], auto simp: t_def)
      qed
    qed (rule True)      
    with SN_rel_ext_trans[of P Q R E M, folded M'_def, folded A_def]
    have "\<not> SN_rel_ext P Q R E M" by blast
    thus ?thesis ..
  next
    case False
    then obtain k where k: "\<And> i. i \<ge> k \<Longrightarrow> t i \<notin> {top_s, top_ns}" 
      unfolding INFM_nat_le by auto
    {
      fix i
      assume i: "i \<ge> k"
      hence "2 * i \<ge> k" by auto
      from k[OF this, unfolded t_def f_def j_def] 
      have nmem: "(g i, h i) \<notin> Q'" by auto
      have sub: "\<And> A B. (A \<union> B)^* \<subseteq> B^* \<union> relto A (A \<union> B)" by regexp
      from set_mp[OF sub gh[of i, unfolded QE_def Int_Un_distrib2]] nmem
      have gh: "(g i, h i) \<in> (E \<inter> M')^*" unfolding QE_def Q'_def Int_Un_distrib2 by blast
      from i have "2 * i + 1 \<ge> k" by auto
      from k[OF this, unfolded t_def f_def j_def]
      have nmem: "(h i, g (Suc i)) \<notin> P'" by auto
      with hg[of i] have hg: "(h i, g (Suc i)) \<in> R \<inter> M'" unfolding PR_def P'_def by auto
      note gh hg
    } note gh = this
    def f \<equiv> "\<lambda> i. g (k + i)"
    {
      fix i
      have "(f i, f (Suc i)) \<in> relto (R \<inter> M') (E \<inter> M')"
        unfolding f_def using gh[of "k + i"] by auto
    }
    hence "\<not> SN_on (relto (R \<inter> M') (E \<inter> M')) {f 0}" by (rule steps_imp_not_SN_on)
    with M[of k]
    show ?thesis unfolding f_def by auto
  qed
qed



context size_preserving_trs
begin

(*Lemma 1 of Akihisa's draft*)
interpretation subteq_redpair: SN_order_pair "suptrel R" "supteqrel R"
by (standard) (auto intro: SN_suptrel refl_supteqrel trans_supteqrel trans_suptrel
                           supteqrel_suptrel_comp suptrel_supteqrel_comp)

end

abbreviation "relstep R E \<equiv> relto (rstep R) (rstep E)"

lemma
  "relstep R E = rel_rstep (R, E)"
by simp

context
  fixes shp :: "'f \<Rightarrow> 'f"
begin

interpretation sharp_syntax .

(*Lemma 2 of Akihisa's draft.*)
lemma lemma2_pre:
  assumes "wf_trs R" and "l \<cdot> \<sigma> \<in> Tinf (relstep R E)" and "(l, r) \<in> R \<union> E"
    and "\<not> SN_on (relstep R E) {r \<cdot> \<sigma>}"
  shows "\<exists>u \<unlhd> r. u \<cdot> \<sigma> \<in> Tinf (relstep R E) \<and> (\<sharp> l, \<sharp> u) \<in> DP \<sharp> R \<union> DP \<sharp> E"
 s o r r y

text \<open>lemma2 is wrong, since the defined symbols are taken separately for each 
  TRS!
  Consider R = {a -> b}, E = {b -> a}, then DP R = DP E = {}.\<close>
lemma lemma2:
  fixes R E
  defines "M \<equiv> Tinf (relstep R E)"
  assumes "wf_trs R" and "s \<in> M"
      and "\<not> SN_on (relstep R E) {t}"
  shows "(s,t) \<in> rrstep R \<Longrightarrow> \<exists>t' \<unlhd> t. t' \<in> M \<and> (\<sharp> s, \<sharp> t') \<in> rrstep (DP \<sharp> R)"
    and "(s,t) \<in> rrstep E \<Longrightarrow> \<exists>t' \<unlhd> t. t' \<in> M \<and> (\<sharp> s, \<sharp> t') \<in> rrstep (DP \<sharp> E)"
 s o r r y


text \<open>The following lemma Tinf_nrrstep is wrong, but perhaps one can use 
  @{thm Tinf_imp_SN_nr_first_root_step_rel[of _ False "{}" _ "{}", unfolded qrstep_rstep_conv nrqrstep_nrrstep rqrstep_rrstep_conv]}
  instead. Counterexample: E = {}, R = {f(a) -> f(a), a -> b}, s = f(a), t = f(b).\<close>

lemma Tinf_nrrstep:
  assumes s: "s \<in> Tinf (relstep R E)" and st: "(s,t) \<in> nrrstep (R \<union> E)"
  shows "t \<in> Tinf (relstep R E)"
  oops

lemma not_SN_on_pred:
  assumes "\<not> SN_on R {b}" and "(a, b) \<in> R"
  shows "\<not> SN_on R {a}"
using step_preserves_SN_on and assms by metis

lemma not_SN_on_rel_pred:
  assumes "(a, b) \<in> E\<^sup>*" and "\<not> SN_on (relto R E) {b}"
  shows "\<not> SN_on (relto R E) {a}"
proof -
  obtain f where "f 0 = b" and *: "\<And>i. (f i, f (Suc i)) \<in> relto R E" using assms by auto
  moreover def g \<equiv> "\<lambda>i. if i = 0 then a else f i"
  ultimately have "g 0 \<in> {a}" and "(b, f 1) \<in> relto R E" by auto
  moreover from relcompI [OF \<open>(a, b) \<in> E\<^sup>*\<close> this(2)]
    have "(a, f 1) \<in> relto R E" by (auto simp: g_def)
  moreover then have "\<forall> i. (g i, g (Suc i)) \<in> relto R E" using * by (auto simp: g_def)
  ultimately show ?thesis by auto
qed

lemma not_SN_on_succ:
  assumes "\<not> SN_on R {a}"
  shows "\<exists>b. (a, b) \<in> R \<and> \<not> SN_on R {b}"
proof -
  obtain f where "f 0 = a" and "\<And>i. (f i, f (Suc i)) \<in> R" using assms by auto
  then have "(a, f 1) \<in> R" and "\<not> SN_on R {f 1}" by (auto simp: SN_defs)
  then show ?thesis by auto
qed

lemma not_SN_on_rel_succ:
  assumes "\<not> SN_on (relstep R E) {s}"
  shows "\<exists>t u. (s, t) \<in> (rstep E)\<^sup>* \<and> (t, u) \<in> (rstep R) \<and> \<not> SN_on (relstep R E) {u}"
proof -
  obtain v where "(s, v) \<in> relstep R E" and v: "\<not> SN_on (relstep R E) {v}"
    using not_SN_on_succ [OF assms] by auto
  moreover then obtain t and u
    where "(s, t) \<in> (rstep E)\<^sup>*" and "(t, u) \<in> rstep R" and "(u, v) \<in> (rstep E)\<^sup>*" by auto
  moreover then have "\<not> SN_on (relstep R E) {u}" using not_SN_on_rel_pred [OF _ v] by auto
  ultimately show ?thesis by auto
qed

(* currently not needed   
lemma sharp_nrrstep_imp_nrrstep:
  assumes "inj shp" and "(\<sharp> s, \<sharp> t) \<in> nrrstep R"
  shows "(s, t) \<in> nrrstep R"
proof -
  {
    fix t
    have "sharp_term (the_inv \<sharp>) (\<sharp> t) = t"
      by (cases t, auto simp: the_inv_f_f[OF assms(1)])
  }
  from nrrstep_imp_sharp_nrrstep[OF assms(2), of "the_inv \<sharp>", unfolded this]
  show ?thesis .
qed

lemma sharp_inner_rewriting:
  assumes "inj shp"
  shows "(\<sharp> s, \<sharp> t) \<in> nrrstep R \<longleftrightarrow> (s, t) \<in> nrrstep R"
  using sharp_nrrstep_imp_nrrstep[OF assms] nrrstep_imp_sharp_nrrstep by metis
*)

definition "iacchain_part P Q R E s v n \<equiv>
  (\<forall>i. v i \<in> Tinf (relstep R E)) \<and>
  s = v 0 \<and>
  (\<forall>i<n. (\<sharp> (v i), \<sharp> (v (Suc i))) \<in> rrstep Q \<union> nrrstep E) \<and>
  (\<sharp> (v n), \<sharp> (v (Suc n))) \<in> rrstep (DP \<sharp> R) \<union> nrrstep R"

lemma iacchain_partI[intro]:
  assumes "\<And>i. v i \<in> Tinf (relstep R E)"
      and "s = v 0"
      and "\<And>i. i<n \<Longrightarrow> (\<sharp> (v i), \<sharp> (v (Suc i))) \<in> rrstep Q \<union> nrrstep E"
      and "(\<sharp> (v n), \<sharp> (v (Suc n))) \<in> rrstep (DP \<sharp> R) \<union> nrrstep R"
  shows "iacchain_part P Q R E s v n"
  using assms unfolding iacchain_part_def by auto

lemma iacchain_partD[dest]:
  assumes "iacchain_part P Q R E s v n"
  shows "\<And>i. v i \<in> Tinf (relstep R E)"
    and "s = v 0"
    and "\<And>i. i<n \<Longrightarrow> (\<sharp> (v i), \<sharp> (v (Suc i))) \<in> rrstep Q \<union> nrrstep E"
    and "(\<sharp> (v n), \<sharp> (v (Suc n))) \<in> rrstep (DP \<sharp> R) \<union> nrrstep R"
  using assms unfolding iacchain_part_def using PiE by auto

definition "starts_iacchain P Q R E s \<equiv> \<exists>v n. iacchain_part P Q R E s v n"

lemma starts_iacchain_pred:
  assumes s: "s : Tinf (relstep R E)"
      and st: "(\<sharp> s, \<sharp> t) \<in> rrstep Q \<union> nrrstep E"
      and t: "starts_iacchain P Q R E t"
  shows "starts_iacchain P Q R E s"
proof -
  obtain v n where v: "iacchain_part P Q R E t v n" using t unfolding starts_iacchain_def by auto
  def v' \<equiv> "\<lambda>i. if i = 0 then s else v (i-1)"
  show ?thesis unfolding starts_iacchain_def
  proof(intro exI iacchain_partI)
    note * = iacchain_partD[OF v]
    { fix i show "v' i \<in> Tinf (relstep R E)" unfolding v'_def using s *(1) by auto
    }
    { fix i assume "i < Suc n" thus "(\<sharp> (v' i), \<sharp> (v' (Suc i))) \<in> rrstep Q \<union> nrrstep E"
        apply(cases i) unfolding v'_def using st * by auto
    }
    show "s = v' 0" unfolding v'_def by auto
    show "(\<sharp> (v' (Suc n)), \<sharp> (v' (Suc (Suc n)))) \<in> rrstep (DP \<sharp> R) \<union> nrrstep R" using * unfolding v'_def by auto
  qed
qed

(*Lemma 3 of Akihisa's draft.*)
lemma lemma3:
  assumes E: "size_preserving_trs E" and R: "wf_trs R"
  defines "M \<equiv> Tinf (relstep R E)"
  assumes s: "s \<in> M" and st: "(s, t) \<in> (rstep E)\<^sup>*" and tu: "(t, u) \<in> rstep R"
    and u: "\<not> SN_on (relstep R E) {u}"
  shows "starts_iacchain (DP \<sharp> R) (DP \<sharp> E) R E s"
proof -
  interpret sp: size_preserving_trs E by fact
  let ?lextwo = "{(ab, ab'). lex_two (\<lambda>x y. (x, y) \<in> suptrel E) (\<lambda>x y. (x, y) \<in> supteqrel E) (\<lambda>a b. (b :: nat) < a) ab ab'}"
  have SN1: "SN {(x,y). (x,y) \<in> suptrel E}" using sp.SN_suptrel by auto
  have SN: "SN ?lextwo"
    by (rule lex_two[OF supteqrel_suptrel_comp SN1 SN_nat_gt])
  obtain n where "(s, t) \<in> (rstep E) ^^ n" using st by auto
  then show ?thesis using s tu u
  proof (induct "(s, n)" arbitrary: s t u n rule: SN_induct [OF SN])
    case 1
      hence s: "s \<in> M"
        and st: "(s, t) \<in> (rstep E) ^^ n"
        and tu: "(t, u) \<in> rstep R"
        and u: "\<not> SN_on (relstep R E) {u}"
        by auto
      note IH = 1(1)
      show ?case
      proof(cases "n")
        case 0
          hence [simp]: "s = t" using st by auto
          show ?thesis
          proof(cases rule: rstep_cases[OF tu])
            case 1
              then obtain t' where st': "(\<sharp> s, \<sharp> t') \<in> rrstep (DP \<sharp> R)" and t': "t' \<in> M"
                using lemma2(1)[OF R s[unfolded M_def] u] unfolding M_def by auto
              thus ?thesis using s unfolding starts_iacchain_def M_def
                by (intro exI[of _ "\<lambda>i. if i = 0 then s else t'"] exI, auto)
            next case 2
              hence "(\<sharp> t, \<sharp> u) \<in> nrrstep R" using nrrstep_imp_sharp_nrrstep by auto
              moreover from s have u: "u \<in> M" unfolding M_def
                apply(rule Tinf_nrrstep) unfolding nrrstep_union using 2 by auto
              ultimately show ?thesis using s unfolding starts_iacchain_def M_def
                by(intro exI[of _ "\<lambda>i. if i = 0 then s else u"] exI, auto)
          qed
        next case (Suc n')
          with st obtain s' where ss': "(s,s') \<in> rstep E" and s't: "(s',t) \<in> (rstep E) ^^ n'"
            by (meson relpow_Suc_D2)
          from relpow_imp_rtrancl[OF s't] tu have "(s',u) \<in> relstep R E" by auto
          hence s': "\<not> SN_on (relstep R E) {s'}" by(rule not_SN_on_pred[OF u])
          show ?thesis
          proof(cases rule: rstep_cases[OF ss'])
            case 1
              from lemma2(2)[OF R s[unfolded M_def] s' this]
              obtain s'' where sub: "s'' \<unlhd> s'" and s's'': "(\<sharp> s, \<sharp> s'') \<in> rrstep (DP \<sharp> E)" and s'': "s'' \<in> M"
                unfolding M_def by auto
              show ?thesis
              proof(cases "s'' = s'")
                case True hence s': "s' \<in> M" using s'' by auto
                  have "(s,s') \<in> supteqrel E" using ss' unfolding supteqrel_def by auto
                  hence "((s,n),(s',n')) \<in> ?lextwo" using Suc by auto
                  note IH = IH[OF this s't s' tu u]
                  show ?thesis
                    apply (rule starts_iacchain_pred[OF s[unfolded M_def] _ IH])
                    using s's'' True by auto
                next case False hence "s'' \<lhd> s'" using sub by auto
                  hence "(s,s'') \<in> suptrel E" using ss' unfolding suptrel_def by auto
                  hence *: "\<And>m. ((s,n),(s'',m)) \<in> ?lextwo" by auto
                  from s''[unfolded M_def]
                  have "\<not> SN_on (relstep R E) {s''}" by (simp add: Tinf_def)
                  from not_SN_on_rel_succ[OF this]
                  obtain t' u' m
                  where 1: "(s'', t') \<in> rstep E ^^ m" and 2: "(t', u') \<in> rstep R"
                    and u': "\<not> SN_on ((rstep E)\<^sup>* O rstep R O (rstep E)\<^sup>*) {u'}" by auto
                  show ?thesis apply (rule starts_iacchain_pred[OF s[unfolded M_def]])
                    using s's'' IH[OF *[folded Suc] 1 s'' 2 u'] by auto
              qed
            next case 2
              from s have s': "s' \<in> M" unfolding M_def
                apply (rule Tinf_nrrstep) using 2 nrrstep_union by auto
              have "(s,s') \<in> supteqrel E" using ss' unfolding supteqrel_def by auto
              hence "((s,n),(s',n')) \<in> ?lextwo" using Suc by auto
              note IH = IH[OF this s't s' tu u]
              show ?thesis apply(rule starts_iacchain_pred[OF s[unfolded M_def]])
                using nrrstep_imp_sharp_nrrstep[OF 2] IH by auto
          qed
      qed
  qed
qed

definition "min_iacchain P Q R Rw E s t (\<sigma> :: (nat \<Rightarrow> ('f,'v)subst)) \<longleftrightarrow>
  (\<forall>i. (s i, t i) \<in> P \<union> Q) \<and>
  (\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>*) \<and>
  ((\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> P) \<or>
   (\<exists>\<^sub>\<infinity>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> relto (rstep R) (rstep (R \<union> Rw \<union> E)))) \<and>
  (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i \<cdot> \<sigma> i})"

text \<open>The following characterization admits to reuse lemmas such as
  @{thm ichain_split} which are crucial for reduction pair processors, etc.\<close>
lemma min_iacchain_via_ichain: "min_iacchain P Q R Rw E s t \<sigma> \<longleftrightarrow>
  ichain (False,False,P,Q,{},R,Rw \<union> E) s t \<sigma> \<and> (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i \<cdot> \<sigma> i})"
  unfolding min_iacchain_def by (auto simp: ichain.simps ac_simps)

lemma min_iacchain_split: assumes chain: "min_iacchain P Q R Rw E s t \<sigma>"
  and nchain: "\<not> min_iacchain (Ps \<inter> (P \<union> Q)) (P \<union> Q - Ps) (Rs \<inter> (R \<union> Rw)) (R \<union> Rw - Rs) E s t \<sigma>"
  shows "\<exists> i. min_iacchain (P - Ps) (Q - Ps) (R - Rs) (Rw - Rs) E (shift s i) (shift t i) (shift \<sigma> i)"
proof -
  have Rw: "R \<union> Rw = (Rs \<inter> (R \<union> Rw)) \<union> (R \<union> Rw - Rs)" by auto
  have nchain: "\<not> ichain (False,False,Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, Rs \<inter> (R \<union> Rw),  R \<union> Rw - Rs \<union> E) s t \<sigma>"
  proof
    assume ichain: "ichain (False,False,Ps \<inter> (P \<union> Q), P \<union> Q - Ps, {}, Rs \<inter> (R \<union> Rw),  R \<union> Rw - Rs \<union> E) s t \<sigma>"
    hence "min_iacchain (Ps \<inter> (P \<union> Q)) (P \<union> Q - Ps) (Rs \<inter> (R \<union> Rw)) (R \<union> Rw - Rs) E s t \<sigma>"
      unfolding min_iacchain_via_ichain
      using chain[unfolded min_iacchain_via_ichain]
      by (simp add: Rw[symmetric])
    with nchain show False by auto
  qed
  from chain have ichain: "ichain (False,False,P,Q,{},R,Rw \<union> E) s t \<sigma>" 
    unfolding min_iacchain_via_ichain by auto
  from ichain_split_gen[OF ichain nchain] obtain i
    where ichain: "ichain (False,False,P - Ps, Q - Ps, {}, R - Rs, Rw - Rs \<union> E) (shift s i) (shift t i) (shift \<sigma> i)" ..
  have "min_iacchain (P - Ps) (Q - Ps) (R - Rs) (Rw - Rs) E (shift s i) (shift t i) (shift \<sigma> i)"
    unfolding min_iacchain_via_ichain
    by (rule conjI[OF ichain], rule allI, 
    rule SN_on_mono[OF _ relto_mono[OF rstep_mono[of _ "R \<union> Rw"] subset_refl]],
    insert chain[unfolded min_iacchain_via_ichain], auto)
  thus ?thesis ..
qed

lemma min_iacchain_split_top: assumes chain: "min_iacchain P Q R Rw E s t \<sigma>"
  and nchain: "\<not> min_iacchain (Ps \<inter> (P \<union> Q)) (P \<union> Q - Ps) {} (R \<union> Rw) E s t \<sigma>"
  shows "\<exists> i. min_iacchain (P - Ps) (Q - Ps) R Rw E (shift s i) (shift t i) (shift \<sigma> i)"
  using min_iacchain_split[OF chain, of Ps "{}"] nchain by auto

definition "min_iacchain_alt P Q R Rw E s t \<longleftrightarrow>
  (\<forall>i. (s i, t i) \<in> rrstep (P \<union> Q)) \<and>
  (\<forall>i. (t i, s (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>*) \<and>
  ((\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> rrstep P) \<or>
   (\<exists>\<^sub>\<infinity>i. (t i, s (Suc i)) \<in> relstep R (R \<union> Rw \<union> E))) \<and>
  (\<forall>i. SN_on (relstep (R \<union> Rw) E) {t i})"

lemma min_iacchain_altI:
  "min_iacchain P Q R Rw E s t \<sigma> \<Longrightarrow> min_iacchain_alt P Q R Rw E (\<lambda>i. s i \<cdot> \<sigma> i) (\<lambda>i. t i \<cdot> \<sigma> i)"
by (auto simp: min_iacchain_def min_iacchain_alt_def INFM_mono rrstepI)

lemma min_iacchain_altD:
  assumes "min_iacchain_alt P Q R Rw E s t"
  shows "\<exists>s t \<sigma>. min_iacchain P Q R Rw E s t \<sigma>"
proof -
  { fix i :: nat
    have "\<exists>u v \<sigma>. (u, v) \<in> P \<union> Q \<and> s i = u \<cdot> \<sigma> \<and> t i = v \<cdot> \<sigma>"
      using assms by (auto simp: min_iacchain_alt_def elim!: rrstepE dest!: spec [of _ i]) }
  then obtain u and v and \<sigma>
    where *: "\<forall>i. (u i, v i) \<in> P \<union> Q \<and> s i = u i \<cdot> \<sigma> i \<and> t i = v i \<cdot> \<sigma> i" by metis
  consider (P) "\<exists>\<^sub>\<infinity>i. (s i, t i) \<in> rrstep P" | (R) "\<exists>\<^sub>\<infinity>i. (t i, s (Suc i)) \<in> relstep R (R \<union> Rw \<union> E)"
    using assms by (auto simp: min_iacchain_alt_def)
  then show ?thesis
  proof (cases)
    def I \<equiv> "\<lambda>i. LEAST j. j \<ge> i \<and> (\<exists>u v \<sigma>. (u, v) \<in> P \<and> s j = u \<cdot> \<sigma> \<and> t j = v \<cdot> \<sigma>)" (is "\<lambda>i. LEAST j. ?P i j")
    case P
    then have P: "\<forall>i. \<exists>j. ?P i j" by (force simp: INFM_nat_le elim!: rrstepE)
    then have "\<forall>i. \<exists>u v \<sigma>. (u, v) \<in> P \<and> s (I i) = u \<cdot> \<sigma> \<and> t (I i) = v \<cdot> \<sigma>"
      by (auto simp: I_def dest!: LeastI_ex)
    then obtain u' and v' and \<sigma>'
      where **: "\<forall>i. (u' i, v' i) \<in> P \<and> s (I i) = u' i \<cdot> \<sigma>' i \<and> t (I i) = v' i \<cdot> \<sigma>' i" by metis
    def U \<equiv> "\<lambda>i. if \<exists>j. i = I j then u' (LEAST j. i = I j) else u i"
    def V \<equiv> "\<lambda>i. if \<exists>j. i = I j then v' (LEAST j. i = I j) else v i"
    def S \<equiv> "\<lambda>i. if \<exists>j. i = I j then \<sigma>' (LEAST j. i = I j) else \<sigma> i"
    { fix i :: nat
      have "(U i, V i) \<in> P \<union> Q \<and> (V i \<cdot> S i, U (Suc i) \<cdot> S (Suc i)) \<in> (rstep (R \<union> Rw \<union> E))\<^sup>* \<and>
        SN_on (relstep (R \<union> Rw) E) {V i \<cdot> S i}"
        using assms and * and **
        by (auto simp: min_iacchain_alt_def U_def V_def S_def) (metis (mono_tags, lifting) LeastI_ex)+ }
    moreover
    { fix i :: nat
      have "(U (I i), V (I i)) \<in> P" using ** by (auto simp: U_def V_def)
      moreover have "I i \<ge> i" using P by (auto simp: I_def dest!: LeastI_ex)
      ultimately have "\<exists>j\<ge>i. (U j, V j) \<in> P" by blast }
    ultimately show ?thesis by (auto simp: min_iacchain_def INFM_nat_le)
  next
    case R
    then have "min_iacchain P Q R Rw E u v \<sigma>"
      using assms and * by (auto simp: min_iacchain_alt_def min_iacchain_def)
    then show ?thesis by blast
  qed
qed

definition "strictly_finite P Q R E \<longleftrightarrow> \<not> (\<exists>s t \<sigma>. min_iacchain P Q R {} E s t \<sigma>)"
definition "weakly_finite P Q R E \<longleftrightarrow> \<not> (\<exists>s t \<sigma>. min_iacchain P Q {} R E s t \<sigma>)"

lemma strictly_finite_imp_SN_rel_ext:
  assumes "strictly_finite P Q R E"
  shows "SN_rel_ext (rrstep P) (rrstep Q) (rstep R) (rstep E) (\<lambda>t. SN_on (relstep R E) {t})"
    (is "SN_rel_ext ?P ?Q ?R ?E ?M")
proof (rule ccontr)
  let ?rel = "SN_rel_ext_step ?P ?Q ?R ?E"
  assume "\<not> ?thesis"
  from this[unfolded SN_rel_ext_def] obtain f ty where
    steps: "\<And> i. (f i, f (Suc i)) \<in> ?rel (ty i)"
   and min: "\<And> i. ?M (f i)"
   and inf1: "INFM i. ty i \<in> {top_s,top_ns}"
   and inf2: "INFM i. ty i \<in> {top_s,normal_s}" by blast
  obtain p where p: "\<And> i. p i = (ty i \<in> {top_s,top_ns})" by auto
  interpret infinitely_many p
    by (unfold_locales, insert inf1, unfold INFM_nat_le p)
  let ?ind = "infinitely_many.index p"
  let ?g = "\<lambda> i. f (?ind i)"
  let ?h = "\<lambda> i. f (Suc (?ind i))"
  let ?prop = "\<lambda> s t \<sigma> i. (s, t) \<in> P \<union> Q \<and> ?g i = s \<cdot> \<sigma> \<and> ?h i = t \<cdot> \<sigma> \<and> (ty (?ind i) = top_s \<longrightarrow> (s, t) \<in> P)"
  {
    fix i
    have "ty (?ind i) \<in> {top_s,top_ns}"
      using index_p[of i] unfolding p .
    hence one: "(?g i, ?h i) \<in> ?P \<union> ?Q" and two: "ty (?ind i) = top_s \<Longrightarrow> (?g i, ?h i) \<in> ?P" using steps[of "?ind i"]
      by auto      
    have "\<exists> s t \<sigma>. ?prop s t \<sigma> i"
    proof (cases "ty (?ind i) = top_s")
      case True
      with two[OF True] have "(?g i, ?h i) \<in> rrstep P" by simp
      then show ?thesis by (auto elim!: rrstepE)
    next
      case False
      moreover from one have "(?g i, ?h i) \<in> rrstep (P \<union> Q)" by (auto simp: rrstep_union)
      ultimately show ?thesis by (auto elim!: rrstepE)
    qed
  }
  from choice[OF allI[OF this]] obtain s where "\<forall> i. \<exists> t \<sigma>. ?prop (s i) t \<sigma> i" ..
  from choice[OF this] obtain t where "\<forall> i. \<exists> \<sigma>. ?prop (s i) (t i) \<sigma> i" ..
  from choice[OF this] obtain \<sigma> where st\<sigma>: "\<And> i. ?prop (s i) (t i) (\<sigma> i) i" by auto
  from st\<sigma> have PQ: "\<And> i. (s i, t i) \<in> P \<union> Q" by auto
  from st\<sigma> have s: "\<And> i. s i \<cdot> \<sigma> i = ?g i" by auto
  from st\<sigma> have t: "\<And> i. t i \<cdot> \<sigma> i = ?h i" by auto
  from st\<sigma> have P: "\<And> i. ty (?ind i) = top_s \<Longrightarrow> (s i, t i) \<in> P" by auto
  {
    fix i
    from st\<sigma>[of i] have "?M (t i \<cdot> \<sigma> i)" using min[of "Suc (?ind i)"] by auto
  } note min = this
  let ?f = "\<lambda> i j. f (Suc (?ind i) + j)"
  let ?n = "\<lambda> i. ?ind (Suc i) - Suc (?ind i)"
  let ?RE = "qrstep False {} (R \<union> E)"
  have "ichain (False, False, P, Q, {}, R, E) s t \<sigma>"
    unfolding ichain_alternative
  proof (rule exI[of _ ?f], rule exI[of _ ?n], intro conjI allI impI, rule PQ)
    fix i
    show "?f i 0 = t i \<cdot> \<sigma> i" unfolding t by simp
  next
    fix i
    show "?f i (?n i) = s (Suc i) \<cdot> \<sigma> (Suc i)" unfolding s
      using index_ordered[of i] by auto
  next
    fix i j
    let ?k = "Suc (?ind i) + j"
    assume "j < ?n i"
    with index_not_p_between[of i ?k]
    have "\<not> p ?k" by auto
    hence "ty ?k \<in> {normal_s,normal_ns}" unfolding p by (cases "ty ?k", auto)
    with steps[of ?k] show "(?f i j, ?f i (Suc j)) \<in> ?RE" unfolding qrstep_union
      by auto
  next
    let ?L = "\<lambda> i. (s i, t i) \<in> P"
    let ?R = "\<lambda> i. \<exists> j < ?n i. (?f i j, ?f i (Suc j)) \<in> qrstep False {} R"
    show "(INFM i. ?L i) \<or> (INFM i. ?R i)" 
      unfolding INFM_disj_distrib[symmetric] 
      unfolding INFM_nat_le
    proof (intro allI)
      fix i
      from inf2[unfolded INFM_nat_le]
      obtain k where k: "k \<ge> ?ind i" and ty: "ty k \<in> {top_s,normal_s}" by auto
      from index_surj[OF k]
      obtain j j' where kj: "k = ?ind j + j'" and j': "?ind j + j' < ?ind (Suc j)" by auto
      note ty = ty[unfolded kj]
      from k[unfolded kj] j' have lt: "?ind i < ?ind (Suc j)" by auto
      {
        assume "j < i"
        hence "Suc j \<le> i" by auto
        from index_ordered_le[OF this] lt have False by auto
      }
      hence j: "i \<le> j" by presburger
      show "\<exists> j \<ge> i. ?L j \<or> ?R j"
      proof (intro exI conjI, rule j)
        show "?L j \<or> ?R j"
        proof (cases j')
          case 0
          from index_p[of j, unfolded p]
            ty[unfolded 0] have "ty (?ind j) = top_s" by auto
          from P[OF this] show ?thesis by auto
        next
          case (Suc j'')
          from index_not_p_between[of j "?ind j + j'", OF _ j']
          have "ty (?ind j + j') \<notin> {top_s,top_ns}" unfolding p Suc by auto
          with ty have ty: "ty (?ind j + j') = normal_s" by auto
          have j'': "j'' < ?ind (Suc j) - Suc (?ind j)" using j' unfolding Suc by auto
          have "?R j"
            by (rule exI[of _ j''], rule conjI[OF j''],
              insert steps[of "?ind j + j'", unfolded ty], unfold Suc, auto)
          thus ?thesis by auto
        qed
      qed
    qed
  qed auto
  then have "min_iacchain P Q R {} E s t \<sigma>"
    using min unfolding min_iacchain_via_ichain by auto
  with assms show False by (auto simp: strictly_finite_def)
qed

lemma strictly_finite_SN:
  assumes "strictly_finite P Q R E"
  defines "M' \<equiv> \<lambda>t. SN_on (relstep R E) {t}"
  defines "M \<equiv> {(s, t). M' t}"
  defines "S \<equiv> ((rrstep Q \<union> rstep E) \<inter> M)\<^sup>* O ((rrstep P \<union> rstep R) \<inter> M)"
  shows "SN S"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain f where "\<And>i. (f i, f (Suc i)) \<in> S" by (auto simp: SN_defs)
  from chain_reshuffling [of f, OF this [unfolded S_def M_def], folded M_def]
    consider (ichain) "\<not> SN_rel_ext (rrstep P) (rrstep Q) (rstep R) (rstep E) M'" |
      (min) "(\<exists>t. M' t \<and> \<not> SN_on (relto (rstep R \<inter> M) (rstep E \<inter> M)) {t})" by blast
  then show False
  proof (cases)
    case ichain
    with strictly_finite_imp_SN_rel_ext [OF assms(1)] show False by (auto simp: M'_def)
  next
    case min
    have "rstep R \<inter> M \<subseteq> rstep R" and "rstep E \<inter> M \<subseteq> rstep E" by auto
    from SN_on_mono [OF _ relto_mono [OF this]] and min show ?thesis by (auto simp: M'_def)
  qed
qed

(*Theorem 1 of Akihisa's draft.*)
lemma ACDP_transformation:
  fixes E :: "('f, 'v) trs"
  assumes "size_preserving_trs E"
    and "wf_trs R"
    and fin: "strictly_finite (DP \<sharp> R) (DP \<sharp> E) R E"
  shows "SN (relstep R E)"
proof (rule ccontr)
  def M' \<equiv> "\<lambda>t. SN_on (relstep R E) {t}"
  def M \<equiv> "{(s::('f, 'v) term, t). M' t}"
  def S \<equiv> "((rrstep (DP \<sharp> E) \<union> rstep E) \<inter> M)\<^sup>* O ((rrstep (DP \<sharp> R) \<union> rstep R) \<inter> M)"
  assume "\<not> SN (relstep R E)"
  then obtain s where "s \<in> Tinf (relstep R E)" by (auto simp: SN_def dest: not_SN_imp_subt_Tinf)
  then show False
  proof (induct "\<sharp> s" arbitrary: s rule: SN_induct [OF strictly_finite_SN [OF fin]])
    case (1 s)
    then have s: "s \<in> Tinf (relstep R E)" by auto
    moreover then obtain t and u
      where "(s, t) \<in> (rstep E)\<^sup>*" and "(t, u) \<in> (rstep R)" and "\<not> SN_on (relstep R E) {u}"
      by (auto simp: Tinf_def dest: not_SN_on_rel_succ)
    from lemma3 [OF \<open>size_preserving_trs E\<close> \<open>wf_trs R\<close> s this, unfolded starts_iacchain_def]
      obtain v and n
      where "iacchain_part (DP \<sharp> R) (DP \<sharp> E) R E s v n" by blast
    then have "(\<sharp> s, \<sharp> (v (Suc n))) \<in> S" and "v (Suc n) \<in> Tinf (relstep R E)" s o r r y
    with "1.hyps" show False by (auto simp: S_def M_def M'_def)
  qed
qed

(*end of sharp_syntax*)
end

end
