(*
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2015, 2016)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015, 2016)
License: LGPL (see file COPYING.LESSER)
*)

chapter \<open>Exact Tree Automata Completion\<close>

theory Exact_Tree_Automata_Completion
imports
  "../Rewriting/Trs"
  "../Rewriting/Multihole_Context"
  Tree_Automata
begin

(* TODO: move, but where? *)
lemma finite_subset_induct' [consumes 2, case_names empty insert]:
  assumes "finite F" and "F \<subseteq> A"
  assumes empty: "P {}"
    and insert: "\<And>a F. finite F \<Longrightarrow> F \<subseteq> A \<Longrightarrow> a \<in> A \<Longrightarrow> a \<notin> F \<Longrightarrow> P F \<Longrightarrow> P (insert a F)"
  shows "P F"
proof -
  have "P F \<and> F \<subseteq> A" using assms by (induct rule: finite_subset_induct) auto
  then show ?thesis ..
qed

(* TODO: duplicate of Trs.map_funs_term_inj but with more general type *)
lemma map_funs_term_inj':
  assumes "inj (fg :: 'f \<Rightarrow> 'g)"
  shows "inj (map_funs_term fg)"
proof -
  {
    fix s t :: "('f,'v) term"
    assume "map_funs_term fg s = map_funs_term fg t"
    hence "s = t"
    proof (induct s arbitrary: t)
      case (Var x) with assms show ?case by (induct t) auto
    next
      case (Fun f ss) thus ?case
      proof (induct t)
        case (Var y) thus ?case by auto
      next
        case (Fun g ts)
        hence A: "map (map_funs_term fg) ss = map (map_funs_term fg) ts" by simp
        hence len_eq:"length ss = length ts" by (rule map_eq_imp_length_eq)
        from A have "!!i. i < length ss \<Longrightarrow> (map (map_funs_term fg) ss)!i = (map (map_funs_term fg) ts)!i" by auto
        with len_eq have eq: "!!i. i < length ss \<Longrightarrow> map_funs_term fg  (ss!i) = map_funs_term fg (ts!i)" using nth_map by auto
        have in_set: "!!i. i < length ss \<Longrightarrow> (ss!i) \<in> set ss" by auto
        from Fun have "\<forall>i < length ss. (ss!i) = (ts!i)" using in_set eq by auto
        with len_eq have "ss = ts" using nth_equalityI[where xs = ss and ys = ts] by simp
        have "f = g" using Fun \<open>inj fg\<close> unfolding inj_on_def by auto
        with \<open>ss = ts\<close> show ?case by simp
      qed
    qed
  }
  thus ?thesis unfolding inj_on_def by auto
qed

abbreviation "fground F t \<equiv> ground t \<and> funas_term t \<subseteq> F"

abbreviation "fterm \<equiv> map_funs_term Inl"

abbreviation "fctxt \<equiv> map_funs_ctxt Inl"

abbreviation "qterm \<equiv> map_funs_term Inr"

abbreviation "State f \<equiv> Fun (Inr f) []"

abbreviation "fqterm f qs \<equiv>  Fun (Inl f) (map State qs)"

type_synonym ('f, 'q, 'v) crterm = "('f + 'q + ('f, 'v) term, 'v) term"

lemma fterm_no_State: "fterm s \<noteq> State q" by (induct s) (auto)

lemma fterm_no_Inr: "fterm s \<noteq> Fun (Inr f) ts" by (induct s) (auto)

lemma fterm_contains_no_Inr: "fterm s \<noteq> C\<langle>Fun (Inr f) ts\<rangle>"
by (induct C, simp add: fterm_no_Inr) (metis fterm_no_Inr map_funs_term_ctxt_decomp)

lemma ground_term_no_State:
  assumes "ground t"
  shows "\<forall> q. fterm t \<noteq> State q"
using assms by (induct t) auto

definition ancestors :: "'f sig \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) terms \<Rightarrow> ('f, 'v) terms"
where
  "ancestors F R L = { t | t s. (t, s) \<in> (rstep R)\<^sup>* \<and> s \<in> L \<and> fground F t }"

lemma ancestors_lang_mono:
  "L \<subseteq> L' \<Longrightarrow> ancestors F R L \<subseteq> ancestors F R L'"
by (auto simp: ancestors_def)

lemma ancestors_refl:
  assumes "s \<in> L" and "fground F s"
  shows "s \<in> ancestors F R L"
using assms unfolding ancestors_def by blast

fun growing_rule :: "('f, 'v) rule \<Rightarrow> bool"
where
  "growing_rule (l, r) \<longleftrightarrow> (\<forall>x \<in> vars_term r. \<forall>p \<in> varposs l. Var x = l |_ p \<longrightarrow> size p \<le> 1)"

definition "growing R \<longleftrightarrow> (\<forall>r \<in> R. growing_rule r)"

lemma Var_Fun_pos_size_One:
  assumes "Var x = Fun f ts |_ p" and "p \<in> poss (Fun f ts)" and "size p = 1"
  shows "\<exists>i < length ts. ts ! i = Var x"
using assms
apply (cases p)
apply auto
apply (case_tac x22)
apply auto
done

fun star :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "star c (Var x) = Fun c []"
| "star c (Fun f ts) = Fun f (map (star c) ts)"

fun star_ctxt :: "'f \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt"
where
  "star_ctxt c Hole = Hole"
| "star_ctxt c (More f ss C ts) = More f (map (star c) ss) (star_ctxt c C) (map (star c) ts)"

lemma star_subst_conv:
  "star a t = t \<cdot> (\<lambda>x. Fun a [])"
by (induct t) auto

lemma star_ctxt_apply_term [simp]:
  "star c (C\<langle>t\<rangle>) = (star_ctxt c C)\<langle>star c t\<rangle>"
by (induct C) simp_all

lemma star_subst_apply_term [simp]:
  "star c (t \<cdot> \<sigma>) = t \<cdot> (star c \<circ> \<sigma>)"
by (induct t) simp_all

lemma star_rstep:
  assumes "(s, t) \<in> rstep R" shows "(star c s, star c t) \<in> rstep R"
using assms by (force)

lemma star_rsteps:
  assumes "(s, t) \<in> (rstep R)\<^sup>*" shows "(star c s, star c t) \<in> (rstep R)\<^sup>*"
using assms by (induct) (auto intro: star_rstep rtrancl_into_rtrancl)

lemma ground_star [simp]: "ground (star c t)" by (induct t) simp_all

lemma ground_star_ident [simp]:
  assumes "ground t"
  shows "star c t = t"
using assms by (induct t) (auto simp: map_idI)

lemma funas_term_star_conv:
  "funas_term (star c t) = (if ground t then funas_term t else {(c, 0)} \<union> funas_term t)"
by (induct t) (auto split: if_splits)

definition "sig_rules F c = {TA_rule f (replicate n c) c | f n. (f, n) \<in> F}"

fun ground_instances_rules :: "'f sig \<Rightarrow> 'f \<Rightarrow> ('f, 'v) term \<Rightarrow> (('f ,'v) term, 'f) ta_rule set"
where
  "ground_instances_rules F c (Var x) = sig_rules F (Fun c [])"
| "ground_instances_rules F c (Fun f ts) =
    {TA_rule f (map (star c) ts) (star c (Fun f ts))} \<union> \<Union>(ground_instances_rules F c ` set ts)"

lemma ground_instances_rules_mono:
  assumes "s \<unlhd> t"
  shows "ground_instances_rules F c s \<subseteq> ground_instances_rules F c t"
using assms by (induct s) auto

definition "ground_instances_ta F c t =
  \<lparr> ta_final = {star c t}, ta_rules = ground_instances_rules F c t, ta_eps = {} \<rparr>"

lemma subst_sig_term_sig_imp:
  assumes "\<forall>x \<in> vars_term t. funas_term (\<sigma> x) \<subseteq> F"
    and "funas_term t \<subseteq> F"
  shows "funas_term (t \<cdot> \<sigma>) \<subseteq> F"
using assms by (induct t) (auto, blast)

section \<open>Tree automata\<close>
(*
The datatype for tree automata is already defined in Tree_Automata.
*)

(* A TA from a set of ta_rules with empty ta_final and ta_eps. *)
abbreviation "TA \<Delta> \<equiv> \<lparr> ta_final = {}, ta_rules = \<Delta>, ta_eps = {} \<rparr>"

abbreviation "empty_ta \<equiv> TA {}"

definition funas_ta_rule :: "('q, 'f) ta_rule \<Rightarrow> 'f sig" where
  "funas_ta_rule r = { (r_root r, length (r_lhs_states r)) }"

definition funas_ta :: "('q, 'f) ta \<Rightarrow> 'f sig" where
  "funas_ta A = \<Union>(funas_ta_rule ` ta_rules A)"

definition states :: "('q, 'f) ta \<Rightarrow> 'q set" where 
  "states A = \<Union> (r_states ` ta_rules A) \<union> \<Union> ((\<lambda> (q,q'). {q,q'}) ` ta_eps A) \<union> ta_final A"

definition ta_union :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q, 'f) ta"
where
  "ta_union A B = \<lparr>
    ta_final = ta_final A \<union> ta_final B,
    ta_rules = ta_rules A \<union> ta_rules B,
    ta_eps = ta_eps A \<union> ta_eps B
  \<rparr>"

lemma ta_union_simps [simp]:
  "ta_final (ta_union A B) = ta_final A \<union> ta_final B"
  "ta_rules (ta_union A B) = ta_rules A \<union> ta_rules B"
  "ta_eps (ta_union A B) = ta_eps A \<union> ta_eps B"
by (auto simp: ta_union_def)

lemma funas_ta_ta_union:
  assumes "funas_ta X \<subseteq> Y" and "funas_ta Z \<subseteq> W"
  shows "funas_ta (ta_union X Z) \<subseteq> Y \<union> W"
using assms by (auto simp: funas_ta_def)

lemma ta_union_comm [simp]: "ta_union A B = ta_union B A" by (auto simp: ta_union_def)

fun reachable_states :: "('q, 'f) ta \<Rightarrow> ('f + 'q, 'v) term \<Rightarrow> 'q set"
where
  "reachable_states A (Fun (Inr q) []) = { q'. (q,q') \<in> (ta_eps A)\<^sup>* }"
| "reachable_states A (Fun (Inl f) ts) = {q' | q' q qs.
    TA_rule f qs q \<in> ta_rules A \<and>
    (q,q') \<in> (ta_eps A)\<^sup>* \<and>
    length qs = length ts \<and>
    (\<forall> i < length ts. qs ! i \<in> reachable_states A (ts ! i)) }"
|  "reachable_states A _ = {}"

subsection \<open>Map states\<close>

definition map_r_states :: "('q \<Rightarrow> 'r) \<Rightarrow> ('q, 'f) ta_rule \<Rightarrow> ('r, 'f) ta_rule" where
  "map_r_states f r = TA_rule (r_root r) (map f (r_lhs_states r)) (f (r_rhs r))"

lemma funas_ta_rule_map_r_states:
  assumes "funas_ta_rule r \<subseteq> Y"
  shows "funas_ta_rule (map_r_states g r) \<subseteq> Y"
using assms by (auto simp: funas_ta_rule_def map_r_states_def)

definition map_states :: "('q \<Rightarrow> 'r) \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('r, 'f) ta" where
  "map_states f A = \<lparr>
    ta_final = f ` ta_final A,
    ta_rules = (map_r_states f) ` ta_rules A,
    ta_eps = {(f q, f q') | q q'. (q, q') \<in> ta_eps A}
  \<rparr>"

lemma ta_final_map_states_iff [simp]:
  "ta_final (map_states g A) = g ` ta_final A" by (auto simp: map_states_def)

lemma ta_rules_map_states_iff [simp]:
  assumes "inj g"
  shows "TA_rule f (map g qs) (g q) \<in> ta_rules (map_states g A) \<longleftrightarrow> TA_rule f qs q \<in> ta_rules A"
using assms by (force simp: map_states_def map_r_states_def inj_on_def)

lemma ta_rules_map_statesE [elim]:
  assumes "TA_rule f qs q \<in> ta_rules (map_states g A)"
  obtains qs' q' where "qs = map g qs'" and "q = g q'" and "TA_rule f qs' q' \<in> ta_rules A"
using assms by (auto simp: map_states_def map_r_states_def)

lemma ta_eps_map_statesE [elim]:
  assumes "(g q, x) \<in> ta_eps (map_states g A)"
  obtains q' where "x = g q'"
using assms by (auto simp: map_states_def)

lemma ta_eps_map_states2E [elim]:
  assumes "(x, g q) \<in> ta_eps (map_states g A)"
  obtains q' where "x = g q'"
using assms by (auto simp: map_states_def)

lemma rtrancl_ta_eps_map_statesE [elim]:
  assumes "(g q, x) \<in> (ta_eps (map_states g A))\<^sup>*"
  obtains q' where "x = g q'"
using assms by (induct) (auto)

lemma rtrancl_ta_eps_map_states2E [elim]:
  assumes "(x, g q) \<in> (ta_eps (map_states g A))\<^sup>*"
  obtains q' where "x = g q'"
using assms by (induct rule: converse_rtrancl_induct) auto

lemma ta_eps_map_states_iff' [simp]:
  assumes "inj g"
  shows "(g q, g q') \<in> ta_eps (map_states g A) \<longleftrightarrow> (q, q') \<in> ta_eps A"
  (is "?L \<longleftrightarrow> ?R")
using assms by (auto simp: inj_on_def map_states_def)

lemma ta_eps_map_states_iff [simp]:
  assumes "inj g"
  shows "(g q, g q') \<in> (ta_eps (map_states g A))\<^sup>* \<longleftrightarrow> (q, q') \<in> (ta_eps A)\<^sup>*"
  (is "?L \<longleftrightarrow> ?R")
proof
  assume "?L"
  then show "?R" using assms
  proof (induct "g q" "g q'" arbitrary: q')
    case (rtrancl_into_rtrancl)
    then show ?case
      by (auto simp: inj_on_def map_states_def elim!: rtrancl_ta_eps_map_statesE[of g]) force
  qed (auto simp: inj_on_def)
next
  assume "?R"
  then show "?L"
  proof (induct)
    case (step y z)
    then show ?case
      using rtrancl_into_rtrancl[of _ "g y" "ta_eps (map_states g A)"]
      by (auto simp: map_states_def)
  qed (auto)
qed

lemma funas_ta_map_states:
  assumes "funas_ta X \<subseteq> Y"
  shows "funas_ta (map_states g X) \<subseteq> Y"
proof -
  { fix r
    assume "r \<in> ta_rules X"
    then have "funas_ta_rule r \<subseteq> Y" using assms unfolding funas_ta_def by blast
    then have "funas_ta_rule (map_r_states g r) \<subseteq> Y" using funas_ta_rule_map_r_states
      by blast
  }
  then have "\<forall>r \<in> ta_rules X. funas_ta_rule (map_r_states g r) \<subseteq> Y" by blast
  then have "\<forall>r \<in> ta_rules (map_states g X). funas_ta_rule r \<subseteq> Y"
    by (auto simp: map_states_def)
  then show ?thesis by (auto simp: funas_ta_def)
qed

fun states_term :: "('f + 'q, 'v) term \<Rightarrow> 'q set" where
  "states_term (Fun (Inr q) ts) = { q } \<union> \<Union>(states_term ` set ts)"
| "states_term (Fun (Inl f) ts) = \<Union>(states_term ` set ts)"
| "states_term (Var x) = {}"

fun map_states_term :: "('q \<Rightarrow> 'r) \<Rightarrow> ('f + 'q, 'v) term \<Rightarrow> ('f + 'r, 'v) term" where
  "map_states_term g (Fun (Inr q) ts) = Fun (Inr (g q)) (map (map_states_term g) ts)"
| "map_states_term g (Fun (Inl f) ts) = Fun (Inl f) (map (map_states_term g) ts)"
| "map_states_term g (Var x) = ((Var x) :: ('f + 'r, 'v) term)"

lemma map_states_term_fterm_id [simp]:
  "map_states_term g (fterm t) = fterm t"
by (induct t) auto

lemma reachable_states_InlI:
  assumes "TA_rule f qs q \<in> ta_rules A"
    and "(q, q') \<in> (ta_eps A)\<^sup>*"
    and "length qs = length ts"
    and "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i)"
  shows "q' \<in> reachable_states A (Fun (Inl f) ts)"
using assms by auto

definition accessible :: "('q, 'f) ta \<Rightarrow> 'q \<Rightarrow> bool"
where
  "accessible A q \<longleftrightarrow> (\<exists> t :: ('f, 'v) term. ground (fterm t :: ('f + 'q, 'v) term) \<and>
                       q \<in> reachable_states A (fterm t))"

lemma accessibleE [elim]:
  assumes "accessible TYPE('v) A q"
  obtains t :: "('f, 'v) term" where "ground (fterm t)" and "q \<in> reachable_states A (fterm t)"
using assms unfolding accessible_def by auto

(* the language of an automaton *)
definition lang :: "('q, 'f) ta \<Rightarrow> ('f, 'v) terms"
where 
  "lang A = {t. ground t \<and> ta_final A \<inter> reachable_states A (fterm t) \<noteq> {}}"

lemma states_ground_instances_ta:
  assumes "(a, 0) \<in> F"
  shows "states (ground_instances_ta F c t) = (\<Union>r \<in> ground_instances_rules F c t. r_states r)"
using assms
apply (auto simp add: states_def ground_instances_ta_def)
apply (cases t)
apply (auto intro: exI [of _ "TA_rule a (replicate 0 (Fun c [])) (Fun c [])"]
            simp: sig_rules_def r_states_def)
done

lemma funas_ta_rule_sig_rules:
  shows "\<forall>r \<in> sig_rules F c. funas_ta_rule r \<subseteq> F"
proof -
  let ?srs = "sig_rules F c"
  { fix f qs q
    assume "TA_rule f qs q \<in> ?srs"
    then have "(f, length qs) \<in> F" unfolding sig_rules_def by auto
    then have "funas_ta_rule (TA_rule f qs q) \<subseteq> F" unfolding funas_ta_rule_def by simp
  }
  then have "\<forall>r \<in> ?srs. funas_ta_rule r \<subseteq> F" unfolding sig_rules_def by blast
  then show ?thesis by simp
qed

lemma funas_ta_rule_ground_instances_rules:
  "funas_term t \<subseteq> F \<Longrightarrow> r \<in> ground_instances_rules F c t \<Longrightarrow> funas_ta_rule r \<subseteq> F"
by (induct t) (auto simp: sig_rules_def funas_ta_rule_def)

lemma funas_ta_ground_instances_ta_Var:
  shows "funas_ta (ground_instances_ta F c (Var x)) \<subseteq> F"
unfolding ground_instances_ta_def funas_ta_def
using funas_ta_rule_sig_rules [of F "Fun c []"] by auto

lemma funas_ta_ground_instances_ta:
  assumes "funas_term t \<subseteq> F"
  shows "funas_ta (ground_instances_ta F c t) \<subseteq> F"
using funas_ta_rule_ground_instances_rules [OF assms]
by (auto simp: funas_ta_def ground_instances_ta_def)

lemma ta_final_ground_instances_ta [simp]:
  "ta_final (ground_instances_ta F c t) = {star c t}"
by (auto simp: ground_instances_ta_def)

lemma reachable_states_funas_ta:
  assumes "q \<in> reachable_states A (fterm t)"
  shows "funas_term t \<subseteq> funas_ta A"
using assms
proof (induct t arbitrary: q)
  case (Var x)
  show ?case by auto
next
  case (Fun f ts)
  obtain qs q where
    rule: "TA_rule f qs q \<in> ta_rules A" and
    len: "length qs = length (map fterm ts)" and
    qs: "\<forall>i < length (map fterm ts). qs ! i \<in> reachable_states A (map fterm ts ! i)"
    using Fun(2) by auto
  have f: "(f, length ts) \<in> funas_ta A" using funas_ta_def funas_ta_rule_def rule len
    by fastforce
  { fix u
    assume **: "u \<in> set ts"
    obtain i where "u = ts ! i" and "i < length ts" using in_set_idx [OF **] by auto
    then have "qs ! i \<in> reachable_states A (fterm u)" using qs by simp
    from Fun(1) [OF ** this] have "funas_term u \<subseteq> funas_ta A" .
  }
  then have "\<forall>u \<in> set ts. funas_term u \<subseteq> funas_ta A" by auto
  then show ?case using f by auto
qed

lemma ground_instances_ta_Var [simp]:
  "lang (ground_instances_ta F c (Var x)) = {t. fground F t}" (is "?L = ?R")
proof (intro equalityI subsetI)
  fix t
  assume t: "t \<in> ?L"
  then obtain q where "q \<in> reachable_states (ground_instances_ta F c (Var x)) (fterm t)"
    unfolding lang_def by auto
  moreover have "funas_ta (ground_instances_ta F c (Var x)) \<subseteq> F"
    by (simp add: funas_ta_ground_instances_ta_Var)
  ultimately have "funas_term t \<subseteq> F" using reachable_states_funas_ta by fast
  moreover have "ground t" using t unfolding lang_def by blast
  ultimately have "fground F t" by simp
  then show "t \<in> ?R" by auto
next
  fix t
  assume t: "t \<in> ?R"
  then show "t \<in> ?L"
  proof (induct t)
    case (Fun f ts)
    { fix i
      assume "i < length ts"
      moreover then have "ts ! i \<in> {t. fground F t}" using Fun by (auto dest!: nth_mem)
      ultimately have "ts ! i \<in> lang (ground_instances_ta F c (Var x))" using Fun by auto
      then have "Fun c [] \<in> reachable_states (ground_instances_ta F c (Var x)) (fterm (ts ! i))"
        by (auto simp: lang_def)
    }
    then have "Fun c [] \<in> reachable_states (ground_instances_ta F c (Var x)) (fterm (Fun f ts))"
      using Fun(2)
      by (auto simp: ground_instances_ta_def sig_rules_def
               intro: exI [of _ "Fun c []"] exI [of _ "replicate (length ts) (Fun c [])"])
    then show ?case using Fun by (auto simp: lang_def)
  qed (auto)
qed

lemma disj_states_imp_disj_ta:
  assumes "states A \<inter> states B = {}"
  shows "ta_final A \<inter> ta_final B = {} \<and> ta_rules A \<inter> ta_rules B = {} \<and>
    ta_eps A \<inter> ta_eps B = {}"
using assms by (auto simp: states_def r_states_def)

lemma ta_epsD:
  assumes "(q, q') \<in> ta_eps A"
  shows "{q, q'} \<subseteq> states A"
using assms by (auto simp: states_def)

lemma trancl_ta_epsD:
  assumes "(q, q') \<in> (ta_eps A)\<^sup>+"
  shows "{q, q'} \<subseteq> states A"
using assms by (induct) (auto dest: ta_epsD)

lemma disj_states_imp_disj_ta_eps:
  assumes "(q, q') \<in> (ta_eps A \<union> ta_eps B)\<^sup>+" and "states A \<inter> states B = {}"
  shows "(q, q') \<in> (ta_eps A)\<^sup>+ \<union> (ta_eps B)\<^sup>+"
using assms
by (induct) (auto, (blast intro: trancl.trancl_into_trancl dest: trancl_ta_epsD ta_epsD)+)

lemma r_states_in_states:
  assumes "TA_rule f qs q \<in> ta_rules A"
  shows "insert q (set qs) \<subseteq> states A"
using assms by (force simp: states_def r_states_def)

lemma reachable_states_in_states:
  assumes "\<forall>q. t \<noteq> State q"
  shows "reachable_states A t \<subseteq> states A"
using assms
  by (induct A t rule: reachable_states.induct)
     (auto dest!: rtranclD dest: r_states_in_states trancl_ta_epsD)

lemma states_in_disj_rs_imp_rs:
  assumes "length qs = length ts"
    and "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i) \<or> qs ! i \<in> reachable_states B (ts ! i)"
    and "states A \<inter> states B = {}"
    and "TA_rule f qs q' \<in> ta_rules A"
  shows "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i)"
proof -
  { fix i
    assume l: "i < length ts"
    then have *: "qs ! i \<in> reachable_states A (ts ! i) \<or> qs ! i \<in> reachable_states B (ts ! i)"
      using assms(2) by simp
    then have "qs ! i \<in> reachable_states A (ts ! i)"
    proof
      assume rs: "qs ! i \<in> reachable_states B (ts ! i)"
      have "insert q' (set qs) \<subseteq> states A" using r_states_in_states[OF assms(4)] .
      then have "qs ! i \<in> states A" using l assms(1) by auto
      then have not: "qs ! i \<notin> states B" using assms(3) by auto
      show "qs ! i \<in> reachable_states A (ts ! i)"
      proof (cases "\<exists>q. ts ! i = State q")
        case (True)
        then obtain q where "ts ! i = State q" by auto
        then have "(q, qs ! i) \<in> (ta_eps B)\<^sup>*" using reachable_states.simps(1)[of B q] rs by simp
        then show ?thesis
        proof (cases "q = qs ! i")
          case (True)
          then show ?thesis by (simp add: \<open>ts ! i = State q\<close>)
        next
          case (False)
          then show ?thesis using not \<open>(q, qs ! i) \<in> (ta_eps B)\<^sup>*\<close>
            by (auto simp: rtrancl_eq_or_trancl dest: trancl_ta_epsD)
        qed
      next
        case (False)
        then show ?thesis using not reachable_states_in_states rs by auto
      qed
    qed
  }
  then show ?thesis by auto
qed

lemma disj_reachable_states_eq_union:
  fixes t :: "('f + 'q, 'v) term"
  assumes "states A \<inter> states B = {}"
  shows "reachable_states (ta_union A B) t = reachable_states A t \<union> reachable_states B t"
  (is "?L = ?R1 \<union> ?R2")
proof (intro equalityI subsetI)
  fix q
  assume "q \<in> ?L"
  then have "q \<in> ?R1 \<or> q \<in> ?R2"
  using assms
  proof (induct "ta_union A B" t arbitrary: q rule: reachable_states.induct[case_names state "fun"])
    case (state q1)
    then show ?case
    by (auto simp: trancl_into_rtrancl dest!: rtranclD disj_states_imp_disj_ta_eps)
  next
    case ("fun" f ts)
    let ?C = "ta_union A B"
    from "fun"(2) obtain q' qs where trans: "TA_rule f qs q' \<in> ta_rules ?C" and
        eps: "(q', q) \<in> (ta_eps ?C)\<^sup>*" and l': "length qs = length ts" and
        qs: "\<forall>j < length ts. qs ! j \<in> reachable_states ?C (ts ! j)"
        by (auto)
    { fix i
      assume l: "i < length ts"
      from qs have "qs ! i \<in> reachable_states ?C (ts ! i)" using l by simp
      from "fun"(1)[OF l this "fun"(3)]
      have "qs ! i \<in> reachable_states A (ts ! i) \<or> qs ! i \<in> reachable_states B (ts ! i)" .
    }
    then have qs':
      "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i) \<or> qs ! i \<in> reachable_states B (ts ! i)"
      by simp
    have eps': "(q', q) \<in> (ta_eps A)\<^sup>* \<union> (ta_eps B)\<^sup>*" using eps "fun"(3)
      by (auto simp: trancl_into_rtrancl dest!: rtranclD disj_states_imp_disj_ta_eps)
      
    from disj_states_imp_disj_ta[OF "fun"(3)] have "ta_rules A \<inter> ta_rules B = {}" by auto
    then have *: "TA_rule f qs q' \<in> ta_rules A \<union> ta_rules B" using trans by (auto)
    then have "q \<in> reachable_states A (Fun (Inl f) ts) \<union> reachable_states B (Fun (Inl f) ts)"
    proof (cases "TA_rule f qs q' \<in> ta_rules A")
      case (True)
      then show ?thesis
      proof (cases "(q', q) \<in> (ta_eps A)\<^sup>*")
        case (True)
        then show ?thesis
        proof (cases "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i)")
          case (True)
          then show ?thesis
            using \<open>TA_rule f qs q' \<in> ta_rules A\<close> \<open>(q', q) \<in> (ta_eps A)\<^sup>*\<close> l'
            by auto
        next
          case (False)
          then show ?thesis
            using \<open>TA_rule f qs q' \<in> ta_rules A\<close> and qs' and "fun.prems"(2) and
            states_in_disj_rs_imp_rs[OF l', of A B f q'] 
            by blast
        qed
      next
        case (False)
        then show ?thesis
        using eps' and "fun.prems"(2) and
        r_states_in_states[OF \<open>TA_rule f qs q' \<in> ta_rules A\<close>]
        by (auto simp: rtrancl_eq_or_trancl) (auto dest!: trancl_ta_epsD)
      qed
    next
      case (False)
      then have transb: "TA_rule f qs q' \<in> ta_rules B" using * by simp
      with states_in_disj_rs_imp_rs[OF l', of B A f q'] fun.prems(2) qs'
        have qsb: "\<forall>i < length ts. qs ! i \<in> reachable_states B (ts ! i)" by blast
      show ?thesis
      proof (cases "(q', q) \<in> (ta_eps A)\<^sup>*")
        case (True)
        then show ?thesis
          using transb True qsb fun.prems(2) l'
          by (auto simp: rtrancl_eq_or_trancl) (auto dest!: r_states_in_states trancl_ta_epsD)
      next
        case (False)
        then have epsb: "(q', q) \<in> (ta_eps B)\<^sup>*" using eps' by auto
        with states_in_disj_rs_imp_rs[OF l', of B A f q']
        show ?thesis
          using \<open>TA_rule f qs q' \<in> ta_rules B\<close> fun.prems(2) qs' l'
          by (simp only: reachable_states.simps) blast
      qed
    qed
    then show ?case by simp
  qed (auto)
  then show "q \<in> ?R1 \<union> ?R2" by simp
next
  fix q
  assume "q \<in> ?R1 \<union> ?R2"
  then show "q \<in> ?L"
  proof
    assume "q \<in> ?R1"
    then show "q \<in> ?L" using assms
    proof (induct A t arbitrary: q rule: reachable_states.induct[case_names state "fun"])
      case (state A q1)
      then show ?case by (auto simp: in_rtrancl_UnI)
    next
      case ("fun" A f ts)
      from "fun"(2) obtain q' qs where trans: "TA_rule f qs q' \<in> ta_rules A" and
        eps: "(q', q) \<in> (ta_eps A)\<^sup>*" and l': "length qs = length ts" and
        qs: "\<forall>j < length ts. qs ! j \<in> reachable_states A (ts ! j)"
        by (auto)
      from trans have "TA_rule f qs q' \<in> ta_rules (ta_union A B)" by (auto)
      moreover
      from eps have "(q', q) \<in> (ta_eps (ta_union A B))\<^sup>*" by (auto simp: in_rtrancl_UnI)
      moreover
      { fix i
        assume l: "i < length ts"
        from qs have "qs ! i \<in> reachable_states A (ts ! i)" using l by simp
        from "fun"(1)[OF l this "fun"(3)]
        have "qs ! i \<in> reachable_states (ta_union A B) (ts ! i)" .
      }
      then have "\<forall> i < length ts. qs ! i \<in> reachable_states (ta_union A B) (ts ! i)" by simp
      ultimately show ?case using l' by auto
    qed (auto)
  next
    assume "q \<in> ?R2"
    then show "q \<in> ?L" using assms
    proof (induct B t arbitrary: q rule: reachable_states.induct[case_names state "fun"])
      case (state B q1)
      then show ?case by (auto simp: in_rtrancl_UnI)
    next
      case ("fun" B f ts)
      from "fun"(2) obtain q' qs where trans: "TA_rule f qs q' \<in> ta_rules B" and
        eps: "(q', q) \<in> (ta_eps B)\<^sup>*" and l': "length qs = length ts" and
        qs: "\<forall>j < length ts. qs ! j \<in> reachable_states B (ts ! j)"
        by (auto)
      from trans have "TA_rule f qs q' \<in> ta_rules (ta_union A B)" by (auto)
      moreover
      from eps have "(q', q) \<in> (ta_eps (ta_union A B))\<^sup>*" by (auto simp: in_rtrancl_UnI)
      moreover
      { fix i
        assume l: "i < length ts"
        from qs have "qs ! i \<in> reachable_states B (ts ! i)" using l by simp
        from "fun"(1)[OF l this "fun"(3)]
        have "qs ! i \<in> reachable_states (ta_union A B) (ts ! i)" .
      }
      then have "\<forall> i < length ts. qs ! i \<in> reachable_states (ta_union A B) (ts ! i)" by simp
      ultimately show ?case using l' by auto
    qed (auto)
  qed
qed

lemma disj_states_imp_ta_final_reachable_states_disj:
  assumes "states A \<inter> states B = {}"
    and "\<forall> q. t \<noteq> State q"
  shows "ta_final A \<inter> reachable_states B t = {}"
proof -
  have "ta_final A \<subseteq> states A" by (simp add: states_def r_states_def)
  with assms(1) reachable_states_in_states[OF assms(2), of B]
  show ?thesis by auto
qed

lemma lang_disj_union_eq_union_lang:
  assumes "states A \<inter> states B = {}"
  shows "lang (ta_union A B) = lang A \<union> lang B" (is "?L = ?R1 \<union> ?R2")
using assms
proof (intro equalityI subsetI)
  fix t
  assume "t \<in> ?L"
  moreover have "{t. ground t \<and> ta_final A \<inter> reachable_states B (fterm t) \<noteq> {}} = {}"
  and "{t. ground t \<and> ta_final B \<inter> reachable_states A (fterm t) \<noteq> {}} = {}"
    by (simp_all add: assms ac_simps ground_term_no_State
        disj_states_imp_ta_final_reachable_states_disj)
  ultimately show "t \<in> ?R1 \<union> ?R2"
    by (auto simp: lang_def disj_reachable_states_eq_union assms)
next
  fix t
  assume "t \<in> ?R1 \<union> ?R2"
  then show "t \<in> ?L"
  using disj_reachable_states_eq_union[OF assms] by (auto simp: lang_def)
qed

lemma disj_union_empty_ta_final_imp_lang:
  assumes "ta_final B = {}"
    and "states A \<inter> states B = {}"
  shows "lang (ta_union A B) = lang A" (is "?L = ?R")
using assms lang_disj_union_eq_union_lang
by (auto simp: lang_def) (fastforce)+

lemma reachable_states_map_states_iff':
  fixes t :: "('f + 'q, 'v) term"
  assumes "inj g"
  shows "g q \<in> reachable_states (map_states g A) (map_states_term g t) \<longleftrightarrow>
         q \<in> reachable_states A t"
proof
  fix q
  assume "g q \<in> reachable_states (map_states g A) (map_states_term g t)"
  then show "q \<in> reachable_states A t"
  proof (induct t arbitrary: q rule: reachable_states.induct)
    case (2 A f ts)
    have "g q \<in> reachable_states (map_states g A) (Fun (Inl f) (map (map_states_term g) ts))"
      using 2(2) by auto
    then obtain q'' qs where rule: "TA_rule f qs q'' \<in> ta_rules (map_states g A)"
      and eps: "(q'', g q) \<in> (ta_eps (map_states g A))\<^sup>*"
      and l: "length qs = length (map (map_states_term g) ts)"
      and qs: "\<forall>i < length (map (map_states_term g) ts). qs ! i \<in> reachable_states (map_states g A)
                 (map (map_states_term g) ts ! i)" by auto
    from ta_rules_map_statesE [OF rule] obtain qs' p where qs': "qs = map g qs'" and "q'' = g p"
      and rule': "TA_rule f qs' p \<in> ta_rules A" by blast
    have l': "length (map (map_states_term g) ts) = length ts" by simp
    { fix i
      assume i: "i < length ts"
      then have "qs ! i \<in> reachable_states (map_states g A) (map_states_term g (ts ! i))"
        using qs i by auto
      then have "g (qs' ! i) \<in> reachable_states (map_states g A) (map_states_term g (ts ! i))"
        using qs' l i by simp
      from 2(1) [OF i this] have "qs' ! i \<in> reachable_states A (ts ! i)" .
    }
    then have qs'': "\<forall>i < length ts. qs' ! i \<in> reachable_states A (ts ! i)" by auto
    have "(g p, g q) \<in> (ta_eps (map_states g A))\<^sup>*" using eps \<open>q'' = g p\<close> by auto
    then have eps': "(p, q) \<in> (ta_eps A)\<^sup>*" using assms by auto
    have len: "length ts = length qs'" using l l' qs' by simp
    show ?case using rule' qs'' eps' len by auto
  qed (auto simp: assms)
next
  fix q
  assume "q \<in> reachable_states A t"
  then show "g q \<in> reachable_states (map_states g A) (map_states_term g t)"
  proof (induct t arbitrary: q rule: reachable_states.induct)
    case (2 A f ts)
    from 2(2) obtain q' qs where rule: "TA_rule f qs q' \<in> ta_rules A"
      and eps: "(q', q) \<in> (ta_eps A)\<^sup>*"
      and len: "length qs = length ts"
      and qs: "\<forall>i < length ts. qs ! i \<in> reachable_states A (ts ! i)"
      by auto
    have l': "length ts = length (map (map_states_term g) ts)" by simp
    { fix i
      assume i: "i < length ts"
      then have "qs ! i \<in> reachable_states A (ts ! i)" using qs by simp
      from 2(1) [OF i this] have "map g qs ! i \<in> reachable_states (map_states g A)
        (map (map_states_term g) ts ! i)" by (simp add: i len) 
    }
    then have "\<forall>i < length ts. map g qs ! i \<in> reachable_states (map_states g A)
      (map (map_states_term g) ts ! i)" by blast
    then have qs': "\<forall>i < length (map (map_states_term g) ts). map g qs ! i \<in>
      reachable_states (map_states g A) (map (map_states_term g) ts ! i)" unfolding l' by blast
    from eps have eps': "(g q', g q) \<in> (ta_eps (map_states g A))\<^sup>*"
      using ta_eps_map_states_iff [OF assms, symmetric, THEN iffD1] by blast
    have rule': "TA_rule f (map g qs) (g q') \<in> ta_rules (map_states g A)"
      using ta_rules_map_states_iff [OF assms, symmetric, THEN iffD1, OF rule] .
    have len': "length (map g qs) = length (map (map_states_term g) ts)" using len by simp
    from reachable_states_InlI [OF rule' eps' len' qs']
    have "g q \<in> reachable_states (map_states g A)
      (Fun (Inl f) (map (map_states_term g) ts))" by blast
    then show ?case by fastforce
  qed (auto simp: assms)
qed

lemma ground_map_states_term_iff:
  shows "ground t \<longleftrightarrow> ground (map_states_term g t)"
  (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  then show ?R
  proof (induct t)
    case (Fun f ts)
    show ?case
    proof (cases "\<exists> f'. f = Inl f'")
      case (True)
      then obtain f' where "f = Inl f'" by blast
      then have "ground (Fun (Inl f') (map (map_states_term g) ts))" using Fun by auto
      then show ?thesis by (simp add: \<open>f = Inl f'\<close>)
    next
      case (False)
      then obtain q where "f = Inr q" using sum.exhaust_sel by blast
      then have "ground (Fun (Inr (g q)) (map (map_states_term g) ts))" using Fun by auto
      then show ?thesis by (simp add: \<open>f = Inr q\<close>) 
    qed
  qed (auto)
next
  assume ?R
  then show ?L
  proof (induct t)
    case (Fun f ts)
    show ?case
    proof (cases "\<exists> f'. f = Inl f'")
      case (True)
      then obtain f' where "f = Inl f'" by blast
      then have "ground (Fun (Inl f') (map (map_states_term g) ts))" using Fun by auto
      then show ?thesis by (simp add: Fun(1))
    next
      case (False)
      then obtain q where "f = Inr q" using sum.exhaust_sel by blast
      then have "ground (Fun (Inr (g q)) (map (map_states_term g) ts))" using Fun by auto
      then show ?thesis by (simp add: Fun(1))
    qed
  qed (auto)
qed

lemma reachable_states_map_states_exists_map_states_term:
  assumes "g q \<in> reachable_states (map_states g D) t'"
  shows "\<exists>t. t' = (map_states_term g t)"
using assms
proof (induct "map_states g D" t' arbitrary: q rule: reachable_states.induct)
  case (1 q')
  then have "(q', g q) \<in> (ta_eps (map_states g D))\<^sup>*" by simp
  then obtain q'' where "q' = g q''" by blast
  with 1 have "State q' = map_states_term g (State q'')" by simp 
  then show ?case by blast
next
  case (2 f ts)
  then obtain q' qs where rule: "TA_rule f qs q' \<in> ta_rules (map_states g D)"
    and eps: "(q', g q) \<in> (ta_eps (map_states g D))\<^sup>*"
    and len: "length qs = length ts"
    and qs: "\<forall>i < length ts. qs ! i \<in> reachable_states (map_states g D) (ts ! i)"
    by auto
  from ta_rules_map_statesE [OF rule] obtain qs' q'' where *: "qs = map g qs'" and "q' = g q''"
    and rule': "TA_rule f qs' q'' \<in> ta_rules D" by blast
  have **: "\<forall>i < length ts. map g qs' ! i \<in> reachable_states (map_states g D) (ts ! i)"
    using * qs by blast
  { fix i
    assume i: "i < length ts"
    with ** have "map g qs' ! i \<in> reachable_states (map_states g D) (ts ! i)" by blast
    then have "g (qs' ! i) \<in> reachable_states (map_states g D) (ts ! i)" using * i len by auto
    from 2(1) [OF i this] have "\<exists>t. map_states_term g t = ts ! i" by fastforce
  }
  then obtain h where 3: "\<forall>i < length ts. map_states_term g (h i) = ts ! i" by metis
  moreover def us \<equiv> "map h [0..<length ts]"
  ultimately have "map (map_states_term g) us = map (op ! ts) [0..<length ts]"  by simp
  then have "map (map_states_term g) us = ts" by (simp add: map_nth)
  then have "Fun (Inl f) ts = map_states_term g (Fun (Inl f) us)" by simp
  then show ?case by blast
qed (auto)

lemma accessible_map_states:
  assumes "inj g"
  shows "accessible TYPE ('v) D q \<longleftrightarrow> accessible TYPE ('v) (map_states g D) (g q)"
  (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  then obtain t :: "('c, 'v) term" where "ground t" and 2: "q \<in> reachable_states D (fterm t)"
    by (auto simp: accessible_def)
  then have 1: "ground (fterm t)" by simp
  from reachable_states_map_states_iff' [OF assms, symmetric, THEN iffD1, OF 2]
    have "g q \<in> reachable_states (map_states g D) (map_states_term g (fterm t))" .
  moreover have "ground (map_states_term g (fterm t))" using ground_map_states_term_iff [THEN iffD1, OF 1]
    by auto
  ultimately show ?R unfolding accessible_def by (auto simp: accessible_def)
next
  assume ?R
  then obtain t' :: "('c, 'v) term" where g: "ground (fterm t')"
    and *: "g q \<in> reachable_states (map_states g D)  (fterm t')"
      by (auto simp: accessible_def)
  from reachable_states_map_states_exists_map_states_term [OF *] obtain t where
    t: "fterm t' = (map_states_term g (fterm t))" by simp
  have "g q \<in> reachable_states (map_states g D) (map_states_term g (fterm t))" using * unfolding t .
  from reachable_states_map_states_iff' [OF assms, THEN iffD1, OF this]
    have **: "q \<in> reachable_states D (fterm t)" .
  have "ground (fterm t)" using g t by (metis ground_map_funs_term ground_map_states_term_iff)
  then show ?L using ** accessible_def by fast
qed

lemma reachable_states_map_states_iff [simp]:
  fixes t :: "('f, 'v) term"
  assumes "inj g"
  shows "reachable_states (map_states g A) (fterm t) = g ` (reachable_states A (fterm t))"
  (is "?L t = ?R t")
proof (induct t)
  case (Fun f ts)
  then have "?R (Fun f ts) \<subseteq> ?L (Fun f ts)"
  apply auto
  apply (rule_tac x = "g q" in exI)
  apply (rule_tac x = "map g qs" in exI)
  apply (auto simp: assms)
  done
  moreover
  have  "?L (Fun f ts) \<subseteq> ?R (Fun f ts)"
  proof
    fix q :: "'b"
    let ?A = "map_states g A :: ('b, 'f) ta"
    assume "q \<in> ?L (Fun f ts)"
    then obtain qs q' where trans: "TA_rule f qs q' \<in> ta_rules ?A" and
                eps: "(q', q) \<in> (ta_eps ?A)\<^sup>*" and
                len: "length qs = length (map fterm ts)" and
                qs: "\<forall>i < length (map fterm ts). qs ! i \<in> reachable_states ?A (fterm (ts ! i))"
      by auto
    have 1: "TA_rule f (map (inv g) qs) ((inv g) q') \<in> ta_rules A" and q': "q' \<in> range g"
      using trans
      by (auto simp: map_states_def map_r_states_def comp_def assms)
    have 2: "((inv g) q', (inv g) q) \<in> (ta_eps A)\<^sup>*" and q: "q \<in> range g"
      using eps and assms and q'
      by (induct) (auto simp: map_states_def)
    have 3: "length (map (inv g) qs) = length (map fterm ts)" using len by auto
    { fix i
      assume "i < length ts"
      then have "ts ! i \<in> set ts" by simp
      from Fun[OF this]
      have "reachable_states ?A (fterm (ts ! i)) = g ` reachable_states A (fterm (ts ! i))" .
    }
    then have "\<forall>i < length ts.
      reachable_states ?A (fterm (ts ! i)) = g ` reachable_states A (fterm (ts ! i))" by simp
    with qs have "\<forall>i < length (map fterm ts). qs ! i \<in> g ` reachable_states A (fterm (ts ! i))"
      by simp
    then have 4:
      "\<forall>i < length (map fterm ts). map (inv g) qs ! i \<in> reachable_states A (map fterm ts ! i)"
      using len assms by auto
    
    have "((inv g) q) \<in> reachable_states A (fterm (Fun f ts))"
      using reachable_states_InlI[OF 1 2 3 4] by simp
    from imageI[OF this, of g] show "q \<in> ?R (Fun f ts)"
      using assms and f_inv_into_f[OF q] by simp
  qed
  ultimately
  show ?case by blast
qed (simp)

lemma ta_final_map_states_ex:
  assumes "q' \<in> ta_final (map_states g A)"
  shows "\<exists>q. g q = q'"
proof -
  from assms(1) have "q' \<in> g ` ta_final A" unfolding map_states_def by simp
  then show ?thesis by blast
qed

lemma accepted_map_states_iff [simp]:
  assumes "inj g"
  shows "ta_final (map_states g A) \<inter> g ` reachable_states A (fterm t) =
   g ` (ta_final A \<inter> reachable_states A (fterm t))"
using assms injD by fastforce

lemma lang_map_states_iff:
  assumes "inj g"
  shows "lang (map_states g A) = lang A" (is "?L = ?R")
using assms accepted_map_states_iff unfolding lang_def by fastforce

subsection \<open>The one-step move relation of a tree automaton\<close>

(* the one step move relation for TAs *)
inductive_set move :: "('q, 'f) ta \<Rightarrow> ('f + 'q, 'v) term rel" for A
where
  trans: "TA_rule f qs q \<in> ta_rules A \<Longrightarrow> (C\<langle>fqterm f qs\<rangle>, C\<langle>State q\<rangle>) \<in> move A"
| eps: "(q, q') \<in> (ta_eps A)\<^sup>* \<Longrightarrow> (C\<langle>State q\<rangle>, C\<langle>State q'\<rangle>) \<in> move A"

lemma move_mono [mono_set]:
  assumes "B \<subseteq> C"
  shows "(move (D\<lparr> ta_rules := B \<rparr>))\<^sup>* \<subseteq> (move (D\<lparr> ta_rules := C \<rparr>))\<^sup>*"
using assms by (auto elim!: move.induct intro!: move.intros rtrancl_mono)

lemma move_mono':
  assumes "(s, t) \<in> move U"
    and "ta_rules U \<subseteq> ta_rules I"
    and "ta_eps U \<subseteq> ta_eps I"
  shows "(s, t) \<in> move I"
using assms
  by (induct rule: move.induct) (auto simp: move.trans move.eps subsetD dest: rtrancl_mono)

lemma move_mono'_rtrancl:
  assumes "(s, t) \<in> (move U)\<^sup>*"
    and "ta_rules U \<subseteq> ta_rules I"
    and "ta_eps U \<subseteq> ta_eps I"
  shows "(s, t) \<in> (move I)\<^sup>*"
using assms move_mono' [OF _ assms(2) assms(3)] by (induct) force+

lemma move_mono'_trancl:
  assumes "(s, t) \<in> (move U)\<^sup>+"
    and "ta_rules U \<subseteq> ta_rules I"
    and "ta_eps U \<subseteq> ta_eps I"
  shows "(s, t) \<in> (move I)\<^sup>+"
using assms move_mono'_rtrancl [OF _ assms(2) assms(3)] move_mono' [OF _ assms(2) assms(3)]
  rtrancl_into_trancl1 tranclD2 by metis

lemma move_ctxt:
  assumes "(s, t) \<in> move A"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> move A"
proof -
  note * = move.intros(1)[of _ _ _ _ "C \<circ>\<^sub>c D" for C D, simplified]
           move.intros(2)[of _ _ _ "C \<circ>\<^sub>c D" for C D, simplified]
  show ?thesis using assms
    by (induct) (auto intro: *)
qed

lemma move_ctxt_rtrancl:
  assumes "(s, t) \<in> (move A)\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (move A)\<^sup>*"
using assms by (induct) (auto simp: move_ctxt rtrancl.rtrancl_into_rtrancl)

lemma ctxt_closed_move: "ctxt_closed (move A)"
proof
  fix s t :: "('a + 'b, 'd) term" and C
  assume "(s, t) \<in> move A"
  then show "(s, t) \<in> move A \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> move A" by (rule move_ctxt)
qed

lemma move_ground_ground:
  assumes "(s, t) \<in> move A" and "ground t"
  shows "ground s"
using assms
by (induct rule: move.induct) (simp add: move.intros)+

lemma rtrancl_move_ground_ground:
  assumes "(s, t) \<in> (move A)\<^sup>*" and "ground t"
  shows "ground s"
using assms move_ground_ground
by (induct) (auto)

lemma no_non_ground_move:
  assumes "\<not> ground s" and "ground t"
  shows "(s, t) \<notin> move A"
using assms rtrancl_move_ground_ground by auto

lemma distr_move':
  assumes "(s, Fun (Inl f) ts) \<in> move A"
  shows "\<exists>ss. s = Fun (Inl f) ss \<and> length ts = length ss \<and>
    (\<forall>i < length ss. ss ! i = ts ! i \<or> (ss ! i, ts ! i) \<in> move A)"
proof -
  { fix C q v E
    assume move: "(\<exists>q'. v = State q' \<and> (q', q) \<in> (ta_eps A)\<^sup>*)
        \<or> (\<exists>g qs. v = fqterm  g qs \<and> TA_rule g qs q \<in> ta_rules A)"
      and s: "s = C\<langle>v\<rangle>"
      and C: "Fun (Inl f) ts = C\<langle>State q\<rangle>"
    from C obtain us vs D where D: "C = More (Inl f) us D vs" by (cases C) (auto)
    with s obtain ss where 1: "s = Fun (Inl f) ss" by simp
    from D s 1 have ss: "ss = us @ D\<langle>v\<rangle> # vs" by simp
    then obtain i where iD: "ss ! i = D\<langle>v\<rangle>"
      and us: "take i ss = us" and vs: "drop (Suc i) ss = vs"
      using nth_append_length by fastforce
    from C D have *: "ts = us @ D\<langle>State q\<rangle> # vs" by auto
    { fix j
      assume "j < length ss"
      from us ss not_le have len: "length us = i" by fastforce
      have "ss ! j = ts ! j \<or> (ss ! j, ts ! j) \<in> move A"
      proof (cases "j = i")
        case (True)
        then show ?thesis
          using move move.trans * len iD by (fastforce simp: move.eps r_into_rtrancl)
      qed (simp add: * len append_Cons_nth_not_middle rtrancl_eq_or_trancl ss)
    }
    then have "\<exists> ss. ?thesis" by (simp add: * 1 ss)
  }
  then have
    "\<And> C q v. (\<exists>q'. v = State q' \<and> (q', q) \<in> (ta_eps A)\<^sup>*) \<or>
      (\<exists>g qs. v = fqterm g qs \<and> TA_rule g qs q \<in> ta_rules A) \<Longrightarrow>
      Fun (Inl f) ts = C\<langle>State q\<rangle> \<Longrightarrow> s = C\<langle>v\<rangle> \<Longrightarrow>
      \<exists>ss. s = Fun (Inl f) ss \<and> length ts = length ss \<and> 
      (\<forall>i < length ss. ss ! i = ts ! i \<or> (ss ! i, ts ! i) \<in> move A)" by auto
  with assms show ?thesis by (cases rule: move.cases) (blast)+
qed

lemma distr_move:
  assumes "(s, Fun (Inl f) ts) \<in> move A"
  shows "\<exists>ss. s = Fun (Inl f) ss \<and> length ts = length ss \<and>
    (\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*)"
using distr_move' [OF assms] by fastforce

lemma distr_moves':
  assumes "(s, Fun (Inl f) ts) \<in> (move A)\<^sup>+"
  shows "\<exists>ss. s = Fun (Inl f) ss \<and> length ts = length ss \<and>
    (\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*)"
using assms
by (induct s "Fun (Inl f) ts" arbitrary: ts rule: trancl.induct)
   (auto dest!: distr_move, metis (mono_tags) rtrancl_trans)

lemma distr_moves:
  assumes "(s, Fun (Inl f) ts) \<in> (move A)\<^sup>*"
  shows "\<exists>ss. s = Fun (Inl f) ss \<and> length ts = length ss \<and>
    (\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*)"
using assms
by (induct s "Fun (Inl f) ts" arbitrary: ts rule: rtrancl.induct)
   (auto dest!: distr_move, metis (mono_tags) rtrancl_trans)

subsection \<open>Alternative move relation "move'"\<close>

inductive_set move' :: "('q, 'f) ta \<Rightarrow> ('f + 'q, 'v) term rel" for A
where
  trans: "TA_rule f qs q \<in> ta_rules A \<Longrightarrow> (fqterm f qs, State q) \<in> move' A"
| eps: "(q, q') \<in> (ta_eps A)\<^sup>* \<Longrightarrow> (State q, State q') \<in> move' A"
| Fun: "(s, t) \<in> move' A \<Longrightarrow> (Fun g (ss @ s # ts), Fun g (ss @ t # ts)) \<in> move' A"

lemma move'_intros:
  "TA_rule f qs q \<in> ta_rules A \<Longrightarrow> (C\<langle>fqterm f qs\<rangle>, C\<langle>State q\<rangle>) \<in> move' A"
  "(q, q') \<in> (ta_eps A)\<^sup>* \<Longrightarrow> (C\<langle>State q\<rangle>, C\<langle>State q'\<rangle>) \<in> move' A"
by (induct_tac[!] C) (auto intro: move'.intros)

lemma move_eq_move':
  shows "move A = move' A"
proof (intro equalityI subrelI)
  fix s t :: "('a + 'b, 'v) term"
  assume "(s, t) \<in> move A"
  then show "(s, t) \<in> move' A"
    by cases (auto simp: move'_intros)
next
  note * = move_ctxt[of _ _ _ "More f ss \<box> ts" for f ss ts, simplified]
           move.intros[where C = "\<box>", simplified]
  fix s t :: "('a + 'b, 'v) term"
  assume "(s, t) \<in> move' A"
  then show "(s, t) \<in> move A"
    by (induct) (auto intro: *)
qed

lemma no_move_to_nonempty_Inr:
  assumes "(fterm s, Fun (Inr q) (t # ts)) \<in> move A"
  shows "False"
using assms
apply (cases)
defer
using fterm_contains_no_Inr [of s]
apply metis
apply (case_tac C)
apply auto
apply (cases s)
apply auto
done

lemma fterm_args_fterm:
  assumes "fterm s = Fun (Inl f) ss"
  shows "\<exists>ss'. ss = map fterm ss'"
proof -
  from assms obtain f' ss' where "fterm (Fun f' ss') = Fun (Inl f) ss"
    by (metis Term.term.simps(9) funs_term_list.elims term.distinct(1))
  then have "f' = f" and "ss = map fterm ss'" by auto
  then show ?thesis by blast
qed

lemma no_move_to_term_containing_nonempty_Inr:
  assumes "(fterm s, C\<langle>Fun (Inr q) (t # ts)\<rangle>) \<in> move A"
  shows "False"
using assms
proof (induct C arbitrary: s q t ts)
  case (Hole)
  then show ?case using no_move_to_nonempty_Inr by fastforce
next
  case (More f us E vs)
  then show ?case
  proof (cases f)
    case (Inr f')
    have "(fterm s, (More (Inr f') us E vs)\<langle>Fun (Inr q) (t # ts)\<rangle>) \<in> move A"
      using More(2) unfolding Inr .
    moreover have "(More (Inr f') us E vs)\<langle>Fun (Inr q) (t # ts)\<rangle> =
      Fun (Inr f') (us @ E\<langle>Fun (Inr q) (t # ts)\<rangle> # vs)" by auto
    moreover then obtain w ws where "w # ws = us @ E\<langle>Fun (Inr q) (t # ts)\<rangle> # vs"
      by (metis Nil_is_append_conv list.exhaust)
    ultimately have "(fterm s, Fun (Inr f') (w # ws)) \<in> move A" by simp
    from no_move_to_nonempty_Inr [OF this] show ?thesis .
  next
    case (Inl f')
    let ?q' = "Fun (Inr q) (t # ts)"
    let ?ws = "us @ E\<langle>?q'\<rangle> # vs"
    have "(fterm s, (More (Inl f') us E vs)\<langle>?q'\<rangle>) \<in> move A"
      using More(2) unfolding Inl .
    then have "(fterm s, Fun (Inl f') ?ws) \<in> move A" by simp
    from distr_move' [OF this] obtain ss where s: "fterm s = Fun (Inl f') ss" and
       len: "length ?ws = length ss" and
       *: "\<forall>i < length ss. ss ! i = ?ws ! i \<or> (ss ! i, ?ws ! i) \<in> move A" by blast
    obtain ss' where **: "ss = map fterm ss'" using fterm_args_fterm [OF s] by blast
    from * obtain j where
      "map fterm ss' ! j = E\<langle>?q'\<rangle> \<or> (map fterm ss' ! j, E\<langle>?q'\<rangle>) \<in> move A" (is "?s' = _ \<or> _")
      and len': "j < length (map fterm ss')"
      using ** len by (metis (no_types) add_Suc_right length_Cons length_append length_map
                       less_add_Suc1 nth_append_length)
    moreover
    { assume "(map fterm ss' ! j, E\<langle>?q'\<rangle>) \<in> move A"
      then have "(fterm (ss' ! j), E\<langle>?q'\<rangle>) \<in> move A" using len' by auto
      from More(1) [OF this] have ?thesis .
    }
    moreover
    { assume "map fterm ss' ! j = E\<langle>?q'\<rangle>"
      then have "fterm (ss' ! j) = E\<langle>?q'\<rangle>" using len' by auto
      then have ?thesis using fterm_contains_no_Inr by fast
    }
    ultimately show ?thesis by blast
  qed
qed

inductive_set wf_terms where
  Var: "(Var x) \<in> wf_terms"
| State: "(State q) \<in> wf_terms"
| Fun: "\<forall>t \<in> set ts. t \<in> wf_terms \<Longrightarrow> Fun (Inl f) ts \<in> wf_terms"

inductive_set wf_ctxts where
  Hole: "\<box> \<in> wf_ctxts"
| More: "\<forall>t \<in> set (us @ vs). t \<in> wf_terms \<Longrightarrow> E \<in> wf_ctxts \<Longrightarrow> More (Inl f) us E vs \<in> wf_ctxts"

lemma wf_terms_subtermeq:
  assumes "s \<in> wf_terms"
    and "s \<unrhd> t"
  shows "t \<in> wf_terms"
using assms
proof (induct s arbitrary: t)
  case (Var x)
  then have *: "t = Var x" by simp
  show ?case unfolding * by (simp add: wf_terms.intros)
next
  case (State q)
  then have *: "t = State q" using Fun_Nil_supt by fastforce
  show ?case unfolding * by (simp add: wf_terms.intros)
next
  case (Fun ts f)
  then show ?case by (metis subterm.le_less supt_Fun_imp_arg_supteq wf_terms.simps)
qed

lemma wf_terms_Fun_Inl_args:
  assumes "Fun (Inl f) ts \<in> wf_terms"
  shows "\<forall>t \<in> set ts. t \<in> wf_terms"
using assms by (cases) blast

lemma wf_terms_wf_ctxts:
  assumes "s \<in> wf_terms"
    and "s = C\<langle>t\<rangle>"
  shows "C \<in> wf_ctxts"
using assms
proof (induct C arbitrary: s)
  case (Hole)
  show ?case by (simp add: wf_ctxts.intros)
next
  case (More f us E vs)
  let ?ss = "us @ E\<langle>t\<rangle> # vs"
  have *: "s = Fun f ?ss" using More by simp
  obtain f' where **: "f = Inl f'" using More(2) unfolding * by (cases) blast+
  have "s = Fun (Inl f') ?ss" using * ** by auto
  then have "E\<langle>t\<rangle> \<in> wf_terms" using More.prems(1) wf_terms_Fun_Inl_args by force
  from More(1) [OF this] have E: "E \<in> wf_ctxts" by simp
  from More(2) have "\<forall>t \<in> set ?ss. t \<in> wf_terms" unfolding * ** using wf_terms_Fun_Inl_args by fast
  then have "\<forall>t \<in> set (us @ vs). t \<in> wf_terms" by force
  then show ?case using E unfolding ** by (simp add: E wf_ctxts.More)
qed

lemma wf_ctxts_wf_terms_wf_terms:
  assumes "C \<in> wf_ctxts"
    and "s \<in> wf_terms"
  shows "C\<langle>s\<rangle> \<in> wf_terms"
using assms
proof (induct C)
  case (Hole)
  then show ?case by simp
next
  case (More us vs E f)
  let ?ss = "us @ E\<langle>s\<rangle> # vs"
  have "E\<langle>s\<rangle> \<in> wf_terms" using More(3) [OF More(4)] .
  with More(1) have "\<forall>t \<in> set ?ss. t \<in> wf_terms" by auto
  then show ?case by (simp add: wf_terms.Fun)
qed

lemma move_preserves_wf_terms:
  assumes "(s, t) \<in> move A"
    and "s \<in> wf_terms"
  shows "t \<in> wf_terms"
using assms
proof (cases)
  case (trans f qs q C)
  have "C \<in> wf_ctxts" using wf_terms_wf_ctxts [OF assms(2) trans(1)] .
  moreover have "State q \<in> wf_terms" by (simp add: wf_terms.State)
  ultimately show ?thesis using trans(2) wf_ctxts_wf_terms_wf_terms by blast
next
  case (eps q q' C)
  have "C \<in> wf_ctxts" using wf_terms_wf_ctxts [OF assms(2) eps(1)] .
  moreover have "State q' \<in> wf_terms" by (simp add: wf_terms.State)
  ultimately show ?thesis using eps(2) wf_ctxts_wf_terms_wf_terms by blast
qed

lemma moves_preserve_wf_terms:
  assumes "(s, t) \<in> (move A)\<^sup>*"
    and "s \<in> wf_terms"
  shows "t \<in> wf_terms"
using assms move_preserves_wf_terms by (induct) auto

lemma fterm_in_wf_terms:
  "fterm s \<in> wf_terms"
by (induct s) (auto simp: wf_terms.intros)

lemma fqterm_in_wf_terms:
  "fqterm f qs \<in> wf_terms"
by (simp add: wf_terms.intros)

lemma non_empty_Inr_not_in_wf_terms:
  "Fun (Inr q) (t # ts) \<notin> wf_terms"
by (auto, cases rule: wf_terms.cases) (auto)

lemma wf_term_no_moves_to_non_empty_Inr:
  assumes "(s, Fun (Inr q) (t # ts)) \<in> (move A)\<^sup>*" (is "(_, ?t) \<in> ?A")
    and "s \<in> wf_terms"
  shows "False"
using assms
proof (induct s ?t rule: rtrancl.induct [case_names base step])
  case (base)
  then show ?case using non_empty_Inr_not_in_wf_terms by fast
next
  case (step s u)
  have "u \<in> wf_terms" using moves_preserve_wf_terms [OF step(1) step(4)] .
  moreover have "?t \<notin> wf_terms" using non_empty_Inr_not_in_wf_terms .
  ultimately show ?case using step(3) move_preserves_wf_terms by blast
qed

lemma ftrem_no_moves_to_non_empty_Inr:
  assumes "(fterm s, Fun (Inr g) (t # ts)) \<in> (move A)\<^sup>*" (is "(?s, ?t) \<in> ?A")
  shows "False"
proof -
  have "?s \<in> wf_terms" by (simp add: fterm_in_wf_terms)
  from moves_preserve_wf_terms [OF assms this] have "?t \<in> wf_terms" .
  then show ?thesis using non_empty_Inr_not_in_wf_terms by fast
qed

inductive_set move_comp :: "('q, 'f) ta \<Rightarrow> _" for A
where
  Hole [simp]: "(\<box>, \<box>) \<in> move_comp A"
| More: "f = g \<Longrightarrow> (C, D) \<in> move_comp A \<Longrightarrow> length ss\<^sub>1 = length ts\<^sub>1 \<Longrightarrow> length ss\<^sub>2 = length ts\<^sub>2 \<Longrightarrow>
    \<forall>i < length (ts\<^sub>1). (ss\<^sub>1 ! i, ts\<^sub>1 ! i) \<in> (move A)\<^sup>* \<Longrightarrow>
    \<forall>i < length (ts\<^sub>2). (ss\<^sub>2 ! i, ts\<^sub>2 ! i) \<in> (move A)\<^sup>* \<Longrightarrow>
    (More f ss\<^sub>1 C ss\<^sub>2, More g ts\<^sub>1 D ts\<^sub>2) \<in> move_comp A"

lemma move_comp_imp_moves:
  assumes "(C, D) \<in> move_comp A"
  shows "(C\<langle>t\<rangle>, D\<langle>t\<rangle>) \<in> (move A)\<^sup>*"
using assms
apply (induct)
apply (auto simp: nth_append intro!: args_steps_imp_steps [OF ctxt_closed_move])
apply (case_tac "length ts\<^sub>1 = i"; simp)
done

lemma distr_moves'':
  assumes "(s, C\<langle>Fun (Inl f) ts\<rangle>) \<in> (move A)\<^sup>*"
    and "s \<in> wf_terms"
  shows "\<exists>D ss. s = D\<langle>Fun (Inl f) ss\<rangle> \<and> length ts = length ss \<and>
         (\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*) \<and>
         (D\<langle>Fun (Inl f) ts\<rangle>, C\<langle>Fun (Inl f) ts\<rangle>) \<in> (move A)\<^sup>* \<and>
         (D, C) \<in> move_comp A"
using assms
proof (induct C arbitrary: s)
  case (Hole)
  then have *: "(s, Fun (Inl f) ts) \<in> (move A)\<^sup>*" by simp
  from distr_moves [OF this] obtain ss where "s = Fun (Inl f) ss" and
    "length ts = length ss" and
    "\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*"
    by blast
  then show ?case by (intro exI [where x = \<box>]; simp)
next
  case (More g us E vs)
  let ?t = "Fun (Inl f) ts"
  let ?ts = "us @ E\<langle>?t\<rangle> # vs"
  let ?C = "More g us E vs"
  have *: "?C\<langle>?t\<rangle> = Fun g ?ts" by simp
  show ?case
  proof (cases g)
    case (Inl g')
    have "(s, Fun (Inl g') ?ts) \<in> (move A)\<^sup>*" using More(2) [unfolded *, unfolded Inl] .
    from distr_moves [OF this] obtain ss' where s: "s = Fun (Inl g') ss'" and
       len: "length ?ts = length ss'" and
       ss': "\<forall>i < length ss'. (ss' ! i, ?ts ! i) \<in> (move A)\<^sup>*"
       by blast
    moreover def j \<equiv> "length us"
    ultimately have 1: "j < length ss'" and 2: "(ss' ! j, E\<langle>?t\<rangle>) \<in> (move A)\<^sup>*"
      and 3: "?ts ! j = E\<langle>?t\<rangle>" apply auto
      by (metis less_add_Suc1 nth_append_length)
    from wf_terms_Fun_Inl_args [OF More(3) [unfolded s]] 
      have "ss' ! j \<in> wf_terms" using 1 by auto
    from More(1) [OF 2 this] obtain D' ss where j: "ss' ! j = D'\<langle>Fun (Inl f) ss\<rangle>" and
      len': "length ts = length ss" and
      ss: "\<forall>i < length ss. (ss ! i, ts ! i) \<in> (move A)\<^sup>*" and
      4: "(D'\<langle>?t\<rangle>, E\<langle>?t\<rangle>) \<in> (move A)\<^sup>*" and
      mc: "(D', E) \<in> move_comp A" by blast
    let ?ss'' = "take j ss' @ D'\<langle>Fun (Inl f) ts\<rangle> # drop (Suc j) ss'"
    have "s = Fun (Inl g') (take j ss' @ D'\<langle>Fun (Inl f) ss\<rangle> # drop (Suc j) ss')"
      using s j 1 id_take_nth_drop by fastforce
    then have 5: "ss' = take j ss' @ D'\<langle>Fun (Inl f) ss\<rangle> # drop (Suc j) ss'"
      (is "_ = ?ss'") using s by simp
    then have 6: "s = (More (Inl g') (take j ss') D' (drop (Suc j) ss'))\<langle>Fun (Inl f) ss\<rangle>"
      (is "_ = ?D\<langle>_\<rangle>") using s by fastforce
    have 8: "\<forall>i < length us. (ss' ! i, us ! i) \<in> (move A)\<^sup>*" using 1 ss'
      by (metis (no_types, lifting) Suc_less_eq append_Cons_nth_left j_def less_SucI less_trans_Suc)
    have 9: "\<forall>i < length vs. (ss' ! Suc (length us + i), vs ! i) \<in> (move A)\<^sup>*"
      using 1 len ss'
      by (auto)
         (metis (no_types) add.commute add_Suc_right nat_add_left_cancel_less nth_Cons_Suc
          nth_append_length_plus)
    from ss' have "\<forall>i < length ?ss'. (?ss' ! i, ?ts ! i) \<in> (move A)\<^sup>*" using 5 by presburger
    then have 7: "\<forall>i < length ?ss''. (?ss' ! i, ?ts ! i) \<in> (move A)\<^sup>*" by auto
    { fix i
      assume a: "i < length ?ss''"
      then have "(?ss'' ! i, ?ts ! i) \<in> (move A)\<^sup>*"
      proof (cases "i = j")
        case (False)
        then show ?thesis using a 1 5 7
          by (metis (no_types, lifting) length_Cons length_append less_imp_le_nat
              nth_append_take_drop_is_nth_conv)
      next
        case (True)
        then have *: "?ss'' ! i = D'\<langle>?t\<rangle>" by (simp add: 1 less_imp_le_nat nth_append_take)
        from True have "?ts ! i = E\<langle>?t\<rangle>" using 3 by simp
        then show ?thesis unfolding * using 4 by simp
      qed
    }
    then have "\<forall>i < length ?ss''. (?ss'' ! i, ?ts ! i) \<in> (move A)\<^sup>*" by blast
    from args_steps_imp_steps [OF ctxt_closed_move _ this, of "Inl g'"]
      have "(Fun (Inl g') ?ss'', Fun (Inl g') ?ts) \<in> (move A)\<^sup>*" using Inl More.prems(1) s
      by (metis 5 len length_Cons length_append)
    then have "(?D\<langle>?t\<rangle>, ?C\<langle>?t\<rangle>) \<in> (move A)\<^sup>*" unfolding Inl by fastforce
    moreover have "(?D, ?C) \<in> move_comp A" using mc len 8 9
      by (intro move_comp.More) (auto simp: Inl j_def)
    ultimately show ?thesis using 6 len' ss unfolding Inl by blast
  next
    case (Inr q)
    have "(s, Fun (Inr q) ?ts) \<in> (move A)\<^sup>*" using More(2) unfolding * using Inr by blast
    then obtain w ws where "(s, Fun (Inr q) (w # ws)) \<in> (move A)\<^sup>*"
      by (metis Nil_is_append_conv list.exhaust) 
    from wf_term_no_moves_to_non_empty_Inr [OF this More(3)] have "False" .
    then show ?thesis ..
  qed
qed

lemma State_move_imp_State:
  assumes "(State q, s) \<in> move A"
  shows "\<exists> q'. s = State q'"
proof -
  { fix l
    assume "(l, s) \<in> move A" and "l = State q"
    then have ?thesis
    by (induct l s rule: move.induct) (auto elim: ctxt_apply_term.elims)
  }
  with assms show ?thesis by simp
qed

lemma State_move_seq_imp_State:
  assumes "(State q, s) \<in> (move A)\<^sup>*"
  shows "\<exists> q'. s = State q'"
proof -
  { fix l
    assume "(l, s) \<in> (move A)\<^sup>*" and "l = State q"
    then have ?thesis
    by (induct) (auto simp: State_move_imp_State)
  }
  with assms show ?thesis by blast
qed

lemma State_move_imp_ta_eps:
  assumes "(State q, State q') \<in> move A"
  shows "(q, q') \<in> (ta_eps A)\<^sup>*"
proof -
  { fix l r :: "('a + 'b, 'd) term"
    assume "(l, r) \<in> move A" and "l = State q" and "r = State q'"
    then have ?thesis
    proof (induct l r rule: move.induct)
      case (trans f qs q1 C)
      then show ?case using ctxt_apply_term.elims by blast
    next
      case (eps q1 q2 C)
      then have "C = \<box>" using ctxt_apply_term.elims by auto
      with eps show ?case by simp
    qed
  }
  with assms show ?thesis by auto
qed

lemma State_move_seq_imp_ta_eps:
  assumes "(State q, State q') \<in> (move A)\<^sup>*" (is "(_, _) \<in> ?move\<^sup>*")
  shows "(q, q') \<in> (ta_eps A)\<^sup>*"
proof -
  { fix l r
    assume "(l, r) \<in> ?move\<^sup>*" "l = State q" "r = State q'"
    then have ?thesis
    proof (induct l r arbitrary: q q' rule: rtrancl.induct[case_names refl step])
      case (step l m r)
      have "(State q, m) \<in> ?move\<^sup>*" using step(4,1) by simp
      from State_move_seq_imp_State[OF this] obtain q1 where [simp]: "m = State q1" by auto
      show ?case using step by (force dest: State_move_imp_ta_eps)
    qed (simp)
  }
  with assms show ?thesis by blast
qed

lemma ta_eps_reachable_states:
  assumes "p \<in> reachable_states A t" and "(p, q) \<in> (ta_eps A)\<^sup>*"
  shows "q \<in> reachable_states A t"
using assms
by (induct t rule: reachable_states.induct) (auto, meson rtrancl_trans)

lemma move_reachable_states:
  "(t, State q) \<in> (move A)\<^sup>* \<longleftrightarrow> q \<in> reachable_states A t" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  then show ?R
  proof (induct t arbitrary: q)
    case (Var x)
    with rtrancl_move_ground_ground show ?case by fastforce
  next
    case (Fun f ts)
    note IH = this(1)
    {
      fix l r q
      assume "(l, r) \<in> (move A)\<^sup>*" (is "(_, _) \<in> ?move\<^sup>*") "l = Fun f ts" "r = State q"
      then have "q \<in> reachable_states A (Fun f ts)"
      proof (induct l r arbitrary: q rule: rtrancl.induct[case_names refl step])
        case (step l m r)
        show ?case using step(3,1,2) unfolding step(4,5)
        proof (induct rule: move.cases)
          case (trans f' qs q2 C)
          with ctxt_apply_term.elims have "C = \<box>" by auto
          show ?case
          proof (cases m)
            case (Fun g ps)
            then show ?thesis
            proof (cases "\<exists> p. m = State p")
              case (True)
              from True obtain p where True: "m = State p" ..
              from step(2)[OF step(4) True] have 1: "p \<in> reachable_states A (Fun f ts)" .
              from step(3) True step(5) have "(State p, State q) \<in> ?move" by auto
              from State_move_imp_ta_eps[OF this] have 2: "(p, q) \<in> (ta_eps A)\<^sup>*" .
              from ta_eps_reachable_states[OF 1 2] show ?thesis .
            next
              case (False)
              then show ?thesis
              proof (cases "\<exists> h x xs. m = Fun (Inr h) (x # xs)")
                case (False)
                then obtain a us where "m = Fun a us" by (simp add: \<open>C = \<box>\<close> trans)
                from this trans(1) \<open>C = \<box>\<close> have "Fun a us = fqterm f' qs" by simp
                then have "a = Inl f'" and "us = map State qs" by auto
                with \<open>m = Fun a us\<close> have "m = Fun (Inl f') us" by simp
                with step(1) have "(l, Fun (Inl f') us) \<in> ?move\<^sup>*" by simp
                from distr_moves[OF this] obtain ss where "l = Fun (Inl f') ss"
                  and "length us = length ss"
                  and  *: "\<forall>i < length ss. (ss ! i, us ! i) \<in> ?move\<^sup>*" by auto
                from \<open>l = Fun (Inl f') ss\<close> step(4) have "ss = ts" and "f = Inl f'" by auto
                have "a = f" using \<open>a = Inl f'\<close> \<open>f = Inl f'\<close> by simp
                then have "(Fun f ss, Fun f (map State qs)) \<in> ?move\<^sup>*"
                  using \<open>l = Fun (Inl f') ss\<close> \<open>m = Fun a us\<close> \<open>us = map State qs\<close> step
                  by simp
                from * have **: "\<forall>i < length ts. (ts ! i, (map State qs) ! i) \<in> ?move\<^sup>*"
                  by (simp add: \<open>ss = ts\<close> \<open>us = map State qs\<close>)
                { fix i
                  assume len: "i < length ts"
                  then have ti: "ts ! i \<in> set ts" by simp
                  from len ** have qi: "(ts ! i, (map State qs) ! i) \<in> ?move\<^sup>*" by simp
                  with IH[OF ti]
                  have "qs ! i \<in> reachable_states A (ts ! i)"
                    using \<open>length us = length ss\<close> \<open>ss = ts\<close> \<open>us = map State qs\<close> len by auto
                }
                then have "\<forall> i < length ts. qs ! i \<in> reachable_states A (ts ! i)" by auto
                then show ?thesis
                  using \<open>f = Inl f'\<close> \<open>C = \<box>\<close> \<open>length us = length ss\<close> \<open>ss = ts\<close> \<open>us = map State qs\<close>
                  trans by auto
              qed (auto simp: trans \<open>C = \<box>\<close>)
            qed
          qed (auto simp: trans \<open>C = \<box>\<close>)
        next
          case (eps q1 q2 C)
          then show ?case by (cases C, cases f; cases ts) (auto, (metis rtrancl_trans)+)
        qed
      qed auto
    }
    then show ?case using Fun(2) by blast
  qed
next
  assume ?R
  then show ?L
  proof (induct t arbitrary: q rule: reachable_states.induct)
    case (1 A q')
    then show ?case
    using move.intros(2)[of q' q A "\<box>"] by auto
  next
    case (2 A f ts q')
    obtain q qs where trans: "TA_rule f qs q \<in> ta_rules A"
      and eps: "(q, q') \<in> (ta_eps A)\<^sup>*"
      and len: "length qs = length ts"
      and rec: "\<forall> i < length ts. qs ! i \<in> reachable_states A (ts ! i)"
        using 2(2) by auto
    have len': "length ts = length (map State qs)" using len by auto
    have rec': "\<forall>i < length ts. (ts ! i, map State qs ! i) \<in> (move A)\<^sup>*"
      by (simp add: rec "2.hyps" len)
    have 1: "(Fun (Inl f) ts, fqterm f qs) \<in> (move A)\<^sup>*"
      using args_steps_imp_steps[OF ctxt_closed_move len' rec', of "Inl f"] by simp
    have 2: "(fqterm f qs, State q) \<in> move A" using move.intros(1)[OF trans, of "\<box>"]  by simp
    have "(State q, State q') \<in> move A" using move.intros(2)[OF eps, of "\<box>"] by simp
    then have 3: "(State q, State q') \<in> (move A)\<^sup>*" by auto
    show ?case using rtrancl_trans[OF rtrancl.rtrancl_into_rtrancl[OF 1 2] 3] .
  qed (auto)
qed

lemma move_rtrancl_map_states_iff:
  assumes "inj g"
  shows "(fterm t, State q) \<in> (move A)\<^sup>* \<longleftrightarrow> (fterm t, State (g q)) \<in> (move (map_states g A))\<^sup>*"
using assms by (simp add: move_reachable_states inj_image_mem_iff)

lemma map_states_term_State [simp]:
  shows "map_states_term g \<circ> State = State \<circ> g"
by (rule ext) (simp)

lemma inj_map_states_term [intro, simp]:
  assumes "inj g"
  shows "inj (map_states_term g)"
proof (rule injI)
  fix s t :: "('e + 'a, 'f) term"
  assume "map_states_term g s = map_states_term g t"
  then show "s = t"
  using assms
  apply (induct g s arbitrary: t rule: map_states_term.induct)
  apply (auto elim!: map_states_term.elims simp: inj_on_def) by (metis list.inj_map_strong)+
qed

lemma map_append_Cons_exists:
  assumes "xs @ y # ys = map f zs"
    and "inj f"
  shows "\<exists>xs' y' ys'. xs = map f xs' \<and> y = f y' \<and> ys = map f ys' \<and> zs = xs' @ y' # ys'"
proof -
  { fix xs' y' ys'
    assume 1: "xs' = map (inv f) xs" and 2: "ys' = map (inv f) ys" and 3: "y' = inv f y"
    moreover
    then have 4: "length xs' = length xs" and 5: "length ys' = length ys" by auto
    ultimately
    have "xs @ y # ys = (map f xs') @ (f y') # (map f ys')"
      using assms
      by (auto) (metis (no_types, lifting) list.simps(9) map_append map_map o_inv_o_cancel)
    then have "xs = map f xs' \<and> y = f y' \<and> ys = map f ys' \<and> zs = xs' @ y' # ys'"
      using assms 1 2 3 4 5
      by (metis append_eq_conv_conj inj_map_eq_map list.sel(1) list.sel(3) list.simps(9)
          map_append take_map)
  }
  from this[of "map (inv f) xs" "map (inv f) ys" "inv f y"] show ?thesis by blast
qed

lemma move_map_states_iff:
  assumes "inj g"
  shows "(map_states_term g s, map_states_term g t) \<in> move (map_states g A) \<longleftrightarrow> (s, t) \<in> move A"
  (is "?L \<longleftrightarrow> ?R")
proof
  assume "?R"
  then show "?L"
  unfolding move_eq_move'
  proof (induct)
    case (trans f qs q)
    then have "TA_rule f (map g qs) (g q) \<in> ta_rules (map_states g A)" using assms by simp
    from move'.trans[OF this] show ?case by simp
  next
    case (eps q q')
    then have "(g q, g q') \<in> (ta_eps (map_states g A))\<^sup>*" using assms by simp
    from move'.eps[OF this] show ?case by simp
  next
    case (Fun s t f ss ts)
    from move'.Fun[OF this(2), of _ "map (map_states_term g) ss" "map (map_states_term g) ts"]
    show ?case by (cases f) (simp)+
  qed
next
  assume "?L"
  then show "?R"
  unfolding move_eq_move'
  proof (induct "map_states_term g s" "map_states_term g t" arbitrary: s t)
    case (trans f qs q)
    then obtain qs' q' where "qs = map g qs'" and "q = g q'" and
    rule: "TA_rule f qs' q' \<in> ta_rules A" by auto
    with trans assms have "s = Fun (Inl f) (map State qs')" and "t = State q'"
    apply (auto elim!: map_states_term.elims)
    apply (subst (asm) map_states_term_State[symmetric])
    unfolding map_map[symmetric]
    apply (drule map_injective)
    apply (simp)+
    apply (simp add: inj_on_def)
    done
    with rule show ?case by (auto intro: move'.intros)
  next
    case (eps q q')
    then show ?case
    using assms by (auto elim!: map_states_term.elims intro: move'.intros)
  next
    note * = inj_map_states_term [OF assms]
    case (Fun u v f ss ts)
    then show ?case
    using assms
      by (auto dest!: inj_onD elim!: map_states_term.elims)
         (auto dest!: map_append_Cons_exists [OF _ *] map_injective [OF _ *] intro!: move'.Fun)
  qed
qed

lemma disj_move_union_iff:
  assumes "states A \<inter> states B = {}"
  shows "move (ta_union A B) = (move A) \<union> (move B)"
  (is "?L = ?R1 \<union> ?R2")
proof (intro equalityI subrelI)
  fix s t
  assume "(s, t) \<in> ?L"
  then show "(s, t) \<in> ?R1 \<union> ?R2"
  proof (induct)
    case (eps q q' C)
    with assms have "(q, q') \<in> (ta_eps A)\<^sup>* \<union> (ta_eps B)\<^sup>*"
      by (auto simp: rtrancl_eq_or_trancl) (drule disj_states_imp_disj_ta_eps, blast+)
    then show ?case by (auto simp: move.eps)
  qed (auto simp: move.trans)
next
  fix s t
  assume "(s, t) \<in> ?R1 \<union> ?R2"
  then have "(s, t) \<in> ?R1 \<or> (s, t) \<in> ?R2" by simp
  then show "(s, t) \<in> ?L"
  by (auto simp: move.trans elim!: move.cases) (simp add: in_rtrancl_UnI move.eps)+
qed

lemma empty_ta_eps_move_union_iff:
  assumes "ta_eps A = {}"
  shows "move (ta_union A B) = (move A) \<union> (move B)"
  (is "?L = ?R1 \<union> ?R2")
proof (intro equalityI subrelI)
  fix s t
  assume "(s, t) \<in> ?L"
  then show "(s, t) \<in> ?R1 \<union> ?R2"
  proof (induct)
    case (eps q q' C)
    with assms have "(q, q') \<in> (ta_eps B)\<^sup>*" by (auto)
    then show ?case by (auto simp: move.eps)
  qed (auto simp: move.trans)
next
  fix s t
  assume "(s, t) \<in> ?R1 \<union> ?R2"
  then have "(s, t) \<in> ?R1 \<or> (s, t) \<in> ?R2" by simp
  then show "(s, t) \<in> ?L"
  by (auto simp: move.trans elim!: move.cases) (simp add: in_rtrancl_UnI move.eps)+
qed

lemma empty_ta_eps_move_union_iff':
  assumes "ta_eps B = {}"
  shows "move (ta_union A B) = (move A) \<union> (move B)"
using empty_ta_eps_move_union_iff [OF assms] by auto

lemma funs_term_ctxt_apply [simp]:
  "funs_term (C\<langle>t\<rangle>) = funs_ctxt C \<union> funs_term t"
proof (induct t)
  case (Var x) show ?case by (induct C) auto
next
  case (Fun f ts) show ?case by (induct C arbitrary: f ts) auto
qed

lemma move_preserves_signature:
  assumes "funs_term s \<subseteq> Inl ` F \<union> Inr ` states A" (is "_ \<subseteq> ?S")
    and "(s, t) \<in> move A"
  shows "funs_term t \<subseteq> ?S"
using assms(2)
proof (cases rule: move.cases)
  case (trans f qs q C)
  show ?thesis using trans(1,2) assms(1) r_states_in_states[OF trans(3)] by simp
next
  case (eps q q' C)
  show ?thesis
  proof (cases "q = q'")
    case (True)
    then show ?thesis using assms(1) and eps(1,2) by fast
  next
    case (False)
    show ?thesis using eps and assms(1) and trancl_ta_epsD[of q q' A]
      and rtrancl_eq_or_trancl[of q q' "ta_eps A"] by auto
  qed
qed

lemma move_rtrancl_preserves_signature:
  assumes "funs_term s \<subseteq> Inl ` F \<union> Inr ` states A" (is "_ \<subseteq> ?S")
    and "(s, t) \<in> (move A)\<^sup>*"
  shows "funs_term t \<subseteq> ?S"
using assms(2,1) by (induct) (auto simp: move_preserves_signature)

lemma disj_move_imp_ta_union_move:
  assumes "states A \<inter> states B = {}"
    and "(s, t) \<in> move A"
  shows "(s, t) \<in> move (ta_union A B)"
using assms disj_move_union_iff by blast

lemma disj_move_rtrancl_imp_ta_union_move_rtrancl:
  assumes "states A \<inter> states B = {}"
    and "(s, t) \<in> (move A)\<^sup>*"
  shows "(s, t) \<in> (move (ta_union A B))\<^sup>*"
using assms(2) disj_move_union_iff assms(1)
by (induct, auto)
   (simp add: disj_move_union_iff in_rtrancl_UnI rtrancl.rtrancl_into_rtrancl)

lemma disj_move_trancl_imp_ta_union_move_trancl:
  assumes "states A \<inter> states B = {}"
    and "(s, t) \<in> (move A)\<^sup>+"
  shows "(s, t) \<in> (move (ta_union A B))\<^sup>+"
using assms by (meson disj_move_imp_ta_union_move subrelI trancl_mono)

lemma disj_reachable_states_ta_union_reachable_states:
  assumes "q \<in> reachable_states A t"
    and "states A \<inter> states B = {}"
  shows "q \<in> reachable_states (ta_union A B) t"
proof -
  have "(t, State q) \<in> (move A)\<^sup>*" using move_reachable_states [THEN iffD2, OF assms(1)] .
  from disj_move_rtrancl_imp_ta_union_move_rtrancl [OF assms(2) this]
    have "(t, State q) \<in> (move (ta_union A B))\<^sup>*" .
  from move_reachable_states [THEN iffD1, OF this]
    show "q \<in> reachable_states (ta_union A B) t" .
qed

lemma langE [elim]:
  assumes "t \<in> lang A"
  obtains q\<^sub>f where "(fterm t, State q\<^sub>f) \<in> (move A)\<^sup>*" and "q\<^sub>f \<in> ta_final A"
using assms move_reachable_states[symmetric] unfolding lang_def by fast

lemma fterm_subst_distrib [simp]:
  "fterm (t \<cdot> \<sigma>) = fterm t \<cdot> (fterm \<circ> \<sigma>)"
by (induct t) auto

lemma moves_state_subst:
  fixes \<sigma> :: "('f + 'q, 'v) subst"
  assumes "(s \<cdot> \<sigma>, State q) \<in> (move A)\<^sup>*" (is "_ \<in> ?move") and "q \<in> states A" and "linear_term s"
  shows "\<exists>\<theta>. (\<forall>x \<in> vars_term s. (\<sigma> x, (State \<circ> \<theta>) x) \<in> (move A)\<^sup>*) \<and>
             (s \<cdot> (State \<circ> \<theta>), State q :: ('f + 'q, 'v) term) \<in> (move A)\<^sup>* \<and> 
             range \<theta> \<subseteq> states A"
using assms
proof (induct s arbitrary: q)
  case (Var x)
  { fix \<theta> :: "'v \<Rightarrow> 'q"
    assume \<theta>: "\<theta> = (\<lambda>y. q)"
    let ?\<theta> = "\<lambda>x. State (\<theta> x)"
    have "(Var x \<cdot> \<sigma>, Var x \<cdot> ?\<theta>) \<in> ?move" using Var \<theta> by simp
    moreover have "(Var x \<cdot> ?\<theta>, State q) \<in> ?move" using \<theta> by simp
    moreover have "range \<theta> \<subseteq> states A" using Var \<theta> by blast 
    ultimately
    have "(Var x \<cdot> \<sigma>, Var x \<cdot> ?\<theta>) \<in> ?move \<and> (Var x \<cdot> ?\<theta>, State q) \<in> ?move \<and>
          range \<theta> \<subseteq> states A" by simp
  }
  then show ?case by force
next
  case (Fun f ss)
  let ?ss = "map (\<lambda>s. s \<cdot> \<sigma>) ss"
  show ?case
  proof (cases "\<exists>f'. f = Inl f'")
    case (True) (* f is a function symbol *)
    then obtain f' where f': "f = Inl f'" by auto
    have "q \<in> reachable_states A (Fun (Inl f') ?ss)"
      using move_reachable_states [THEN iffD1, OF Fun(2)] unfolding f' by simp
    then obtain qs q' where
      trans: "TA_rule f' qs q' \<in> ta_rules A" and
      eps: "(q',q) \<in> (ta_eps A)\<^sup>*" and
      len: "length qs = length ?ss" and
      qs: "\<forall> i < length ?ss. qs ! i \<in> reachable_states A (?ss ! i)"
      by auto
    have len': "length qs = length ss" using len by simp
    { fix i
      let ?s\<^sub>i = "ss ! i"
      let ?q\<^sub>i = "qs ! i"
      let ?subst = "\<lambda>\<theta>. State \<circ> \<theta>"
      assume l: "i < length ss"
      have "?s\<^sub>i \<in> set ss" using l by simp
      moreover have "(?s\<^sub>i \<cdot> \<sigma>, State ?q\<^sub>i) \<in> ?move" using move_reachable_states qs l by force
      moreover have "?q\<^sub>i \<in> states A" using l len' using r_states_in_states[OF trans] by fastforce
      moreover have "linear_term ?s\<^sub>i" using Fun(4) \<open>?s\<^sub>i \<in> set ss\<close>  by simp
      ultimately have "\<exists>\<theta>. (\<forall> x \<in> vars_term ?s\<^sub>i. (\<sigma> x, (?subst \<theta>) x) \<in> ?move) \<and>
        (?s\<^sub>i \<cdot> (?subst \<theta>), State ?q\<^sub>i) \<in> ?move \<and>
        range \<theta> \<subseteq> states A"
        by (rule Fun(1))
    }
    then obtain \<theta> where **: "\<forall>i < length ss. (\<forall>x \<in> vars_term (ss ! i).
                (\<sigma> x, (State \<circ> (\<theta> i)) x) \<in> ?move) \<and>
               (ss ! i \<cdot> (State \<circ> (\<theta> i)), State (qs ! i)) \<in> ?move \<and>
               range (\<theta> i) \<subseteq> states A" by metis
    have "is_partition (map vars_term ss)" using Fun(4) by simp
    from subst_merge [OF this, of \<theta>]
      obtain \<theta>' where \<theta>': "\<forall>i < length ss. \<forall>x \<in> vars_term (ss ! i). \<theta>' x = \<theta> i x" by blast
    moreover def \<theta>'' \<equiv> "(\<lambda>x. (if x \<in> vars_term (Fun f ss) then \<theta>' x else q))"
    ultimately have *: "\<forall>i < length ss. \<forall>x \<in> vars_term (ss ! i). State (\<theta>'' x) = State (\<theta> i x)"
      by (auto)
    { fix i :: nat
      assume l: "i < length ss"
      from * [THEN spec, THEN mp, of i, OF this]
      have "\<And>x. x \<in> vars_term (ss ! i) \<Longrightarrow> (State (\<theta>'' x) :: ('f + 'q, 'v) term) = State (\<theta> i x)" by simp
      from term_subst_eq [OF this, of "ss ! i" id, simplified]
        have "(\<forall> x \<in> vars_term (ss ! i). (\<sigma> x, (State \<circ> \<theta>'') x) \<in> ?move) \<and>
               (ss ! i \<cdot> (State \<circ> \<theta>''), State (qs ! i)) \<in> ?move"
        using ** * l by (auto simp: o_def)
    }
    then have "\<forall>i < length ss. (\<forall>x \<in> vars_term (ss ! i). (\<sigma> x, (State \<circ> \<theta>'') x) \<in> ?move) \<and>
               (ss ! i \<cdot> (State \<circ> \<theta>''), State (qs ! i)) \<in> ?move" by (auto simp: o_def)
    then have "(\<forall>x \<in> vars_term (Fun f ss). (\<sigma> x, (State \<circ> \<theta>'') x) \<in> ?move)" and
              "(Fun f ss \<cdot> (State \<circ> \<theta>''), Fun f (map State qs)) \<in> ?move"
      using len by (auto intro: args_steps_imp_steps [OF ctxt_closed_move] simp: in_set_conv_nth)
    moreover have "range \<theta>'' \<subseteq> states A" using ** Fun(3) \<theta>'
      by (auto simp: \<theta>''_def in_set_conv_nth) (blast)
    moreover have "(Fun f (map State qs), State q) \<in> ?move"
      using move.trans [OF trans, of \<box>] and move.eps [OF eps, of \<box>]
      by (auto simp: f' dest: r_into_rtrancl rtrancl_trans)
    ultimately show ?thesis by (force simp: o_def)
  next
    case (False) (* f is a state symbol *)
    then obtain f' where f': "f = Inr f'" using sum.exhaust_sel by blast
    have *: "q \<in> reachable_states A (Fun (Inr f') ss \<cdot> \<sigma>)"
      using move_reachable_states [THEN iffD1, OF Fun(2)] unfolding f' .
    show ?thesis
    proof (cases "ss = Nil")
      case (False)
      then obtain x xs where "ss = x # xs" by (metis list.exhaust)
      then have "reachable_states A (Fun (Inr f') ss \<cdot> \<sigma>) = {}"
      using * reachable_states.simps(4) by auto
      then show ?thesis using * by simp
    next
      def \<theta> \<equiv> "(\<lambda>x. q) :: 'v \<Rightarrow> 'q"
      case (True)
      then show ?thesis  using Fun(2,3) by (auto simp: \<theta>_def f')
    qed
  qed
qed

lemma moves_State:
  assumes "(C\<langle>t\<rangle>, State q) \<in> (move A)\<^sup>*" (is "(_, ?q) \<in> ?move")
  shows "\<exists>q'. (t, (State q') :: ('f + 'q, 'v) term) \<in> (move A)\<^sup>* \<and>
              (C\<langle>State q'\<rangle>, State q) \<in> (move A)\<^sup>*"
using assms
proof (induct C arbitrary: q)
  case (More f ss D ts)
  let ?ss = "ss @ D\<langle>t\<rangle> # ts"
  show ?case
  proof (cases "\<exists> f'. f = Inl f'")
    case (True) (* f is a function symbol *)
    then obtain f' where f': "f = Inl f'" by auto
    have "q \<in> reachable_states A (More f ss D ts)\<langle>t\<rangle>"
      using move_reachable_states [THEN iffD1, OF More(2)] .
    then obtain qs q\<^sub>1 where
      trans: "TA_rule f' qs q\<^sub>1 \<in> ta_rules A" and
      eps: "(q\<^sub>1, q) \<in> (ta_eps A)\<^sup>*" and
      len: "length qs = length ?ss" and
      qs: "\<forall> i < length ?ss. qs ! i \<in> reachable_states A (?ss ! i)"
      unfolding f' by auto
    have qs': "\<forall> i < length ?ss. (?ss ! i, State (qs ! i)) \<in> ?move"
      using qs move_reachable_states by metis
    then obtain k where 1: "(D\<langle>t\<rangle>, State (qs ! k)) \<in> ?move" and k: "k = length ss"
      by (metis (no_types, lifting) add_Suc_right len length_Cons length_append less_add_Suc1
          nth_append_length)
    let ?q\<^sub>x = "qs ! k"
    obtain q'' where *: "(t, State q'') \<in> ?move" and **: "(D\<langle>State q''\<rangle>, State ?q\<^sub>x) \<in> ?move"
      using 1 More(1) by blast
    have 2: "(Fun f (map State qs), State q\<^sub>1) \<in> ?move" using move.trans [OF trans, of \<box>]
      unfolding f' by auto
    have 3: "(State q\<^sub>1, State q) \<in> ?move" using move.eps [OF eps, of \<box>] by auto
    let ?ss' = "ss @ D\<langle>State q''\<rangle> # ts"
    have len': "length ?ss' = length (map State qs)" using len by simp
    have "(Fun f ?ss', Fun f (map State qs)) \<in> ?move"
    proof (intro args_steps_imp_steps [OF ctxt_closed_move len'] allI impI)
      fix i
      assume l: "i < length ?ss'"
      have "(?ss' ! i, State (qs ! i)) \<in> ?move"
      proof (cases "i = k")
        case (True)
        then show ?thesis using k len' ** by simp
      next
        case (False)
        show ?thesis
        proof (cases "i < k")
          case (True)
          then have "i < length ss" using k by simp
          then show ?thesis
            using k True qs' [THEN spec, of i] by (simp add: nth_append)
        next
          note False' = False
          moreover case (False)
          ultimately show ?thesis
            using k qs' [THEN spec, of i]
            by (metis append_Cons_nth_not_middle l len len' length_map)
        qed
      qed
      then show "(?ss' ! i, (map State qs ! i)) \<in> ?move" using l len' by simp
    qed
    then have "((More f ss D ts)\<langle>State q''\<rangle>, State q) \<in> ?move" using 2 3 by auto
    with * show ?thesis by blast
  next
    case (False) (* f is a state *)
    then obtain f' where f': "f = Inr f'" using sum.exhaust_sel by blast
    have "q \<in> reachable_states A (More (Inr f') ss D ts)\<langle>t\<rangle>"
      using move_reachable_states [THEN iffD1, OF More(2)] unfolding f' .
    moreover have "(More (Inr f') ss D ts)\<langle>t\<rangle> = Fun (Inr f') (ss @ D\<langle>t\<rangle> # ts)" by auto
    ultimately have 1: "q \<in> reachable_states A (Fun (Inr f') (ss @ D\<langle>t\<rangle> # ts))" by simp
    thm Nil_is_append_conv
    obtain x xs where "x # xs = ss @ D\<langle>t\<rangle> # ts" by (metis Nil_is_append_conv list.exhaust) 
    then have "reachable_states A (Fun (Inr f') (ss @ D\<langle>t\<rangle> # ts)) = {}"
      using reachable_states.simps(4) by metis
    with 1 have "False" by simp
    then show ?thesis ..
  qed
qed (auto)

lemma states_map_r_states_iff [simp]:
  assumes "inj g"
  shows "g q \<in> r_states (TA_rule f (map g qs) (g q)) \<longleftrightarrow> q \<in> r_states (TA_rule f qs q)"
unfolding r_states_def using ta_rules_map_states_iff [OF assms] by auto

lemma ta_rules_map_states_iff' [simp]:
  assumes "inj g"
  shows "map_r_states g r \<in> ta_rules (map_states g A) \<longleftrightarrow> r \<in> ta_rules A"
using assms by (cases r) (force simp: map_r_states_def inj_on_def map_states_def)

lemma r_states_map_states_iff [simp]:
  assumes "inj g"
  shows "g q \<in> \<Union>(r_states ` ta_rules (map_states g A)) \<longleftrightarrow>
         q \<in> \<Union>(r_states ` ta_rules A)"
using assms by (auto simp: r_states_def map_states_def map_r_states_def) (auto dest!: injD)

lemma r_states_map_states_ex:
  assumes "q' \<in> \<Union>(r_states ` ta_rules (map_states g A))"
  shows "\<exists>q. g q = q'"
proof -
  from assms have "q' \<in> \<Union>((r_states \<circ> map_r_states g) ` ta_rules A)"
    unfolding map_states_def by simp
  then show ?thesis unfolding r_states_def by (auto simp: map_r_states_def)
qed

lemma ta_eps_map_states_iff'' [simp]:
  assumes "inj g"
  shows "g q \<in> \<Union>((\<lambda>(q, q'). {q, q'}) ` ta_eps (map_states g A)) \<longleftrightarrow>
         q \<in> \<Union>((\<lambda>(q, q'). {q, q'}) ` ta_eps A)"
using assms by (auto simp: map_states_def inj_on_def)

lemma ta_eps_map_states_ex:
  assumes "q' \<in> \<Union>((\<lambda>(q, q'). {q, q'}) ` ta_eps (map_states g A))"
  shows "\<exists>q. g q = q'"
proof -
 from assms have "q' \<in> \<Union>((\<lambda>(q, q'). {g q, g q'}) ` ta_eps A)" unfolding map_states_def by auto
 then show ?thesis by blast
qed

lemma states_map_states_iff [simp]:
  assumes "inj g"
  shows "g q \<in> states (map_states g A) \<longleftrightarrow> q \<in> states A"
unfolding states_def using r_states_map_states_iff [OF assms, of q A] and
  ta_eps_map_states_iff'' [OF assms, of q A]
  inj_image_mem_iff [OF assms] by force

lemma states_map_states_ex:
  assumes "q' \<in> states (map_states g A)"
  shows "\<exists>q. q' = g q"
using assms unfolding states_def using r_states_map_states_ex [of q' g A]
  and ta_eps_map_states_ex [of q' g A] and ta_final_map_states_ex [of q' g A] by fast

lemma non_empty_ta_rules_map_states:
  assumes "ta_rules A \<noteq> {}" and "inj g"
  shows "ta_rules (map_states g A) \<noteq> {}"
using assms by (metis ex_in_conv ta_rules_map_states_iff')

lemma non_empty_ta_rules_ta_union:
  assumes "ta_rules A \<noteq> {}"
  shows "ta_rules (ta_union A B) \<noteq> {}"
using assms by (auto)

lemma states_ta_union_iff [simp]:
  "states (ta_union A B) = states A \<union> states B"
unfolding states_def by auto

lemma accessible_disj_states_ta_union:
  assumes "accessible TYPE('v) A q"
    and "states A \<inter> states B = {}"
  shows "accessible TYPE('v) (ta_union A B) q"
proof -
  obtain t :: "('b, 'v) term" where 1: "ground (fterm t)"
    and 2: "q \<in> reachable_states A (fterm t)"
    using accessibleE [OF assms(1)] by blast
  from disj_reachable_states_ta_union_reachable_states [OF 2 assms(2)] have
    "q \<in> reachable_states (ta_union A B) (fterm t)" .
  with 1 show ?thesis unfolding accessible_def by auto
qed

lemma accessible_ground_move_rtrancl:
  assumes "accessible TYPE('v) A q"
  shows "\<exists>t :: ('b, 'v) term. ground (fterm t) \<and> (fterm t, State q) \<in> (move A)\<^sup>*"
using assms unfolding accessible_def move_reachable_states by simp

lemma vars_move_imp_move:
  assumes "\<forall>x \<in> vars_term t. (\<sigma> x, \<theta> x) \<in> (move A)\<^sup>*"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<theta>) \<in> (move A)\<^sup>*"
using assms by (induct t) (auto simp add: args_steps_imp_steps ctxt_closed_move)

definition "sr R = {u | u l r. u \<lhd> l \<and> (l, r) \<in> R}"

lemma subterm_in_sr:
  assumes "t \<unrhd> s" and "t \<in> sr R"
  shows "s \<in> sr R"
(*using subterm.order.trans[OF assms(1)] assms(2) unfolding sr_def by blast*)
using subterm.le_less_trans[OF assms(1)] assms(2) unfolding sr_def by blast

lemma subterm_funas_term_subset:
  assumes "s \<lhd> t" and "funas_term t \<subseteq> F" shows "funas_term s \<subseteq> F"
using assms by (induct) auto

lemma Fun_in_sr_imp_F:
  assumes "Fun f ts \<in> sr R" and "funas_trs R \<subseteq> F"
  shows "(f, length ts) \<in> F"
using assms by (force simp: sr_def funas_defs dest: subterm_funas_term_subset)

lemma funas_term_sr:
  assumes "t \<in> sr R" and "funas_trs R \<subseteq> F"
  shows "funas_term t \<subseteq> F"
using assms by (force simp: sr_def funas_defs)

lemma subterm_linear:
  assumes "s \<rhd> t" and "linear_term s"
  shows "linear_term t"
using assms by (induct) auto

lemma sr_linear:
  assumes "t \<in> sr R" and "\<forall> (l, r) \<in> R. linear_term l"
  shows "linear_term t"
using assms
unfolding sr_def
by (auto split: prod.splits dest: subterm_linear)

lemma Fun_args_sr:
  assumes "Fun f ts \<in> sr R"
  shows "\<forall>i < length ts. (ts ! i) \<in> sr R"
using assms
by (auto simp: sr_def) (meson arg_subteq nth_mem supt_supteq_trans)

lemma finite_sr:
  fixes R :: "('f, 'v) trs"
  assumes "finite R"
  shows "finite (sr R)"
proof -
  let ?h = "\<lambda>t. {s. t \<rhd> s}"
  have *: "sr R = \<Union>(?h ` lhss R)" (is "_ = \<Union>(?B)") unfolding sr_def by force
  show "finite (sr R)" unfolding *
  proof (intro finite_Union)
     from assms have "finite (lhss R)" by auto
     from finite_imageI [OF this, of ?h] show "finite (?h ` lhss R)" by blast
  next
    fix M
    assume "M \<in> ?B"
    then show "finite M" by (auto simp: finite_subterms supt_supteq_not_supteq)
  qed
qed

lemma fterm_State_funas_imp_ta_fground:
  assumes "(fterm s, State q) \<in> (move A)\<^sup>*" and "funas_ta A \<subseteq> F"
  shows "fground F s"
proof -
  have "ground (State q)" by simp
  with assms(1) have "ground (fterm s)" using rtrancl_move_ground_ground by fastforce
  then have "ground s" by simp
  from move_reachable_states [THEN iffD1, OF assms(1)] have "q \<in> reachable_states A (fterm s)" .
  from reachable_states_funas_ta [OF this] have "funas_term s \<subseteq> funas_ta A" .
  then have "funas_term s \<subseteq> F" using assms(2) by blast
  then show "fground F s" using \<open>ground s\<close> by simp
qed

section \<open>Ground Instances\<close>

definition ground_instances :: "'f sig \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) terms"
where "ground_instances F t = { t \<cdot> \<sigma> | \<sigma>. fground F (t \<cdot> \<sigma>) }"
lemma ground_instances_Var [simp]:
  "ground_instances F (Var x) = {t. fground F t}"
by (auto simp: ground_instances_def)

lemma term_ground_instances_Fun:
  fixes ts :: "('f, 'v) term list"
  shows "ground_instances F (Fun f ts) \<subseteq> { Fun f ss | ss. (f, length ts) \<in> F \<and>
    length ts = length ss \<and> (\<forall>i < length ss. ss ! i \<in> ground_instances F (ts ! i))}"
proof -
  { fix ss :: "('f, 'v) term list"
    assume *: "is_partition (map vars_term ts)" and f: "(f, length ts) \<in> F"
      and "\<forall>x\<in>set ts. linear_term x"
      and l: "length ts = length ss"
      and "\<forall>i<length ss.
             \<exists>\<sigma>. ss ! i = ts ! i \<cdot> \<sigma> \<and>
                 (\<forall>x\<in>vars_term (ts ! i). ground (\<sigma> x)) \<and>
                 funas_term (ts ! i \<cdot> \<sigma>) \<subseteq> F" (is "\<forall>i<length ss. \<exists>\<sigma>. ?P i \<sigma>")
    then obtain \<tau> where **: "\<forall>i < length ss. ?P i (\<tau> i)" by metis
    from subst_merge [OF *, of \<tau>] obtain \<sigma>
      where ***: "\<forall>i<length ts. \<forall>x\<in>vars_term (ts ! i). \<sigma> x = \<tau> i x"
      by auto
    then have 4: "\<forall>i <length ts. ts ! i \<cdot> \<sigma> = ts ! i \<cdot> \<tau> i" by (auto intro: term_subst_eq)
    with ** have "ss = map (\<lambda>t. t \<cdot> \<sigma>) ts" using l by (auto intro!: nth_equalityI)
    moreover have "\<forall>x\<in>set ts. \<forall>x\<in>vars_term x. ground (\<sigma> x)"
      using ** *** l by (auto simp: in_set_conv_nth)
    moreover have "(f, length ss) \<in> F" using l f by auto
    moreover have "(\<Union>x\<in>set ts. funas_term (x \<cdot> \<sigma>)) \<subseteq> F"
      using ** 4 l by (auto simp: in_set_conv_nth) blast
    ultimately have "\<exists>\<sigma>. ss = map (\<lambda>t. t \<cdot> \<sigma>) ts \<and>
              (\<forall>x\<in>set ts. \<forall>x\<in>vars_term x. ground (\<sigma> x)) \<and>
              (f, length ss) \<in> F \<and>
              (\<Union>x\<in>set ts. funas_term (x \<cdot> \<sigma>)) \<subseteq> F" by blast
  }
  then show ?thesis using assms by (auto simp: ground_instances_def)
                                   (metis UN_subset_iff nth_mem)
qed

lemma linear_term_ground_instances_Fun:
  fixes ts :: "('f, 'v) term list"
  assumes "linear_term (Fun f ts)"
  shows "ground_instances F (Fun f ts) = { Fun f ss | ss. (f, length ts) \<in> F \<and>
    length ts = length ss \<and> (\<forall>i < length ss. ss ! i \<in> ground_instances F (ts ! i))}"
proof -
  { fix ss :: "('f, 'v) term list"
    assume *: "is_partition (map vars_term ts)" and f: "(f, length ts) \<in> F"
      and "\<forall>x\<in>set ts. linear_term x"
      and l: "length ts = length ss"
      and "\<forall>i<length ss.
             \<exists>\<sigma>. ss ! i = ts ! i \<cdot> \<sigma> \<and>
                 (\<forall>x\<in>vars_term (ts ! i). ground (\<sigma> x)) \<and>
                 funas_term (ts ! i \<cdot> \<sigma>) \<subseteq> F" (is "\<forall>i<length ss. \<exists>\<sigma>. ?P i \<sigma>")
    then obtain \<tau> where **: "\<forall>i < length ss. ?P i (\<tau> i)" by metis
    from subst_merge [OF *, of \<tau>] obtain \<sigma>
      where ***: "\<forall>i<length ts. \<forall>x\<in>vars_term (ts ! i). \<sigma> x = \<tau> i x"
      by auto
    then have 4: "\<forall>i <length ts. ts ! i \<cdot> \<sigma> = ts ! i \<cdot> \<tau> i" by (auto intro: term_subst_eq)
    with ** have "ss = map (\<lambda>t. t \<cdot> \<sigma>) ts" using l by (auto intro!: nth_equalityI)
    moreover have "\<forall>x\<in>set ts. \<forall>x\<in>vars_term x. ground (\<sigma> x)"
      using ** *** l by (auto simp: in_set_conv_nth)
    moreover have "(f, length ss) \<in> F" using l f by auto
    moreover have "(\<Union>x\<in>set ts. funas_term (x \<cdot> \<sigma>)) \<subseteq> F"
      using ** 4 l by (auto simp: in_set_conv_nth) blast
    ultimately have "\<exists>\<sigma>. ss = map (\<lambda>t. t \<cdot> \<sigma>) ts \<and>
              (\<forall>x\<in>set ts. \<forall>x\<in>vars_term x. ground (\<sigma> x)) \<and>
              (f, length ss) \<in> F \<and>
              (\<Union>x\<in>set ts. funas_term (x \<cdot> \<sigma>)) \<subseteq> F" by blast
  }
  then show ?thesis using assms by (auto simp: ground_instances_def)
                                   (metis UN_subset_iff nth_mem)
qed

lemma Fun_in_ground_instances_VarD:
  assumes "Fun f ss \<in> ground_instances F (Var x)" and "s \<in> set ss"
  shows "s \<in> ground_instances F (Var x)"
using assms by (fastforce simp: ground_instances_def)

lemma args_in_ground_instances_FunD:
  assumes "Fun f ss \<in> ground_instances F (Fun f ts)" and "length ss = length ts" and "i < length ss"
  shows "ss ! i \<in> ground_instances F (ts ! i)"
using assms by (fastforce simp: ground_instances_def)

lemma ground_term_in_ground_instances_Var:
  assumes "fground F s"
  shows "s \<in> ground_instances F (Var x)"
using assms unfolding ground_instances_def by auto

lemma linear_args_ground_instances:
  assumes "linear_term (Fun f ts)" and "f = g" and "(f, length ts) \<in> F"
    and "length ts = length ss" and "\<forall>i < length ss. ss ! i \<in> ground_instances F (ts ! i)"
  shows "Fun g ss \<in> ground_instances F (Fun f ts)"
proof -
  have "\<forall>i < length ss. \<exists> \<sigma>. ss ! i = (ts ! i) \<cdot> \<sigma> \<and> ground ((ts ! i) \<cdot> \<sigma>)
    \<and> funas_term ((ts ! i) \<cdot> \<sigma>) \<subseteq> F"
    using assms(5) by (auto simp: ground_instances_def)
  then obtain \<sigma> where \<sigma>i: "\<forall>i < length ss. (ss ! i) = (ts ! i) \<cdot> \<sigma> i \<and> ground ((ts ! i) \<cdot> \<sigma> i)
    \<and> funas_term ((ts ! i) \<cdot> \<sigma> i) \<subseteq> F"
    by metis
  from subst_merge[of ts \<sigma>] assms(1) obtain \<tau>
    where "\<forall>i < length ts. \<forall>x \<in> vars_term (ts ! i). \<tau> x = \<sigma> i x" by auto
  then have "\<forall>i < length ts. ss ! i = (ts ! i) \<cdot> \<tau> \<and> fground F ((ts ! i) \<cdot> \<tau>)"
    using \<sigma>i assms(4) term_subst_eq by fastforce
  then have "Fun g ss = Fun f ts \<cdot> \<tau>" and "fground F (Fun f ts \<cdot> \<tau>)"
    using assms by (auto simp: in_set_conv_nth, metis map_nth_eq_conv) blast
  then show ?thesis using assms(1) by (auto simp: ground_instances_def)
qed

locale etac =
  fixes c :: "'f" (* a fresh constant function symbol *)
    and F :: "'f sig"
    and R :: "('f, 'v) trs"
    and a :: "'f" (* the signature has to contain at least one constant *)
    and A :: "('q, 'f) ta"
  assumes nonempty: "(a, 0) \<in> F"
    and fresh: "(c, 0) \<notin> F"
    and rinf: "funas_trs R \<subseteq> F"
    and ainf: "funas_ta A \<subseteq> F"
    and novar: "\<forall> (l, r) \<in> R. is_Fun l"
    and linear: "\<forall> (l, r) \<in> R. linear_term l \<and> linear_term r"
    and growing: "growing R"
    and accessible: "\<forall>q \<in> states A. accessible TYPE('v) A q"
    and finiteF: "finite F"
    and finiteR: "finite R"
    and finite_qA: "finite (states A)"
    and finite_tA: "finite (ta_rules A)"
begin

abbreviation star_symbol ("\<star>") where "\<star> \<equiv> Fun c []"

definition restrict :: "('v \<Rightarrow> ('f, 'q, 'v) crterm) \<Rightarrow> 'v set \<Rightarrow> 'v \<Rightarrow> ('f, 'q, 'v) crterm"
where
  "restrict \<theta> V = (\<lambda>x. (if x \<in> V then \<theta> x else State (Inr \<star>)))"

lemma restrict_state_subst_eq:
  "t \<cdot> \<theta> = t \<cdot> (restrict \<theta> (vars_term t))"
proof (induct t)
  case (Var x)
  then show ?case by (auto simp: restrict_def)
next
  case (Fun f ts)
  { fix i
    assume "i < length ts"
    then have "ts ! i \<in> set ts" by auto
    from Fun [OF this] have "ts ! i \<cdot> \<theta> = ts ! i \<cdot> (restrict \<theta> (vars_term (ts ! i)))"
      by (simp add: term_subst_eq_conv)
  }
  then have "\<forall>i < length ts. ts ! i \<cdot> \<theta> = ts ! i \<cdot> (restrict \<theta> (vars_term (ts ! i)))" by blast
  then show ?case using restrict_def term_subst_eq_conv by fastforce
qed

section \<open>Matching and propagating auxiliary tree automaton 'mp_ta'\<close>

(*abbreviation \<sigma>\<^sub>c :: "'v \<Rightarrow> ('f, 'v) term" where "\<sigma>\<^sub>c \<equiv> \<lambda>_. \<star>"*)

(* matching and propagating TA, called 'B' in AMs slides *)
definition mp_ta :: "(('f ,'v) term, 'f) ta"
where
"mp_ta = TA (
   { TA_rule f (map (\<lambda>t. star c t) ts) (star c t) | t f ts. t \<in> sr R \<and> t = Fun f ts } \<union> (* matching rules *)
   sig_rules F \<star> (* propagation rules *)
 )"

lemma star_rule_in_mp_ta:
  assumes "(f, n) \<in> F"
  shows "(Fun (Inl f) (replicate n (State \<star>)), State \<star>) \<in> move mp_ta"
  (is "(Fun (Inl f) ?ss, ?c) \<in> move ?A")
proof -
  from assms  have "TA_rule f (replicate n \<star>) \<star> \<in> ta_rules ?A"        
    by (auto simp: mp_ta_def sig_rules_def)
  from move.intros(1)[OF this, of "Hole"] show "(Fun (Inl f) ?ss, ?c) \<in> move ?A" by simp
qed

lemma states_mp_ta:
  shows "states mp_ta = { star c t | t. t \<in> sr R } \<union> { \<star> }" (is "?L = ?R")
proof (intro equalityI subsetI)
  fix q
  assume "q \<in> ?L"
  then show "q \<in> ?R"
  by (auto simp: states_def r_states_def mp_ta_def sig_rules_def) (blast intro: subterm_in_sr)
next
  fix q
  assume "q \<in> ?R"
  then show "q \<in> ?L"
  proof
    assume "q \<in> {star c t |t. t \<in> sr R}"
    then obtain t where "q = star c t" and "t \<in> sr R" by auto
    show "q \<in> ?L"
    proof (cases t)
      case (Var x)
      then have "q = \<star>" using \<open>q = star c t\<close> by simp
      then show "q \<in> ?L" using nonempty 
        by (auto simp: states_def sig_rules_def r_states_def mp_ta_def) (fastforce)+
    next
      case (Fun f ts)
      then show "q \<in> ?L"
        using  \<open>t \<in> sr R\<close> `q = star c t`
        by (force simp: mp_ta_def states_def r_states_def)
    qed
  next
    assume "q \<in> {\<star>}"
    then show "q \<in> ?L" using nonempty 
      by (auto simp: states_def sig_rules_def r_states_def mp_ta_def) (fastforce)+
  qed
qed

lemma non_Var_sr_term_ta_rule_mp_ta:
  assumes "t \<in> sr R" and "t = Fun f ts"
  shows "TA_rule f (map (\<lambda>t. star c t) ts) (star c t) \<in> ta_rules mp_ta"
using assms unfolding mp_ta_def by auto

lemma funas_ta_mp_ta:
  "funas_ta mp_ta \<subseteq> F"
proof -
  let ?mts = "{TA_rule f (map (\<lambda>t. star c t) ts) (star c t) |t f ts. t \<in> sr R \<and> t = Fun f ts}"
  let ?pts = "{TA_rule f (replicate n \<star>) \<star> |f n. (f, n) \<in> F} :: (('f, 'v) term,'f) ta_rule set"
  { fix f qs q
    assume "TA_rule f qs q \<in> ?mts"
    then have "(f, length qs) \<in> F" using Fun_in_sr_imp_F rinf by force
    then have "funas_ta_rule (TA_rule f qs q) \<subseteq> F" unfolding funas_ta_rule_def by simp
  }
  then have "\<forall>r \<in> ?mts. funas_ta_rule r \<subseteq> F" by blast
  moreover
  { fix f qs q
    assume "TA_rule f qs q \<in> ?pts"
    then have "(f, length qs) \<in> F" using Fun_in_sr_imp_F by force
    then have "funas_ta_rule (TA_rule f qs q) \<subseteq> F" unfolding funas_ta_rule_def by simp
  }
  then have "\<forall>r \<in> ?pts. funas_ta_rule r \<subseteq> F" by blast
  ultimately show "funas_ta mp_ta \<subseteq> F" unfolding funas_ta_def mp_ta_def sig_rules_def by force
qed

lemma mp_ta_all_states_accessible:
  assumes "q \<in> states mp_ta"
  shows "accessible TYPE('v) mp_ta q"
proof -
  have "q \<in> { star c t | t. t \<in> sr R } \<union> { \<star> }" using states_mp_ta assms by simp
  then show ?thesis
  proof
    assume "q \<in> {star c t |t. t \<in> sr R}"
    then obtain t where "q = star c t" and "t \<in> sr R" by auto
    then show ?thesis
    proof (induction t arbitrary: q)
      case (Var x)
      then have q: "q = \<star>" by simp
      have "TA_rule a [] \<star> \<in> ta_rules mp_ta" by (simp add: nonempty mp_ta_def sig_rules_def)
      moreover
      have "(\<star>, \<star>) \<in> (ta_eps mp_ta)\<^sup>*" ..
      moreover
      have "\<forall>i < length []. [] ! i \<in> reachable_states mp_ta ([] ! i)" by simp
      ultimately
      have *: "\<star> \<in> reachable_states mp_ta (fterm (Fun a []))" by auto
      have "ground (Fun a [])" using nonempty by auto
      then show ?case using q * unfolding accessible_def using ground_map_funs_term by blast
    next
      case (Fun f ts)
      note IH = this(1)
      { fix i
        assume "i < length ts"
        then have *: "ts ! i \<in> set ts" by simp
        let ?q = "star c (ts ! i)"
        have "ts ! i \<in> sr R" using Fun(3) * using subterm_in_sr by blast
        from IH[OF * _ this, of ?q] have "accessible TYPE('v) mp_ta ?q" by simp
        then have "\<exists>u :: ('f, 'v) term. ground (fterm u :: ('f + ('f, 'v) term, 'v) term) \<and> ?q \<in> reachable_states mp_ta (fterm u)"
          by (simp add: accessible_def)
      }
      then obtain h :: "nat \<Rightarrow> ('f, 'v) term" where *: "\<forall>i < length ts.
        ground (fterm (h i) :: ('f + ('f, 'v) term, 'v) term)
        \<and> (star c (ts ! i)) \<in> reachable_states mp_ta (fterm (h i))"
        by metis
      def us \<equiv> "map h [0..<length ts]"
      have 1: "TA_rule f (map (\<lambda>t. star c t) ts) (star c (Fun f ts)) \<in> ta_rules mp_ta"
        using Fun(3) by (auto simp: mp_ta_def)
      have 2: "length us = length ts" by (simp add: us_def)
      have **: "\<forall>i < length us. ground (fterm (us ! i)) \<and>
        star c (ts ! i) \<in> reachable_states mp_ta (fterm (us ! i))" using * by (simp add: us_def)
      then have "\<forall>i < length us. (fterm (us ! i), State (star c (ts ! i))) \<in> (move mp_ta)\<^sup>*"
        by (auto simp: move_reachable_states)
      with 2 have "(Fun (Inl f) (map fterm us), fqterm f (map (\<lambda>t. star c t) ts)) \<in> (move mp_ta)\<^sup>*"
        by (intro args_steps_imp_steps[OF ctxt_closed_move]) auto
      then have 3: "(fterm (Fun f us), fqterm f (map (\<lambda>t. star c t) ts)) \<in> (move mp_ta)\<^sup>*" by simp
      have ***: "ground (fterm (Fun f us))" using us_def ** Fun_in_sr_imp_F [OF Fun(3)] by auto
      have "(\<box>\<langle>fqterm f (map (\<lambda>t. star c t) ts)\<rangle>, \<box>\<langle>State (star c (Fun f ts))\<rangle>) \<in> move mp_ta"
        using move.trans[OF 1] by blast
      then have 4: "(fqterm f (map (\<lambda>t. star c t) ts), State (star c (Fun f ts))) \<in> (move mp_ta)\<^sup>*" by auto
      have "(fterm (Fun f us), State (star c (Fun f ts))) \<in> (move mp_ta)\<^sup>*" using rtrancl_trans [OF 3 4] .
      from move_reachable_states [THEN iffD1, OF this]
        have "star c (Fun f ts) \<in> reachable_states mp_ta (fterm (Fun f us))" .
      then show ?case unfolding accessible_def Fun(2) using *** by blast
    qed
  next
    assume "q \<in> {\<star>}"
    then have q: "q = \<star>" ..
    have "TA_rule a [] \<star> \<in> ta_rules mp_ta" by (simp add: nonempty mp_ta_def sig_rules_def)
    moreover
    have "(\<star>, \<star>) \<in> (ta_eps mp_ta)\<^sup>*" ..
    moreover
    have "\<forall>i < length []. [] ! i \<in> reachable_states mp_ta ([] ! i)" by simp
    ultimately
    have *: "\<star> \<in> reachable_states mp_ta (fterm (Fun a []))" by auto
    have "ground (Fun a [])" using nonempty by auto
    then show ?thesis using q * unfolding accessible_def
      using ground_map_funs_term by blast 
  qed
qed

lemma ground_term_accepted_in_c:
  assumes "fground F s"
  shows "(fterm s, State \<star>) \<in> (move mp_ta)\<^sup>*" (is "(_, ?c) \<in> (move ?A)\<^sup>*")
using assms
proof (induct s)
  case (Fun f ss)
  note * = this
  note IH = *(1)
  let ?ss = "replicate (length ss) ?c"
  from Fun(2) have "funas_term (Fun f ss) \<subseteq> F" by blast
  then have "(f, length ss) \<in> F" by simp
  from star_rule_in_mp_ta[OF this] have 1: "(Fun (Inl f) ?ss, ?c) \<in> move ?A" .
  { fix i
    assume "i < length ss"
    from this have "ss ! i \<in> set ss" by simp
    with Fun(2) have "fground F (ss ! i)" by auto
    from \<open>ss ! i \<in> set ss\<close> \<open>funas_term (Fun f ss) \<subseteq> F\<close> have "funas_term (ss ! i) \<subseteq> F" by auto
    from IH[OF \<open>ss ! i \<in> set ss\<close> \<open>fground F (ss ! i)\<close>] have "(fterm (ss ! i), ?c) \<in> (move ?A)\<^sup>*" .
  }
  then have "\<forall> i < length ss. (fterm (ss ! i), State \<star>) \<in> (move ?A)\<^sup>*" by simp
  then have "(Fun (Inl f) (map fterm ss), Fun (Inl f) ?ss) \<in> (move ?A)\<^sup>*"
    by (intro args_steps_imp_steps[OF ctxt_closed_move]) (simp)+
  with 1 show ?case by (simp)    
qed (simp)

lemma ta_eps_mp_ta_eq:
  assumes "(q, q') \<in> (ta_eps mp_ta)\<^sup>*"
  shows "q = q'"
using assms by (auto simp: mp_ta_def)

lemma matching_ta_rule:
  assumes "TA_rule f qs (Fun g (map (\<lambda>x. star c x) ts)) \<in> ta_rules mp_ta"
    and "Fun g ts \<in> sr R" and "(f, length qs) \<in> F"
  shows "f = g \<and> length qs = length ts \<and> (\<forall>i < length qs. qs ! i = star c (ts ! i))"
using assms fresh rinf
apply (auto simp: mp_ta_def sig_rules_def dest!: Fun_in_sr_imp_F)
using map_eq_imp_length_eq by fastforce (metis map_nth_conv)

lemma mp_ta_states_accept_ground_instances:
  assumes "t \<in> sr R" and "funas_term s \<subseteq> F"
  shows "s \<in> ground_instances F t \<longleftrightarrow> (fterm s, State (star c t)) \<in> (move mp_ta)\<^sup>+" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L
  with assms show ?R
  proof (induct s arbitrary: t)
    case (Var x)
    then show ?case unfolding ground_instances_def by fastforce
  next
    case (Fun f ss)
    note * = this
    note IH = *(1)
    note s_ground = *(4)
    let ?c = "State \<star>"
    let ?ss = "replicate (length ss) ?c"
    let ?A = "mp_ta"
    show ?case
    proof (cases t)
      case [simp]: (Var y)
      have "(f, length ss) \<in> F" using *(3) unfolding ground_instances_def by auto
      from star_rule_in_mp_ta[OF this] have "(Fun (Inl f) ?ss, ?c) \<in> move ?A" .
      moreover
      have "(Fun (Inl f) (map fterm ss), Fun (Inl f) ?ss) \<in> (move ?A)\<^sup>*"
        using IH[of _ t] Fun_in_ground_instances_VarD [of f ss F y] Fun
        by (intro args_steps_imp_steps [OF ctxt_closed_move])
           (simp add: UN_subset_iff trancl_into_rtrancl)+
      ultimately show ?thesis by (auto dest: rtrancl_into_trancl1)
    next
      case (Fun g ts)
      with s_ground have eq: "g = f" and len: "length ss = length ts"
        unfolding ground_instances_def by auto
      { fix i
        assume "i < length ss"
        with len have "i < length ts" by simp
        from Fun \<open>t \<in> sr R\<close> eq have sr: "Fun f ts \<in> sr R" by simp
        with \<open>i < length ts\<close> have "(ts ! i) \<in> set (args (Fun f ts))" by simp
        with sr \<open>i < length ts\<close> have "(ts ! i) \<in> sr R"
          using subterm_in_sr[of "Fun f ts" "ts ! i"] by simp
        from *(3) \<open>i < length ss\<close> have "funas_term (ss ! i) \<subseteq> F"
          using subterm_in_sr[of "Fun f ts" "ts ! i"] by (simp add: UN_subset_iff)
        have "ss ! i \<in> ground_instances F (ts ! i)"
          using eq \<open>i < length ss\<close> len s_ground Fun args_in_ground_instances_FunD [of f ss F ts]
          by simp
        then have "(fterm (ss ! i), State (star c (ts ! i))) \<in> (move ?A)\<^sup>*"
          using IH[of "ss ! i" "ts ! i"] \<open>(ts ! i) \<in> sr R\<close> \<open>i < length ss\<close>
          by (simp add: `funas_term (ss ! i) \<subseteq> F`)          
      }
      then have "\<forall> i < length ss. (fterm (ss ! i), State (star c (ts ! i))) \<in> (move ?A)\<^sup>*" by simp
      then have 1: "(Fun (Inl f) (map fterm ss), Fun (Inl f) (map (State \<circ> (\<lambda>x. star c x)) ts)) \<in>
        (move ?A)\<^sup>*" by (intro args_steps_imp_steps [OF ctxt_closed_move]) (simp add: len)+
      have "TA_rule f (map (\<lambda>t. star c t) ts) (star c t) \<in> ta_rules ?A"
        using mp_ta_def \<open>t \<in> sr R\<close> eq Fun by auto
      from move.intros(1) [OF this, of "Hole"]
      have "(fqterm f (map (\<lambda>t. star c t) ts), State (star c t)) \<in> move ?A" by simp
      then have 2: "(Fun (Inl f) (map (State \<circ> (\<lambda>t. star c t)) ts), State (star c t)) \<in> move ?A" by simp
      from 1 2 have "(Fun (Inl f) (map fterm ss), State (star c t)) \<in> (move ?A)\<^sup>+"
        by (simp add: rtrancl_into_trancl1 [OF 1 2])
      then show ?thesis by auto       
    qed
  qed
next
  assume ?R
  {
    assume "(fterm s, State (star c t)) \<in> (move mp_ta)\<^sup>*"
    then have ?L
    using assms
    proof (induct t arbitrary: s)
      case (Var x)
      then have "ground (State (star c (Var x)))" by auto
      from rtrancl_move_ground_ground [OF Var(1) this] have "ground (map_funs_term Inl s)" by simp
      then have "fground F s" using Var(3) by simp
      from ground_term_in_ground_instances_Var [OF this]
      show ?case .
    next
      case (Fun f ts)
      note * = this
      show ?case
      proof (cases s)
        case (Var x)
        then show ?thesis using rtrancl_move_ground_ground [OF Fun.prems(1)] by fastforce
      next
        case (Fun g ss)
        with * show ?thesis
          unfolding move_reachable_states
          apply (auto dest!: ta_eps_mp_ta_eq matching_ta_rule)
          apply (intro linear_args_ground_instances)
          using linear
          apply (auto dest: sr_linear)
          apply (auto dest!: Fun_args_sr)
          by (simp add: UN_subset_iff)
      qed
    qed
  }
  with \<open>?R\<close> show ?L by (auto simp: trancl_into_rtrancl)
qed

lemma disj_states_A_mp_ta:
  "states (map_states Inl A) \<inter> states (map_states Inr mp_ta) = {}"
by (auto simp: map_states_def states_def r_states_def map_r_states_def)

lemma c_in_states_mp_ta:
  shows "\<star> \<in> states mp_ta"
using nonempty unfolding states_def mp_ta_def sig_rules_def r_states_def by force

lemma sr_Fun_in_states_mp_ta:
  assumes "t \<in> sr R" and "t = Fun f ts"
  shows "star c t \<in> states mp_ta"
using assms unfolding mp_ta_def states_def r_states_def by force

lemma finite_ta_rules_mp_ta:
  "finite (ta_rules mp_ta)"
proof -
have "{Fun f ts | f ts. Fun f ts \<in> sr R} \<subseteq> sr R" (is "?A \<subseteq> _") by blast
  from finite_subset [OF this finite_sr [OF finiteR]] have "finite ?A" .
  let ?h = "\<lambda>t :: ('f, 'v) term. (case t of Fun f ts \<Rightarrow> TA_rule f (map (\<lambda>t. star c t) ts) (star c t))"
  have *: "{TA_rule f (map (\<lambda>t. star c t) ts) (star c t) |t f ts. t \<in> sr R \<and> t = Fun f ts} = ?h ` ?A"
    (is "?B = _") by force
  from finite_imageI [OF \<open>finite ?A\<close>, of ?h] have 1: "finite ?B" using * by auto
  let ?f = "\<lambda>(f, n). TA_rule f (replicate n \<star>) \<star>"
  have **: "{TA_rule f (replicate n \<star>) \<star> |f n. (f, n) \<in> F} = ?f ` F"
    (is "?C = _") by auto
  from finite_imageI [OF finiteF, of ?f] have 2: "finite ?C" unfolding ** [symmetric] .
  show "finite (ta_rules mp_ta)" unfolding mp_ta_def sig_rules_def using 1 2 by auto
qed

lemma finite_states_mp_ta:
  "finite (states mp_ta)"
proof -
  have "finite (\<Union>(r_states ` ta_rules mp_ta))"
  proof (intro finite_Union)
    from finite_imageI [OF finite_ta_rules_mp_ta, of r_states]
      show "finite (r_states ` ta_rules mp_ta)" .
  next
    fix M
    assume "M \<in> r_states ` ta_rules mp_ta"
    then show "finite M" by (auto simp: r_states_def)
  qed
  moreover have "finite (\<Union>((\<lambda>(q, q'). {q, q'}) ` ta_eps mp_ta))" by (auto simp: mp_ta_def)
  moreover have "finite (ta_final mp_ta)" by (auto simp: mp_ta_def)
  ultimately show ?thesis unfolding states_def by blast
qed

lemma non_empty_ta_rules_mp_ta: "ta_rules mp_ta \<noteq> {}"
by (force simp: mp_ta_def sig_rules_def intro: nonempty)

section \<open>Second auxiliary tree automaton 'cr_ta'\<close>

definition cr_ta :: "('q + ('f, 'v) term, 'f) ta" where
  "cr_ta = ta_union (map_states Inl A) (map_states Inr mp_ta)"

lemma non_empty_ta_rules_cr_ta:
  "ta_rules cr_ta \<noteq> {}"
proof -
  from non_empty_ta_rules_map_states [OF non_empty_ta_rules_mp_ta inj_Inr]
    have "ta_rules (map_states Inr mp_ta) \<noteq> {}" .
  from non_empty_ta_rules_ta_union [OF this]
    show ?thesis unfolding cr_ta_def by auto
qed

lemma funas_ta_cr_ta:
  "funas_ta cr_ta \<subseteq> F"
proof -
  have *: "funas_ta (map_states Inr mp_ta) \<subseteq> F" using funas_ta_mp_ta funas_ta_map_states by blast
  have **: "funas_ta (map_states Inl A) \<subseteq> F" using ainf funas_ta_map_states by blast
  from funas_ta_ta_union [OF ** *] show ?thesis unfolding cr_ta_def by simp
qed

lemma lang_cr_ta_eq_lang_A:
  "lang cr_ta = lang A"
proof -
  have "ta_final (map_states Inr mp_ta) = {}"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    moreover have "ta_final mp_ta = {}" by (simp add: mp_ta_def)
    ultimately show "False" by (auto simp: map_states_def) 
  qed
  then show ?thesis
    using disj_union_empty_ta_final_imp_lang[OF _ disj_states_A_mp_ta]
    by (simp add: cr_ta_def lang_map_states_iff)
qed

lemma cr_ta_states_accept_ground_instances:
  assumes "t \<in> sr R" and "funas_term s \<subseteq> F"
  shows "s \<in> ground_instances F t \<longleftrightarrow> (fterm s, State (Inr (star c t))) \<in> (move cr_ta)\<^sup>+"
  (is "?L \<longleftrightarrow> ?R")
proof -
  let ?A = "map_states Inl A"
  let ?B = "map_states Inr mp_ta"
  let ?q' = "star c t"
  let ?q = "Inr ?q'"
  let ?t = "fterm s"
  show ?thesis
  proof
    have *: "states ?B \<inter> states ?A = {}" using disj_states_A_mp_ta by auto
    assume ?L
    then have "(?t, State ?q') \<in> (move mp_ta)\<^sup>*"
      using mp_ta_states_accept_ground_instances [OF assms] by simp
    then have "(?t, State ?q) \<in> (move (map_states Inr mp_ta))\<^sup>*"
      using move_rtrancl_map_states_iff[OF inj_Inr, of s "?q'" mp_ta] by simp
    then have **: "(?t, State ?q) \<in> (move (ta_union ?A ?B))\<^sup>*"
      using disj_move_rtrancl_imp_ta_union_move_rtrancl [OF *, of "?t" "State ?q"] by auto
    have "(?t, State ?q) \<in> (move (ta_union ?A ?B))\<^sup>+"
      using ** by (simp add: fterm_no_State rtrancl_eq_or_trancl)
    then show ?R by (auto simp: cr_ta_def)
  next
    assume ?R
    then have "(?t, State ?q) \<in> (move cr_ta)\<^sup>*" by simp
    from move_reachable_states [THEN iffD1, OF this]
      have "?q \<in> reachable_states cr_ta ?t" .
    then have "?q \<in> reachable_states (ta_union ?A ?B) ?t"
      by (simp add: cr_ta_def)
    then have "?q \<in> reachable_states ?A ?t \<or> ?q \<in> reachable_states ?B ?t"
      using disj_reachable_states_eq_union [OF disj_states_A_mp_ta] by force
    then have "?q \<in> reachable_states ?B ?t" by fastforce
    with reachable_states_map_states_iff [OF inj_Inr, of mp_ta s, symmetric]
    have "?q' \<in> reachable_states mp_ta ?t" by fast
    from move_reachable_states [THEN iffD2, OF this]
      have "(?t, State ?q') \<in> (move mp_ta)\<^sup>*" .
    moreover have "fterm s \<noteq> State ?q'" by (simp add: fterm_no_State)
    ultimately have "(fterm s, State ?q') \<in> (move mp_ta)\<^sup>+"
        by (simp add: \<open>fterm s \<noteq> State ?q'\<close> rtrancl_eq_or_trancl)
    then have "(fterm s, State (star c t)) \<in> (move mp_ta)\<^sup>+" .
    with mp_ta_states_accept_ground_instances [OF assms, symmetric] show ?L by simp
  qed
qed

fun qi :: "('f, 'v) term \<Rightarrow> ('v \<Rightarrow> 'q + ('f, 'v) term) \<Rightarrow> ('f, 'v) term \<Rightarrow> 'q + ('f, 'v) term"
where
  "qi t g (Var x) = (if x \<in> vars_term t then g x else Inr \<star>)"
| "qi t g (Fun f ts) = Inr (star c (Fun f ts))"

lemma states_cr_ta_iff:
  shows "states cr_ta = states (map_states Inl A) \<union> states (map_states Inr mp_ta)"
unfolding cr_ta_def by auto

lemma finite_states_cr_ta:
  "finite (states cr_ta)"
proof -
  from states_map_states_iff [OF inj_Inl]
    have *: "states (map_states Inl A) = Inl ` states A" (is "?A = _")
    using states_map_states_ex by fastforce
  from finite_imageI [OF finite_qA, of Inl]
    have 1: "finite ?A" unfolding * [symmetric] .
  from states_map_states_iff [OF inj_Inr]
    have **: "states (map_states Inr mp_ta) = Inr ` states mp_ta" (is "?B = _")
    using states_map_states_ex by fastforce
  from finite_imageI [OF finite_states_mp_ta, of Inr]
    have 2: "finite ?B" unfolding ** [symmetric] .
  thm states_cr_ta_iff finite_qA finite_Un
  from finite_Un [THEN iffD2, of ?A ?B] 1 2 have "finite (?A \<union> ?B)" by blast
  then show ?thesis unfolding states_cr_ta_iff .
qed

lemma finite_ta_rules_cr_ta:
  "finite (ta_rules cr_ta)"
proof -
  have [simp]: "ta_rules (map_states Inl A) = map_r_states Inl ` ta_rules A"
    (is "_ = ?A") by (auto simp: map_states_def)
  have [simp]: "ta_rules (map_states Inr mp_ta) = map_r_states Inr ` ta_rules mp_ta"
    (is "_ = ?B") by (auto simp: map_states_def)
  from finite_imageI [OF finite_tA] have 1: "finite ?A" by blast
  from finite_imageI [OF finite_ta_rules_mp_ta] have 2: "finite ?B" by blast
  have "ta_rules cr_ta = ?A \<union> ?B" by (simp add: cr_ta_def)
  with 1 2 finite_Un [symmetric] show ?thesis unfolding cr_ta_def by simp
qed

lemma c_in_states_cr_ta:
  shows "Inr \<star> \<in> states cr_ta"
using c_in_states_mp_ta by (auto simp: cr_ta_def)

lemma sr_Fun_in_states_cr_ta:
  assumes "t \<in> sr R" and "t = Fun f ts"
  shows "Inr (star c t) \<in> states cr_ta"
using sr_Fun_in_states_mp_ta [OF assms] by (auto simp: cr_ta_def)

lemma cr_ta_all_states_accessible:
  assumes "q \<in> states cr_ta"
  shows "accessible TYPE('v) cr_ta q"
proof -
  from assms states_cr_ta_iff have *: "q \<in> states (map_states Inl A) \<or>
    q \<in> states (map_states Inr mp_ta)" by auto
  { assume a: "q \<in> states (map_states Inl A)"
    from states_map_states_ex [OF this] obtain q' where q': "q = Inl q'" ..
    have a': "q' \<in> states A" using a unfolding q' using states_map_states_iff [OF inj_Inl] by fast
    with accessible have "accessible TYPE('v) A q'" ..
    from accessible_map_states [OF inj_Inl, of A q'] this have
      "accessible TYPE('v) (map_states Inl A) q" unfolding q' by auto
    from accessible_disj_states_ta_union [OF this disj_states_A_mp_ta]
      have "accessible TYPE('v) cr_ta q" unfolding cr_ta_def .
  }
  moreover
  { assume a: "q \<in> states (map_states Inr mp_ta)"
    from states_map_states_ex [OF this] obtain q' where q': "q = Inr q'" ..
    from disj_states_A_mp_ta have disj:
      "states (map_states Inr mp_ta) \<inter> states (map_states Inl A) = {}" by blast
    have a': "q' \<in> states mp_ta" using a unfolding q' using states_map_states_iff [OF inj_Inr]
      by fast
    from mp_ta_all_states_accessible [OF this] have "accessible TYPE('v) mp_ta q'" .
    from accessible_map_states [OF inj_Inr, of mp_ta q'] this have
      "accessible TYPE('v) (map_states Inr mp_ta) q" unfolding q' by auto
    from accessible_disj_states_ta_union [OF this disj]
      have "accessible TYPE('v) cr_ta q" unfolding cr_ta_def by simp
  }
  ultimately show ?thesis using * unfolding states_cr_ta_iff [symmetric] by auto
qed

lemma finite_aux:
  assumes "finite Q"
  shows "finite {(f, qs) |f ls r qs. (Fun f ls, r) \<in> R \<and> set qs \<subseteq> Q \<and> length qs = length ls}"
    (is "finite ?A")
proof -
  have *: "?A = \<Union>{{(f, qs) |qs. set qs \<subseteq> Q \<and> length qs = length ls} | f ls r. (Fun f ls, r) \<in> R}"
    (is "_ = \<Union>?B") by auto
  show ?thesis unfolding *
  proof (intro finite_Union)
    let ?h = "\<lambda>(l, r). (case l of Fun f ls \<Rightarrow> {(f, qs) |qs. set qs \<subseteq> Q \<and> length qs = length ls} |
      Var _ \<Rightarrow> {})"
    have **: "?B = ?h ` R"
      using novar by (auto simp: image_iff split: term.splits prod.splits) fastforce+
    show "finite ?B" unfolding ** using finite_imageI [OF finiteR] .
  next
    fix M
    assume "M \<in> ?B"
    then show "finite M" using finite_lists_length_eq [OF assms] by auto
  qed
qed

lemma finite_lhs_ta_rules_over_finite_states:
  assumes "finite Q"
  shows "finite { TA_rule f qs q | f ls r qs q.
    (Fun f ls, r) \<in> R \<and> set qs \<subseteq> Q \<and> length qs = length ls \<and> q \<in> Q}" (is "finite ?A")
proof -
  let ?h = "\<lambda>((f, qs), q). TA_rule f qs q"
  note * = finite_imageI [OF finite_cartesian_product [OF finite_aux [OF assms] assms], of ?h]
  show ?thesis by (rule finite_subset [OF _ *]) (force simp: image_iff)
qed

section \<open>The exactly completed tree automaton \<open>C\<^sub>R\<close>\<close>

inductive_set \<Delta>\<^sub>C where
  base: "r \<in> ta_rules cr_ta \<Longrightarrow> r \<in> \<Delta>\<^sub>C"
| step: "(Fun f ls, r) \<in> R \<Longrightarrow>
   range \<theta> \<subseteq> states cr_ta \<Longrightarrow>
   q \<in> states cr_ta \<Longrightarrow>
   (fterm r \<cdot> (State \<circ> \<theta>), (State q) :: ('f, 'q, 'v) crterm)
     \<in> (move (cr_ta\<lparr> ta_rules := \<Delta>\<^sub>C \<rparr>))\<^sup>* \<Longrightarrow> TA_rule f (map (\<lambda>li. qi r \<theta> li) ls) q \<in> \<Delta>\<^sub>C"

lemma \<Delta>\<^sub>C_subseteq_finite_set:
  "\<Delta>\<^sub>C \<subseteq> ta_rules cr_ta \<union> { TA_rule f qs q |f ls r qs q. (Fun f ls, r) \<in> R \<and> set qs \<subseteq> states cr_ta
    \<and> length qs = length ls \<and> q \<in> states cr_ta}" (is "_ \<subseteq> _ \<union> ?\<Delta>")
proof
  fix r
  assume "r \<in> \<Delta>\<^sub>C"
  then show "r \<in> ta_rules cr_ta \<union> ?\<Delta>"
  proof (induct)
    case (base r')
    then show ?case by auto
  next
    case (step f ls r' \<theta> q)
    { fix q
      assume "q \<in> set (map (qi r' \<theta>) ls)"
      then have "q \<in> states cr_ta"
      apply auto
      apply (case_tac x)
      using step(1,2) c_in_states_cr_ta sr_Fun_in_states_cr_ta
      apply (auto simp: sr_def)
      by blast
    }
    then have 2: "set (map (qi r' \<theta>) ls) \<subseteq> states cr_ta" by blast
    have 3: "length (map (qi r' \<theta>) ls) = length ls" by simp
    from step(1,3) 2 3 have "TA_rule f (map (qi r' \<theta>) ls) q \<in> ?\<Delta>" by blast
    then show ?case by auto
  qed
qed

lemma finite_\<Delta>\<^sub>C:
  "finite \<Delta>\<^sub>C"
using finite_qA and finite_tA
proof -
  let ?\<Delta> = "{ TA_rule f qs q |f ls r qs q. (Fun f ls, r) \<in> R \<and> set qs \<subseteq> states cr_ta
    \<and> length qs = length ls \<and> q \<in> states cr_ta}"
  from finite_ta_rules_cr_ta
    and finite_lhs_ta_rules_over_finite_states [OF finite_states_cr_ta]
    have "finite (ta_rules cr_ta \<union> ?\<Delta>)" by blast
  from finite_subset [OF \<Delta>\<^sub>C_subseteq_finite_set this] show ?thesis .
qed

lemma funas_ta_rule_\<Delta>\<^sub>C:
  assumes "r \<in> \<Delta>\<^sub>C"
  shows "funas_ta_rule r \<subseteq> F"
using assms
proof (cases)
  case (base)
  then show ?thesis using funas_ta_cr_ta unfolding funas_ta_def by blast
next
  case (step f ls r \<theta> q)
  from step(2) have "funas_rule (Fun f ls, r) \<subseteq> F" using rinf by (simp add: funas_trs_def) blast
  then have "funas_term (Fun f ls) \<subseteq> F" by (simp add: funas_rule_def)
  then show ?thesis using step(1) by (auto simp: funas_ta_rule_def)
qed

definition "C\<^sub>R = cr_ta\<lparr> ta_rules := \<Delta>\<^sub>C \<rparr>"

lemma funas_ta_cr:
  "funas_ta C\<^sub>R \<subseteq> F"
unfolding C\<^sub>R_def funas_ta_def using funas_ta_rule_\<Delta>\<^sub>C by auto

lemma cr_ta_trans_subseteq_cr_trans:
  "ta_rules cr_ta \<subseteq> ta_rules C\<^sub>R" (is "?L \<subseteq> ?R")
proof
  fix t
  assume "t \<in> ?L"
  then show "t \<in> ?R" by (auto simp: C\<^sub>R_def intro: \<Delta>\<^sub>C.intros(1))
qed

lemma cr_ta_move_subseteq_cr_move:
  "move cr_ta \<subseteq> move C\<^sub>R" (is "?L \<subseteq> ?R")
proof
  fix s t
  assume "(s, t) \<in> ?L"
  then show "(s, t) \<in> ?R"
  apply (auto simp: C\<^sub>R_def)
  apply (induct rule: move.induct)
  apply (auto intro: move.intros)
  using subsetD[OF cr_ta_trans_subseteq_cr_trans] C\<^sub>R_def
  apply (auto simp: move.trans)
  done
qed

lemma cr_ta_move_rtrancl_subseteq_cr_move_rtrancl:
  "(move cr_ta)\<^sup>* \<subseteq> (move C\<^sub>R)\<^sup>*"
using cr_ta_move_subseteq_cr_move rtrancl_mono by blast

lemma reachable_states_cr_ta_subseteq_reachable_states_cr:
  "reachable_states cr_ta t \<subseteq> reachable_states C\<^sub>R t" (is "?L \<subseteq> ?R")
proof
  fix q
  assume "q \<in> ?L"
  from move_reachable_states [THEN iffD2, OF this] have "(t, State q) \<in> (move cr_ta)\<^sup>*" .
  then have "(t, State q) \<in> (move C\<^sub>R)\<^sup>*" using cr_ta_move_rtrancl_subseteq_cr_move_rtrancl by force
  from move_reachable_states [THEN iffD1, OF this] show "q \<in> ?R" .
qed

lemma cr_ta_lang_subseteq_cr_lang:
  "lang cr_ta \<subseteq> lang C\<^sub>R" (is "?L \<subseteq> ?R")
proof
  fix t
  assume "t \<in> ?L"
  then show "t \<in> ?R"
    apply (auto simp: lang_def)
    using C\<^sub>R_def reachable_states_cr_ta_subseteq_reachable_states_cr by fastforce
qed

lemma range_subseteq_qi:
  assumes "range \<theta> \<subseteq> states cr_ta" and "t \<in> sr R"
  shows "qi r \<theta> t \<in> states cr_ta"
using assms c_in_states_cr_ta sr_Fun_in_states_mp_ta cr_ta_def
by (cases t) (auto)

lemma r_states_deltac_subseteq_states_cr_ta:
  assumes "r \<in> \<Delta>\<^sub>C"
  shows "r_states r \<subseteq> states cr_ta"
using assms
proof (induct rule: \<Delta>\<^sub>C.induct)
  case (base r)
  then show ?case by (auto simp: states_def)
next
  case (step f ls r \<theta> q)
  then show ?case using range_subseteq_qi by (auto simp: r_states_def sr_def) blast
qed

lemma states_cr_eq [simp]:
  "states C\<^sub>R = states cr_ta" (is "?R = ?L")
proof -
  have *: "ta_rules C\<^sub>R = \<Delta>\<^sub>C" and
    **: "ta_final C\<^sub>R = ta_final cr_ta" and
    ***: "ta_eps C\<^sub>R = ta_eps cr_ta" by (simp add: C\<^sub>R_def)+
  show ?thesis
  proof (intro equalityI subsetI)
    fix q
    assume "q \<in> ?L"
    then show "q \<in> ?R"
      using cr_ta_trans_subseteq_cr_trans ** ***
      by (auto simp: states_def r_states_def)
  next
    fix q
    assume "q \<in> ?R"
    then show "q \<in> ?L"
      unfolding states_def *** ** * using r_states_deltac_subseteq_states_cr_ta
      by (auto) (fastforce simp: states_def)
  qed
qed

subsection \<open>An iterative inductive version of \<open>\<Delta>\<^sub>C\<close>\<close>

inductive it_k :: "nat \<Rightarrow> ('q + ('f, 'v) term, 'f) ta_rule \<Rightarrow> bool" where
  base: "r \<in> ta_rules cr_ta \<Longrightarrow> it_k k r"
| step: "(Fun f ls, r) \<in> R \<Longrightarrow>
    range \<theta> \<subseteq> states cr_ta \<Longrightarrow>
    q \<in> states cr_ta \<Longrightarrow>
    \<forall>r \<in> \<Delta>. (\<exists>n \<le> k. it_k n r) \<Longrightarrow>
    (fterm r \<cdot> (State \<circ> \<theta>), (State q) :: ('f, 'q, 'v) crterm)
      \<in> (move (cr_ta\<lparr> ta_rules := \<Delta> \<rparr>))\<^sup>* \<Longrightarrow>
    it_k (Suc k) (TA_rule f (map (\<lambda>l. qi r \<theta> l) ls) q)"

lemma it_k_Suc_mono:
  assumes "it_k n r"
  shows "it_k (Suc n) r"
using assms
apply (induct)
apply (auto intro: it_k.base)
apply (rule it_k.step)
apply auto
apply force
done

lemma it_k_mono: "m \<le> n \<Longrightarrow> it_k m r \<Longrightarrow> it_k n r"
apply (induct "n - m" arbitrary: m n)
apply auto
by (metis add_Suc diff_add_inverse2 it_k_Suc_mono le_add2 le_add_diff_inverse2)

lemma it_k_imp_\<Delta>\<^sub>C:
  assumes "it_k k r"
  shows "r \<in> \<Delta>\<^sub>C"
using assms
proof (induct)
  case (base r)
  then show ?case by (simp add: \<Delta>\<^sub>C.base)
next
  case (step f ls r \<theta> q \<Delta> k)
  from step(4) have "\<Delta> \<subseteq> \<Delta>\<^sub>C" by blast
  from move_mono [OF this, of cr_ta] and step(5) have
    "(fterm r \<cdot> (State \<circ> \<theta>), (State q) :: ('f, 'q, 'v) crterm)
      \<in> (move (cr_ta\<lparr>ta_rules := \<Delta>\<^sub>C\<rparr>))\<^sup>*" by blast  
  with step show ?case using \<Delta>\<^sub>C.step by simp
qed

lemma it_k_0 [simp]:
  "{r. it_k 0 r} = ta_rules cr_ta"
by (auto elim: it_k.cases simp: it_k.base)

lemma finite_it_k_0: "finite {r. it_k 0 r}" using finite_ta_rules_cr_ta it_k_0 by presburger

lemma it_k_0_subseteq_it_k: "{r. it_k 0 r} \<subseteq> {r. it_k k r}" using it_k_mono by force

lemma \<Delta>\<^sub>C_imp_ex_it_k:
  assumes "r \<in> \<Delta>\<^sub>C"
  shows "\<exists>k. it_k k r"
using assms
proof (induct)
  case (base r)
  from it_k.base [OF this] obtain k where "it_k k r" by simp
  then show ?case by blast
next
  case (step f ls r \<theta> q)
  let ?\<Delta> = "{r. \<exists>k. it_k k r}"
  have * [simp]: "{x \<in> \<Delta>\<^sub>C. \<exists>k. it_k k x} = ?\<Delta>" using it_k_imp_\<Delta>\<^sub>C by blast
  let ?A = "(\<lambda>r. SOME k. it_k k r) ` ?\<Delta>"
  have "?\<Delta> \<subseteq> \<Delta>\<^sub>C" using * by blast
  from finite_subset [OF this finite_\<Delta>\<^sub>C] have "finite ?\<Delta>" .
  from finite_imageI [OF this] have "finite ?A" by blast
  moreover from non_empty_ta_rules_cr_ta have "{r. it_k 0 r} \<noteq> {}" by simp
  then have "?A \<noteq> {}" by blast
  ultimately have "finite ?A" and "?A \<noteq> {}" .
  then have *: "\<And>x. x \<in> ?A \<Longrightarrow> x \<le> Max ?A" by simp
  { fix x k assume "it_k k x" and x: "x \<in> ?\<Delta>"
    then have "it_k (SOME k. it_k k x) x" by (intro someI)
    from it_k_mono [OF _ this, of "Max ?A"]
      have "it_k (Max ?A) x" using * and x by auto }
  then have Delta: "?\<Delta> = {r. it_k (Max ?A) r}" by auto
  have "\<forall>r \<in> ?\<Delta>. \<exists>k \<le> Max ?A. it_k k r" by (subst Delta) auto
  from it_k.step [OF _ _ _ this] show ?case using step by (auto) blast
qed

primrec clean_term :: "('f, 'v) term \<Rightarrow> ('f, 'v) term" where
  "clean_term (Var x) = Fun a []"
| "clean_term (Fun f ts) = (if (f, length ts) \<in> F then Fun f (map clean_term ts) else Fun a [])"

lemma clean_term_apply_subst:
  assumes "funas_term u \<subseteq> F"
  shows "clean_term (u \<cdot> \<sigma>) = u \<cdot> (clean_term \<circ> \<sigma>)"
using assms by (induct u) (auto)
  
lemma rstep_clean_term_imp:
  assumes "(s, t) \<in> rstep R"
  shows "(clean_term s, clean_term t) \<in> (rstep R)\<^sup>="
using assms
proof
  fix C \<sigma> l r
  assume *: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and **: "(l, r) \<in> R"
  from ** show "(clean_term s, clean_term t) \<in> (rstep R)\<^sup>=" unfolding *
  proof (induct C)
    case (Hole)
    then have "funas_term l \<subseteq> F" and "funas_term r \<subseteq> F" using funas_trs_def funas_rule_def rinf 
      by fastforce+
    then have 1: "clean_term (l \<cdot> \<sigma>) = l \<cdot> (clean_term \<circ> \<sigma>)"
      and 2: "clean_term (r \<cdot> \<sigma>) = r \<cdot> (clean_term \<circ> \<sigma>)"
      using clean_term_apply_subst by fast+
    from Hole have "(l \<cdot> (clean_term \<circ> \<sigma>), r \<cdot> (clean_term \<circ> \<sigma>)) \<in> rstep R" by fast
    then show ?case using 1 2 by simp
  next
    case (More f ts D ss)
    then show ?case
      apply auto
      apply (rule rstep.ctxt [where C = "More f (map clean_term ts) \<box> (map clean_term ss)", simplified])
      apply simp
      done
  qed
qed

lemma fgroun_clean_term [simp]:
  "fground F (clean_term t)"
using nonempty by (induct t) (auto)

lemma fground_imp_clean_term_eq:
  assumes "fground F s"
  shows "clean_term s = s"
using assms by (induct s, auto) (simp add: SUP_le_iff map_idI)

lemma rstep_fground_iff:
  assumes "fground F s" and "fground F t"
  shows "(s, t) \<in> ((rstep R) \<inter> {(s, t). fground F s \<and> fground F t})\<^sup>* \<longleftrightarrow> (s, t) \<in> (rstep R)\<^sup>*"
  (is "_ \<in> ?RTF \<longleftrightarrow> _")
proof
  assume "(s, t) \<in> ?RTF"
  then show "(s, t) \<in> (rstep R)\<^sup>*" using assms by (induct) (auto)
next
  assume "(s, t) \<in> (rstep R)\<^sup>*"
  then have "(s, clean_term t) \<in> ?RTF"
  using assms(1)
  proof (induct)
    case (step u t)
    note IH = step(3) [OF step(4)]
    have "(clean_term u, clean_term t) \<in> (rstep R)\<^sup>*" using rstep_clean_term_imp [OF step(2)] by auto
    moreover have "fground F (clean_term u)" by simp
    moreover have "fground F (clean_term t)" by simp
    ultimately have "(clean_term u, clean_term t) \<in> ?RTF"
      by (metis (mono_tags, lifting) IntI Un_iff mem_Collect_eq r_into_rtrancl
          rstep_clean_term_imp rtrancl_reflcl case_prodI step.hyps(2)) 
    then show ?case using IH by fastforce
  next
    case (base)
    from fground_imp_clean_term_eq [OF this] have "clean_term s = s" .
    then show ?case by auto
  qed
  then show "(s, t) \<in> ?RTF" using fground_imp_clean_term_eq [OF assms(2)] by auto
qed

lemma ancestors_subseteq_lang_cr:
  shows "ancestors F R (lang A) \<subseteq> lang C\<^sub>R" (is "?L \<subseteq> ?R")
proof
  fix t
  assume "t \<in> ?L"
  then obtain s where 1: "(t, s) \<in> (rstep R)\<^sup>*" and 2: "s \<in> (lang A)" and tfg: "fground F t"
    by (auto simp: ancestors_def)
  obtain q where "q \<in> reachable_states A (fterm s)" using 2 unfolding lang_def by blast
  from reachable_states_funas_ta [OF this] have sfg: "fground F s" using ainf 2
    by (auto simp: lang_def)
  from rstep_fground_iff [OF tfg sfg, THEN iffD2, OF 1] have
    "(t, s) \<in> (rstep R \<inter> {(s, t). fground F s \<and> fground F t})\<^sup>*" (is "(t, s) \<in> ?RTF") .
  then show "t \<in> ?R"
  proof (induct rule: converse_rtrancl_induct [case_names base step])
    case (base)
    then show ?case using cr_ta_lang_subseteq_cr_lang lang_cr_ta_eq_lang_A 2 by blast
  next
    case (step t u)
    from step(3) obtain q\<^sub>f where uq: "(fterm u, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" and qf: "q\<^sub>f \<in> ta_final C\<^sub>R" ..
    obtain C f ls r \<sigma> where t: "t = C\<langle>Fun f ls \<cdot> \<sigma>\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and
      rule: "(Fun f ls, r) \<in> R" using step(1) novar by fast
    with uq have "(fterm (C\<langle>r \<cdot> \<sigma>\<rangle>), State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" by blast
    then have "((fctxt C)\<langle>fterm (r \<cdot> \<sigma>)\<rangle>, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" (is "(?D\<langle>_\<rangle>, _) \<in> _") by simp
    from moves_State [OF this]
      obtain q where **: "(fterm (r \<cdot> \<sigma>), State q) \<in> (move C\<^sub>R)\<^sup>*"
      and "3'": "(?D\<langle>State q\<rangle>, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" by blast
    then have *: "(fterm r \<cdot> (fterm \<circ> \<sigma>), State q) \<in> (move C\<^sub>R)\<^sup>*"
      unfolding fterm_subst_distrib by simp
    have lin: "linear_term (fterm r)" using linear rule by fastforce
    from step(3) have "ground u" using lang_def by blast
    with u have g: "ground (r \<cdot> \<sigma>)" by simp
    have 1: "fground F (Fun f ls \<cdot> \<sigma>)" using step(1) unfolding t by simp
    then have gli: "\<forall>i < length ls. ground ((ls ! i) \<cdot> \<sigma>)" by fastforce
    then have "\<forall>x \<in> vars_term (Fun f ls). funas_term (\<sigma> x) \<subseteq> F"
      using subst_image_subterm[of _ f ls \<sigma>] supt_imp_funas_subset 1 by blast
    then have "\<forall>x \<in> vars_term (Fun f ls). funas_term (\<sigma> x) \<subseteq> F" using nonempty by auto
    moreover have "funas_term (Fun f ls) \<subseteq> F" using lhs_wf rule rinf by blast
    ultimately have "funas_term (Fun f ls \<cdot> \<sigma>) \<subseteq> F"  using subst_sig_term_sig_imp by blast
    then have fli: "\<forall>i < length ls. funas_term ((ls ! i) \<cdot> \<sigma>) \<subseteq> F" by force
    { fix x
      assume x: "r \<cdot> \<sigma> = Var x"
      from move_reachable_states [THEN iffD1, OF **] have "q \<in> reachable_states C\<^sub>R (fterm (r \<cdot> \<sigma>))" .
      moreover then have "reachable_states C\<^sub>R (fterm (r \<cdot> \<sigma>)) = {}" using x reachable_states.simps(3)
        by fastforce
      ultimately have "q \<in> states C\<^sub>R" by auto
    }
    moreover
    { fix f ts
      assume f: "r \<cdot> \<sigma> = Fun f ts"
      with ** have "(fterm (Fun f ts), State q) \<in> (move C\<^sub>R)\<^sup>*" by auto
      from move_reachable_states [THEN iffD1, OF this]
        have "q \<in> reachable_states C\<^sub>R (fterm (Fun f ts))" .
      then obtain f qs q' where
        trans: "TA_rule f qs q' \<in> ta_rules C\<^sub>R" and
        eps: "(q', q) \<in> (ta_eps C\<^sub>R)\<^sup>*" and
        len: "length qs = length (map fterm ts)" and
        qs: "\<forall>i < length (map fterm ts). qs ! i \<in> reachable_states C\<^sub>R ((map fterm ts) ! i)"
        using reachable_states.simps(2)[of C\<^sub>R f "map fterm ts"] by force
      have "q' \<in> states C\<^sub>R" using r_states_in_states [OF trans] by blast
      with eps have "q \<in> states C\<^sub>R"
      proof (cases "q' = q")
        case (False)
        with eps have "(q', q) \<in> (ta_eps C\<^sub>R)\<^sup>+"
          by (simp add: rtrancl_eq_or_trancl)
        from trancl_ta_epsD [OF this] have "{q', q} \<subseteq> states C\<^sub>R" .
        then show ?thesis by blast
      qed (auto)
    }
    ultimately have state: "q \<in> states C\<^sub>R" using Term.term.exhaust by blast
    then have state': "q \<in> states cr_ta" by simp
    obtain \<theta> where vars: "\<forall>x \<in> vars_term (fterm r). ((fterm \<circ> \<sigma>) x, (State \<circ> \<theta>) x) \<in> (move C\<^sub>R)\<^sup>*" and
      4: "(fterm r \<cdot> ((State \<circ> \<theta>) :: ('f + 'q + ('f, 'v) term, 'v) subst), State q) \<in> (move C\<^sub>R)\<^sup>*" and
      \<theta>: "range \<theta> \<subseteq> states C\<^sub>R"
      using moves_state_subst [OF * state lin] by force
    from \<theta> have 3: "range \<theta> \<subseteq> states cr_ta" using C\<^sub>R_def states_cr_eq by auto
    from 4 have 5: "(fterm r \<cdot> ((State \<circ> \<theta>) :: ('f + 'q + ('f, 'v) term, 'v) subst), State q) \<in>
      (move (cr_ta\<lparr>ta_rules := \<Delta>\<^sub>C\<rparr>))\<^sup>*" unfolding C\<^sub>R_def by auto
    have rule': "TA_rule f (map (\<lambda>li. qi r \<theta> li) ls) q \<in> \<Delta>\<^sub>C" using \<Delta>\<^sub>C.intros(2) [OF rule 3 state' 5] .
    let ?ls = "map (fterm \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ls"
    let ?qs = "map (State \<circ> (qi r \<theta>)) ls"
    { fix i
      assume l: "i < length (map fterm ls)"
      then have l': "i < length ls" by simp
      have "(fterm (ls ! i) \<cdot> (fterm \<circ> \<sigma>), State (qi r \<theta> (ls ! i))) \<in> (move C\<^sub>R)\<^sup>*"
      proof (cases "ls ! i \<in> Var ` vars_term r")
        case (True)
        then have eq: "State (qi r \<theta> (ls ! i)) = fterm (ls ! i) \<cdot> (State \<circ> \<theta>)" by fastforce
        from True vars have "(fterm (ls ! i) \<cdot> (fterm \<circ> \<sigma>), fterm (ls ! i) \<cdot> (State \<circ> \<theta>)) \<in> (move C\<^sub>R)\<^sup>*"
          by fastforce
        then show ?thesis by (simp add: eq)
      next
        let ?t = "fterm (ls ! i)"
        let ?\<sigma> = "(fterm \<circ> \<sigma>)"
        let ?t' = "?t \<cdot> ?\<sigma>"
        case (False)
        then have eq: "qi r \<theta> (ls ! i) = Inr (star c (ls ! i))" by (cases "ls ! i") auto
        have 1: "ls ! i \<in> sr R" using rule sr_def [of R] l' by force
        moreover have 2: "funas_term (ls ! i \<cdot> \<sigma>) \<subseteq> F" using fli l' by blast
        moreover have 3: "ground (ls ! i \<cdot> \<sigma>)" using gli l' by blast
        ultimately have 4: "(ls ! i) \<cdot> \<sigma> \<in> ground_instances F (ls ! i)" using ground_instances_def
          by blast
        have "(fterm (ls ! i \<cdot> \<sigma>), State (Inr (star c (ls ! i)))) \<in> (move cr_ta)\<^sup>+"
          using cr_ta_states_accept_ground_instances [OF 1 2, THEN iffD1, OF 4] .
        then have "(fterm (ls ! i \<cdot> \<sigma>), State (Inr (star c (ls ! i)))) \<in> (move cr_ta)\<^sup>*" by fast
        then have "(fterm (ls ! i \<cdot> \<sigma>), State (Inr (star c (ls ! i)))) \<in> (move C\<^sub>R)\<^sup>*"
          using cr_ta_move_rtrancl_subseteq_cr_move_rtrancl by auto
        then show ?thesis unfolding eq fterm_subst_distrib[of "ls ! i" \<sigma>] .
      qed
    }
    then have qis: "\<forall>i < length (map fterm ls).
      (fterm (ls ! i) \<cdot> (fterm \<circ> \<sigma>), State (qi r \<theta> (ls ! i))) \<in> (move C\<^sub>R)\<^sup>*" by blast
    moreover have "length (map fterm ls) = length ?ls" by simp
    ultimately have qis': "\<forall> i < length ?ls. (?ls ! i, ?qs ! i) \<in> (move C\<^sub>R)\<^sup>*" by (auto simp: o_def)
    have len': "length ?ls = length ?qs" by simp
    from args_steps_imp_steps [OF ctxt_closed_move len' qis', of "Inl f"]
      have 1: "(Fun (Inl f) ?ls, Fun (Inl f) ?qs) \<in> (move C\<^sub>R)\<^sup>*" .
    have "TA_rule f (map (qi r \<theta>) ls) q \<in> ta_rules C\<^sub>R" using rule' C\<^sub>R_def by force
    from move.trans[OF this, of \<box>] have 2: "(Fun (Inl f) ?qs, State q) \<in> (move C\<^sub>R)\<^sup>*" by auto
    have 3: "fterm t = ?D\<langle>Fun (Inl f) ?ls\<rangle>" using t by simp
    have "(Fun (Inl f) ?ls, State q) \<in> (move C\<^sub>R)\<^sup>*" using rtrancl_trans[OF 1 2] .
    from ctxt_closedE [OF ctxt_closed_rtrancl[OF ctxt_closed_move] this]
      have "(fterm t, ?D\<langle>State q\<rangle>) \<in> (move C\<^sub>R)\<^sup>*" using 3 by presburger
    then have "(fterm t, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" using "3'" by fastforce
    then have "q\<^sub>f \<in> reachable_states C\<^sub>R (fterm t)" by (simp add: move_reachable_states)
    with qf have "ta_final C\<^sub>R \<inter> reachable_states C\<^sub>R (fterm t) \<noteq> {}" by auto
    moreover have "ground t" using step(1) by blast
    ultimately show ?case unfolding lang_def by auto
  qed
qed

lemma cr_ta_rules:
  "ta_rules C\<^sub>R = ta_rules cr_ta \<union> (\<Delta>\<^sub>C - ta_rules cr_ta)"
unfolding C\<^sub>R_def by (auto simp: \<Delta>\<^sub>C.base)

lemma cr_move_iff'':
  "(s, t) \<in> move C\<^sub>R \<longleftrightarrow> (s, t) \<in> move (ta_union cr_ta (TA (\<Delta>\<^sub>C - ta_rules cr_ta)))"
unfolding ta_union_def using cr_ta_rules by (auto simp: C\<^sub>R_def cr_ta_def ta_union_def)

lemma cr_move_iff:
  "(s, t) \<in> (move C\<^sub>R)\<^sup>+ \<longleftrightarrow> (s, t) \<in> (move (ta_union cr_ta (TA (\<Delta>\<^sub>C - ta_rules cr_ta))))\<^sup>+"
unfolding ta_union_def using cr_ta_rules by (auto simp: C\<^sub>R_def cr_ta_def ta_union_def)

lemma cr_move_iff':
  "(s, t) \<in> (move C\<^sub>R)\<^sup>+ \<longleftrightarrow>
   (s, t) \<in> (move (cr_ta\<lparr> ta_rules := ta_rules cr_ta \<union> (\<Delta>\<^sub>C - ta_rules cr_ta)\<rparr>))\<^sup>+"
  (is "?L \<longleftrightarrow> ?R")
unfolding C\<^sub>R_def using cr_ta_rules by auto (metis Diff_partition \<Delta>\<^sub>C.base cr_ta_rules subsetI)+

lemma cr_moves_imp_it_k_moves:
  assumes "(s, t) \<in> (move C\<^sub>R)\<^sup>+"
  shows "\<exists>k. (s, t) \<in> (move (ta_union cr_ta (TA {r. it_k k r})))\<^sup>+"
proof -
  from assms show ?thesis
  proof (induct)
    case (base t)
    then show ?case
    proof (cases)
      case (trans f qs q C)
      from trans(3) show ?thesis unfolding cr_ta_rules
      proof
        assume "TA_rule f qs q \<in> ta_rules cr_ta"
        then show ?thesis using trans(1,2) 
          by (intro exI [of _ 0]) (simp add: move.trans trancl.r_into_trancl)
      next
        assume "TA_rule f qs q \<in> \<Delta>\<^sub>C - ta_rules cr_ta"
        then have "{TA_rule f qs q} \<subseteq> \<Delta>\<^sub>C - ta_rules cr_ta" by auto
        then obtain n where "{TA_rule f qs q} \<subseteq> {r. it_k n r} - {r. it_k 0 r}"
          using \<Delta>\<^sub>C_imp_ex_it_k by auto
        then have 1: "{TA_rule f qs q} \<subseteq> {r. it_k n r}" by auto
        have "(s, t) \<in> (move (ta_union cr_ta (TA {TA_rule f qs q})))\<^sup>+"
          using trans(1,2) unfolding ta_union_def by (simp add: move.trans trancl.r_into_trancl)
        moreover have "ta_rules (ta_union cr_ta (TA {TA_rule f qs q})) \<subseteq>
          ta_rules (ta_union cr_ta (TA {r. it_k n r}))" using 1 by (auto)
        moreover have "ta_eps (ta_union cr_ta (TA {TA_rule f qs q})) \<subseteq>
          ta_eps (ta_union cr_ta (TA {r. it_k n r}))" by (auto)
        ultimately show "\<exists>k. (s, t) \<in> (move (ta_union cr_ta (TA {r. it_k k r})))\<^sup>+"
          using move_mono'_trancl by blast
      qed
    next
      case (eps q q' C)
      from eps(3) have "(q, q') \<in> (ta_eps cr_ta)\<^sup>*" unfolding C\<^sub>R_def by simp
      moreover then have "(s, t) \<in> (move (ta_union cr_ta (TA {r. it_k 0 r})))\<^sup>+"
        using eps(1,2) by (auto simp: move.eps trancl.r_into_trancl)
      ultimately show ?thesis by blast
    qed
  next
    case (step u t)
    from step(3) obtain n where *: "(s, u) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+" by auto
    from step(2) show ?case
    proof (cases)
      case (trans f qs q C)
      from trans(3) show ?thesis unfolding cr_ta_rules
      proof
        assume "TA_rule f qs q \<in> ta_rules cr_ta"
        then have "TA_rule f qs q \<in> ta_rules (ta_union cr_ta (TA {r. it_k n r}))" by fastforce
        from move.trans [OF this, of C]
          have "(u, t) \<in> move (ta_union cr_ta (TA {r. it_k n r}))" using trans(1,2) by fast
        then have "(u, t) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+" by blast
        then have "(s, t) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+" using * by force
        then show ?thesis by blast
      next
        assume "TA_rule f qs q \<in> \<Delta>\<^sub>C - ta_rules cr_ta"
        then have "{TA_rule f qs q} \<subseteq> \<Delta>\<^sub>C - ta_rules cr_ta" by auto
        then obtain k where "{TA_rule f qs q} \<subseteq> {r. it_k k r} - {r. it_k 0 r}"
          using \<Delta>\<^sub>C_imp_ex_it_k by auto
        then have 1: "{TA_rule f qs q} \<subseteq> {r. it_k k r}" by auto
        have"(u, t) \<in> (move (ta_union cr_ta (TA {TA_rule f qs q})))\<^sup>+"
          using trans(1,2) by (simp add: move.trans trancl.r_into_trancl)
        moreover have "ta_rules (ta_union cr_ta (TA {TA_rule f qs q})) \<subseteq>
          ta_rules (ta_union cr_ta (TA {r. it_k k r}))" using 1 by (auto)
        moreover have "ta_eps (ta_union cr_ta (TA {TA_rule f qs q})) \<subseteq>
          ta_eps (ta_union cr_ta (TA {r. it_k k r}))" by (auto)
        ultimately have k: "(u, t) \<in> (move (ta_union cr_ta (TA {r. it_k k r})))\<^sup>+"
          using move_mono'_trancl by blast
        then show ?thesis
        proof (cases "k \<le> n")
          case (True)
          from it_k_mono [OF this] have s: "{r. it_k k r} \<subseteq> {r. it_k n r}" by blast
          then have "ta_rules (ta_union cr_ta (TA {r. it_k k r})) \<subseteq>
            ta_rules (ta_union cr_ta (TA {r. it_k n r}))" by (auto)
          moreover have "ta_eps (ta_union cr_ta (TA {r. it_k k r})) \<subseteq>
            ta_eps (ta_union cr_ta (TA {r. it_k n r}))" using s by (auto)
          ultimately have "(u, t) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+"
            using move_mono'_trancl k by blast
          then show ?thesis using * trancl_trans by fast
        next
          case (False)
          then have "n \<le> k" by auto
          from it_k_mono [OF this] have s: "{r. it_k n r} \<subseteq> {r. it_k k r}" by blast
          then have "ta_rules (ta_union cr_ta (TA {r. it_k n r})) \<subseteq>
            ta_rules (ta_union cr_ta (TA {r. it_k k r}))" by (auto)
          moreover have "ta_eps (ta_union cr_ta (TA {r. it_k n r})) \<subseteq>
            ta_eps (ta_union cr_ta (TA {r. it_k k r}))" using s by (auto)
          ultimately have "(s, u) \<in> (move (ta_union cr_ta (TA {r. it_k k r})))\<^sup>+"
            using move_mono'_trancl * by blast
          then show ?thesis using k trancl_trans by fast
        qed
      qed
    next
      case (eps q q' C)
      from eps(3) have "(q, q') \<in> (ta_eps cr_ta)\<^sup>*" unfolding C\<^sub>R_def by simp
      then have "(u, t) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+"
        using eps(1,2) by (auto simp: move.eps trancl.r_into_trancl)
      then have "(s, t) \<in> (move (ta_union cr_ta (TA {r. it_k n r})))\<^sup>+" using * by simp
      then show ?thesis by blast
    qed
  qed
qed

subsection \<open>in_k\<close>

(* A predicate which counts how often a specific ta_rule 'r' is used in a sequence. *)
inductive in_k :: "('b, 'a) ta \<Rightarrow> ('b, 'a) ta_rule \<Rightarrow> nat \<Rightarrow>
  ('a + 'b, 'c) term \<Rightarrow> ('a + 'b, 'c) term \<Rightarrow> bool"
for S r
where
  base: "(s, t) \<in> (move S)\<^sup>* \<Longrightarrow> in_k S r 0 s t"
| step: "in_k S r m s t \<Longrightarrow> (t, u) \<in> move (TA {r}) \<Longrightarrow> t \<noteq> u \<Longrightarrow> in_k S r k u v \<Longrightarrow>
  in_k S r (Suc (m + k)) s v"

lemma in_k_refl:
  shows "\<exists>m. in_k S r m s s"
using in_k.base by blast

lemma in_k_move:
  assumes "(s, u) \<in> (move S)\<^sup>*"
    and "in_k S r n u t"
  shows "in_k S r n s t"
using assms(2,1)
proof (induct)
  case (base u t)
  then have "(s, t) \<in> (move S)\<^sup>*" by auto
  then show ?case by (simp add: in_k.base)
qed (simp add: in_k.step)

lemma in_k_trans:
  assumes "in_k S r m s u"
    and "in_k S r n u t"
  shows "in_k S r (m + n) s t"
using assms
proof (induct)
  case (base s u)
  from in_k_move [OF this] show ?case by simp
next
  case (step k s v w l u)
  have "in_k S r (l + n) w t" using step by simp
  then have "in_k S r (Suc (k + (l + n))) s t" using step by (simp add: in_k.step)
  then show ?case by (simp add: add.assoc)
qed

lemma move_trancl_in_k_ex:
  assumes "(s, t) \<in> (move (ta_union S (TA {r})))\<^sup>+"
  shows "\<exists>m. in_k S r m s t"
using assms
proof (induct)
  case (base u)
  then have "(s, u) \<in> move (ta_union (TA {r}) S)" by simp
  then have "(s, u) \<in> move (TA {r}) \<or> (s, u) \<in> move S"
    using empty_ta_eps_move_union_iff [of "TA {r}" S] by auto
  then show ?case
  proof
    assume "(s, u) \<in> move S"
    then have "in_k S r 0 s u" by (simp add: in_k.base r_into_rtrancl) 
    then show ?thesis by auto
  next
    assume "(s, u) \<in> move (TA {r})"
    moreover obtain n m where "in_k S r m s s" and "in_k S r n u u" using in_k_refl by blast+
    ultimately show ?thesis by (cases "s = u") (auto intro: in_k.intros)
  qed
next
  case (step u t)
  then have "(u, t) \<in> move (ta_union (TA {r}) S)" by simp
  then have "(u, t) \<in> move (TA {r}) \<or> (u, t) \<in> move S"
    using empty_ta_eps_move_union_iff [of "TA {r}" S] by auto
  then show ?case
  proof
    assume "(u, t) \<in> move S"
    then have "in_k S r 0 u t" by (simp add: in_k.base r_into_rtrancl)
    moreover obtain n where "in_k S r n s u" using step(3) by blast
    ultimately show ?thesis using in_k_trans by fast
  next
    assume "(u, t) \<in> move (TA {r})"
    moreover obtain n where "in_k S r n s u" using step(3) by blast
    moreover obtain m where "in_k S r m t t" using in_k_refl by blast
    ultimately show ?thesis by (cases "u = t") (auto intro: in_k.intros)
  qed
qed

lemma in_k_Suc_ex:
  assumes "in_k S r (Suc m) s t"
  shows "\<exists>k u v l. (in_k S r k s u \<and> (u, v) \<in> move (TA {r}) \<and> u \<noteq> v \<and> in_k S r l v t) \<and> m = k + l"
using assms by (cases) (blast)

lemma in_kE:
  assumes "in_k S r m s t"
  shows "m = 0 \<or> (\<exists>u v. in_k S r 0 s u \<and> (u, v) \<in> move (TA {r}) \<and> u \<noteq> v \<and> in_k S r (m - 1) v t)"
using assms
proof (induct)
  case (step m s' t' u k v)
  moreover
  { assume "m = 0"
    then have ?case using step by auto
  }
  moreover
  { fix u' v'
    assume "in_k S r 0 s' u'"
      and "(u', v') \<in> move (TA {r})"
      and " u' \<noteq> v'"
      and "in_k S r (m - 1) v' t'"
    moreover then have "in_k S r ( Suc((m - 1) + k)) v' v" using step by (simp add: in_k.step)
    ultimately have ?case by (simp) (metis One_nat_def add_eq_if step(1,3,4,5))
  }
  ultimately show ?case by blast
qed (simp)

lemma in_k_SucE:
  assumes "in_k S r (Suc k) s t"
  shows "\<exists>u v. in_k S r 0 s u \<and> (u, v) \<in> move (TA {r}) \<and> u \<noteq> v \<and> in_k S r k v t"
using in_kE [OF assms] by auto

lemma in_kE2:
  assumes "in_k S r m s t"
  shows "(s, t) \<in> (move (ta_union S (TA {r})))\<^sup>*"
using assms
proof (induct)
  case (base s t)
  show ?case using move_mono'_rtrancl [OF base, of "ta_union S (TA {r})"] by (auto)
next
  case (step m s t u k v) 
  have "(t, u) \<in> move (ta_union S (TA {r}))" using move_mono' [OF step(3)] by (auto)
  then show ?case using step by fastforce
qed

lemma in_ancestors_of_State:
  assumes "(fterm s, State (Inr (star c t))) \<in> (move C\<^sub>R)\<^sup>+"
    and "t \<in> sr R" and "funas_term s \<subseteq> F"
  shows "s \<in> ancestors F R (ground_instances F t)"
proof -
  obtain k where "(fterm s, State (Inr (star c t))) \<in> (move (ta_union cr_ta (TA ({r. it_k k r}))))\<^sup>+"
    using cr_moves_imp_it_k_moves [OF assms(1)] by blast
  then show ?thesis using assms(2,3)
  proof (induct k arbitrary: s t)
    case (0)
    then have "(fterm s, State (Inr (star c t))) \<in> (move cr_ta)\<^sup>+"by (auto simp: move_mono'_trancl)
    then have "s \<in> ground_instances F t" using 0 cr_ta_states_accept_ground_instances by blast
    moreover then have "fground F s" unfolding ground_instances_def by blast
    ultimately show ?case unfolding ancestors_def by blast
  next
    case (Suc k)
    note IH\<^sub>0 = this(1)
    have "{r. it_k (Suc k) r} = ({r. it_k (Suc k) r} - {r. it_k k r}) \<union> {r. it_k k r}"
      (is "?\<Gamma>\<^sub>1 = ?\<Gamma>\<^sub>2 \<union> ?\<Gamma>\<^sub>3") by (auto simp: it_k_Suc_mono)
    let ?A\<^sub>1 = "ta_union cr_ta (TA ?\<Gamma>\<^sub>1)" and ?A\<^sub>2 = "ta_union cr_ta (ta_union (TA ?\<Gamma>\<^sub>2) (TA ?\<Gamma>\<^sub>3))"
    have "ta_rules ?A\<^sub>1 \<subseteq> ta_rules ?A\<^sub>2" and "ta_eps ?A\<^sub>1 \<subseteq> ta_eps ?A\<^sub>2" by auto
    then have 3: "(fterm s, State (Inr (star c t))) \<in> (move ?A\<^sub>2)\<^sup>+"
      using move_mono'_trancl [OF Suc(2)] by auto
    have "finite ?\<Gamma>\<^sub>2" using it_k_imp_\<Delta>\<^sub>C by (intro finite_subset [OF _ finite_\<Delta>\<^sub>C]) auto
    moreover have "?\<Gamma>\<^sub>2 \<subseteq> ?\<Gamma>\<^sub>2" ..
    ultimately show ?case using 3 Suc(3,4)
    proof (induct arbitrary: s t rule: finite_subset_induct')
      case (empty)
      then show ?case using IH\<^sub>0 by (auto simp: move_mono'_trancl)
    next
      case (insert r' \<Gamma>)
      note IH\<^sub>1 = this(5)
      let ?D = "TA (insert r' \<Gamma>)" and ?D\<^sub>1 = "TA \<Gamma>" and ?D\<^sub>2 = "TA {r'}" and ?G = "TA ?\<Gamma>\<^sub>3"
      let ?E = "ta_union cr_ta (ta_union ?D\<^sub>1 ?G)"
      have "?D = ta_union ?D\<^sub>1 ?D\<^sub>2" by auto
      then have *: "ta_union cr_ta (ta_union ?D ?G) = ta_union ?E ?D\<^sub>2" by simp
      have 1: "r' \<in> {r. it_k (Suc k) r}" and 2: "r' \<notin> ta_rules cr_ta"
        using Diff_iff [THEN iffD1, OF insert(3)] it_k_0 it_k_0_subseteq_it_k by auto
      obtain f ls r \<theta> q' \<Delta>' where rule: "r' = TA_rule f (map (qi r \<theta>) ls) q'"
      and r'': "(Fun f ls, r) \<in> R"
      and range: "range \<theta> \<subseteq> states cr_ta"
      and "q' \<in> states cr_ta"
      and \<Delta>': "\<forall>r \<in> \<Delta>'. (\<exists>n \<le> k. it_k n r)"
      and seq: "(fterm r \<cdot> ((State \<circ> \<theta>) :: 'v \<Rightarrow> ('f, 'q, 'v) crterm), State q')
           \<in> (move (cr_ta\<lparr>ta_rules := \<Delta>'\<rparr>))\<^sup>*" using 1 2 by (auto elim: it_k.cases)
      let ?\<theta> = "(State \<circ> \<theta>) :: 'v \<Rightarrow> ('f, 'q, 'v) crterm"
      def \<Delta>\<^sub>k' \<equiv> "(in_k ?E r') :: nat \<Rightarrow> ('f, 'q, 'v) crterm \<Rightarrow> ('f, 'q, 'v) crterm \<Rightarrow> bool"
      from insert(6) have "(fterm s, State (Inr (star c t))) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>+"
        by (simp add: *)
      from move_trancl_in_k_ex [OF this]
        obtain m where "\<Delta>\<^sub>k' m (fterm s) (State (Inr (star c t)))" unfolding \<Delta>\<^sub>k'_def by blast
      then show ?case using insert(2,7,8)
      proof (induction m arbitrary: s t)
        case (0)
        then have "(fterm s, State (Inr (star c t))) \<in> (move ?E)\<^sup>*"
          by (auto simp: \<Delta>\<^sub>k'_def dest: in_k.cases)
        then have "(fterm s, State (Inr (star c t))) \<in> (move ?E)\<^sup>+"
          by (simp add: fterm_no_State rtrancl_eq_or_trancl)
        from IH\<^sub>1 [OF this] 0 show ?case by simp
      next
        case (Suc m')
        note IH\<^sub>2 = this(1) [OF _ Suc(3)]
        let ?s = "fterm s" and ?t = "State (Inr (star c t))"
        obtain qs q where r': "r' = TA_rule f qs q" using rule by auto
        from in_k_SucE [OF Suc(2) [unfolded \<Delta>\<^sub>k'_def]] obtain u v where 1: "\<Delta>\<^sub>k' 0 (fterm s) u"
          and uv: "(u, v) \<in> move (TA {r'})"
          and ne': "u \<noteq> v"
          and tail: "\<Delta>\<^sub>k' m' v ?t" unfolding \<Delta>\<^sub>k'_def by blast
        obtain C where u: "u = C\<langle>fqterm f qs\<rangle>"
          and v: "v = C\<langle>State q\<rangle>"
          and "(C\<langle>fqterm f qs\<rangle>, C\<langle>State q\<rangle>) \<in> move (TA {r'})"
          using uv ne' apply (cases rule: move.cases) using uv by (auto simp: r')
        from 1 have 2: "(?s, C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*" unfolding u \<Delta>\<^sub>k'_def by cases blast
        then obtain s' :: "('f, 'q, 'v) crterm" where s': "s' = ?s" and wf: "s' \<in> wf_terms"
          using fterm_in_wf_terms [of s] by blast
        have "(s', C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*" using 2 unfolding s' by blast
        from distr_moves'' [OF this wf] obtain D ss where s'': "s' = D\<langle>Fun (Inl f) ss\<rangle>"
          and l'': "length (map State qs) = length ss"
          and sqs: "\<forall>i < length ss. (ss ! i, map State qs ! i) \<in> (move ?E)\<^sup>*"
          and "(D\<langle>fqterm f qs\<rangle>, C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*"
          and mc: "(D, C) \<in> move_comp ?E" by force
        from move_comp_imp_moves [OF mc] have dc: "\<forall>t. (D\<langle>t\<rangle>, C\<langle>t\<rangle>) \<in> (move ?E)\<^sup>*" ..
        have len': "length ls = length ss" using l'' rule r' by auto
        have "fterm s = D\<langle>Fun (Inl f) ss\<rangle>" using s' s'' by auto
        then obtain ss' where 9: "ss = map fterm ss'"
          by (metis fterm_args_fterm map_funs_term_ctxt_decomp)
        have x: "fterm s = D\<langle>fterm (Fun f ss')\<rangle>" using s' s'' 9 by auto
        then obtain D' where 12: "D\<langle>fterm (Fun f ss')\<rangle> = (fctxt D')\<langle>fterm (Fun f ss')\<rangle>"
          and D': "D = fctxt D'" using map_funs_term_ctxt_decomp by blast
        then have y: "?s = (fterm D'\<langle>Fun f ss'\<rangle> :: ('f, 'q, 'v) crterm)" by (auto simp: x)
        have "inj fterm" using map_funs_term_inj' [OF inj_Inl [of UNIV]] .
        from inj_onD [OF this y] have s: "s = D'\<langle>Fun f ss'\<rangle>" by auto
        { fix i
          assume i: "i < length ls"
          { assume "ls ! i \<in> Var ` vars_term r"
            moreover def \<tau> \<equiv> "(\<lambda>y. if Var y = ls ! i then ss' ! i else Var y)"
            ultimately have "(ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
              \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i)" by force
          }
          moreover
          { assume a: "ls ! i \<notin> Var ` vars_term r"
            have "qs ! i = qi r \<theta> (ls ! i)" using i rule r' nth_map by auto
            also have "\<dots> = Inr (star c (ls ! i))" using a by (induct r \<theta> "ls ! i" rule: qi.induct) auto
            finally have "qs ! i = Inr (star c (ls ! i))" .
            then have "(fterm (ss' ! i), State (Inr (star c (ls ! i)))) \<in> (move ?E)\<^sup>*"
              using sqs i l'' rule r' by (auto simp: 9)
            moreover have "fterm (ss' ! i) \<noteq> State (Inr (star c (ls ! i)))" using fterm_no_State by fast
            ultimately have b: "(fterm (ss' ! i), State (Inr (star c (ls ! i)))) \<in> (move ?E)\<^sup>+"
              by (auto simp: rtrancl_eq_or_trancl)
            have c: "ls ! i \<in> sr R" unfolding sr_def using i r'' nth_mem by blast
            have "i < length ss'" using i len' 9 by auto
            then have "(ss' ! i) \<unlhd> s" using s by (meson arg_subteq ctxt_supteq nth_mem supteq_trans) 
            then have d: "funas_term (ss' ! i) \<subseteq> F" using Suc by force
            from IH\<^sub>1 [OF b c d] have "ss' ! i \<in> ancestors F R (ground_instances F (ls ! i))" .
            then have "\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
              unfolding ground_instances_def ancestors_def by auto
          }
          ultimately have "\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
            \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i)" by blast
        }
        then have "\<forall>i < length ls. (\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
          \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i))" by blast
        then obtain \<tau>\<^sub>0 where 11: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> (\<tau>\<^sub>0 i)) \<in> (rstep R)\<^sup>*" 
          and 14: "\<forall>i < length ls. (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau>\<^sub>0 i y = ss' ! i)"
          by metis+
        have "is_partition (map vars_term ls)" using linear r'' by fastforce
        from subst_merge [OF this] obtain \<tau>\<^sub>l
          where 16: "\<forall>i < length ls. \<forall>x \<in> vars_term (ls ! i). \<tau>\<^sub>l x = \<tau>\<^sub>0 i x" by blast
        then have \<tau>\<^sub>l: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> \<tau>\<^sub>l) \<in> (rstep R)\<^sup>*"
          using 11 by (metis term_subst_eq)
        have 15: "\<forall>i < length ls. (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau>\<^sub>l y = ss' ! i)"
          using 14 16 by auto
        { fix x
          assume "x \<in> vars_term r - vars_term (Fun f ls)"
          from range have "\<theta> x \<in> states cr_ta" by auto
          from cr_ta_all_states_accessible [OF this, THEN accessible_ground_move_rtrancl]
            have "\<exists>u. ground (fterm u) \<and>
              (fterm u, (State (\<theta> x)) :: ('f, 'q, 'v) crterm) \<in> (move cr_ta)\<^sup>*"
            by auto
        }
        then obtain u where ux: "\<forall>x \<in> vars_term r - vars_term (Fun f ls).
          ground (fterm (u x))
          \<and> (fterm (u x), (State (\<theta> x)) :: ('f, 'q, 'v) crterm) \<in> (move cr_ta)\<^sup>*"
          by metis
        def \<tau>' \<equiv> "\<lambda>y. if y \<in> vars_term r - vars_term (Fun f ls) then u y else Var y"
        def \<tau> \<equiv> "\<lambda>y. if y \<in> vars_term r - vars_term (Fun f ls) then \<tau>' y else \<tau>\<^sub>l y"
        have "\<forall>i < length ls. ls ! i \<cdot> \<tau> = ls ! i \<cdot> \<tau>\<^sub>l" by (auto simp: term_subst_eq_conv \<tau>_def)
        then have *: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*" using \<tau>\<^sub>l by auto
        have l'': "length ss' = length (map (\<lambda>x. x \<cdot> \<tau>) ls)" using 9 len' by auto
        then have "\<forall>i < length ss'. (ss' ! i, (map (\<lambda>x. x \<cdot> \<tau>) ls) ! i) \<in> (rstep R)\<^sup>*" using *
          by auto
        from args_steps_imp_steps [OF ctxt_closed_rstep l'' this, of f] have
          "(s, D'\<langle>Fun f ls \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" by (simp add: rsteps_closed_ctxt s)
        moreover have "(D'\<langle>Fun f ls \<cdot> \<tau>\<rangle>, D'\<langle>r \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" using r'' by blast
        ultimately have r2: "(s, D'\<langle>r \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" by auto
        { fix x
          assume x: "x \<in> vars_term r"
          { assume x': "x \<in> vars_term (Fun f ls)"
            from vars_term_varposs_iff [THEN iffD1, OF this] obtain p where
              p: "p \<in> varposs (Fun f ls)" and p': "Var x = Fun f ls |_ p" by auto
            have p2: "p \<in> poss (Fun f ls)" using varposs_imp_poss [OF p] .
            have p'': "p \<noteq> \<epsilon>" using p' by fastforce
            from bspec [OF growing [unfolded growing_def] r''] have
              "\<forall>x \<in> vars_term r. \<forall>p \<in> varposs (Fun f ls). Var x = (Fun f ls) |_ p \<longrightarrow> size p \<le> 1"
              by auto
            from bspec [OF bspec [OF this x] p] p' have "size p \<le> 1" by blast
            then have "size p = 1" using p'' by (simp add: Suc_leI le_antisym)
            from Var_Fun_pos_size_One [OF p' p2 this] obtain i where i: "i < length ls"
              and x'': "ls ! i = Var x" by auto
            then have "Var x \<cdot> \<tau> = ls ! i \<cdot> \<tau>\<^sub>l" using \<tau>_def x x' by auto
            also have "\<dots> = ss' ! i" using 15 i x x'' by auto
            finally have 14: "Var x \<cdot> \<tau> = ss' ! i" .
            have "State (qs ! i) = State (qi r \<theta> (ls ! i))" using i rule r' nth_map by auto
            also have "\<dots> = State (\<theta> x)"
              unfolding x'' using x by (induct r \<theta> "Var x :: ('f ,'v) term" rule: qi.induct) auto
            also have "\<dots> = Var x \<cdot> ?\<theta>" by simp
            finally have "State (qs ! i) = Var x \<cdot> ?\<theta>" .
            then have 13: "qs ! i = \<theta> x" by auto
            have "(fterm (ss' ! i), State (qs ! i)) \<in> (move ?E)\<^sup>*"
              using sqs len' rule r' i by (auto simp: 9)
            then have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*"
              unfolding 13 14 [symmetric] by auto
          }
          moreover
          { assume x': "x \<notin> vars_term (Fun f ls)"
            then have "\<tau>' x = \<tau> x" using \<tau>'_def \<tau>_def x by auto
            moreover then have *: "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move cr_ta)\<^sup>*"
              by auto (metis DiffI \<tau>'_def ux x x')
            ultimately have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*"
              using x x' move_mono'_rtrancl [OF *] by (auto simp: \<tau>'_def)
          }
          ultimately have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*" by auto
        }
        then have "(fterm r \<cdot> (fterm \<circ> \<tau>), fterm r \<cdot> ?\<theta>) \<in> (move ?E)\<^sup>*"
          by (intro vars_move_imp_move) auto
        then have "(C\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>, C\<langle>fterm r \<cdot> ?\<theta>\<rangle>) \<in> (move ?E)\<^sup>*"
          using move_ctxt_rtrancl by blast
        moreover have "(D\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>, C\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>) \<in> (move ?E)\<^sup>*"
          using dc by auto
        moreover have "D\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle> = fterm D'\<langle>r \<cdot> \<tau>\<rangle>" using D'
          by (metis fterm_subst_distrib map_funs_term_ctxt_distrib)
        ultimately have x2: "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, C\<langle>fterm r \<cdot> ?\<theta>\<rangle>) \<in> (move ?E)\<^sup>*" by fastforce
        have q': "q = q'" using rule r' by simp
        have "?\<Gamma>\<^sub>1 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>2 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>3 \<subseteq> \<Delta>\<^sub>C" using it_k_imp_\<Delta>\<^sub>C by blast+
        then have "\<Gamma> \<subseteq> \<Delta>\<^sub>C" and "r' \<in> \<Delta>\<^sub>C" using insert(2,3) by blast+
        then have tt: "ta_rules (ta_union ?E ?D\<^sub>2) \<subseteq> ta_rules C\<^sub>R"
          by (auto simp: \<Delta>\<^sub>C.base it_k_imp_\<Delta>\<^sub>C C\<^sub>R_def)
        have s2: "(?s, ?t) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>*"
          using in_kE2 [OF Suc(2) [unfolded \<Delta>\<^sub>k'_def]] by blast
        have et: "ta_eps (ta_union ?E ?D\<^sub>2) \<subseteq> ta_eps C\<^sub>R" by (auto simp: C\<^sub>R_def)
        from move_mono'_rtrancl [OF s2 tt this] have "(?s, ?t) \<in> (move C\<^sub>R)\<^sup>*" .
        from fterm_State_funas_imp_ta_fground [OF this funas_ta_cr] have fgs: "fground F s" .
        have "\<Delta>' \<subseteq> ?\<Gamma>\<^sub>3" using \<Delta>' it_k_mono by blast
        from move_mono [OF this, of cr_ta] seq have
          "(fterm r \<cdot> ?\<theta>, State q') \<in> (move (cr_ta\<lparr>ta_rules := ?\<Gamma>\<^sub>3\<rparr>))\<^sup>*" by blast
        from move_mono'_rtrancl [OF this, of ?E] have
          "(fterm r \<cdot> ?\<theta>, State q') \<in> (move ?E)\<^sup>*" by (auto)
        then have "(C\<langle>fterm r \<cdot> ?\<theta>\<rangle>, C\<langle>State q'\<rangle>) \<in> (move ?E)\<^sup>*"
          using move_mono'_rtrancl [of _ _ ?E] move_ctxt_rtrancl by blast
        with x2 have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, C\<langle>State q'\<rangle>) \<in> (move ?E)\<^sup>*" by fastforce
        then have "\<Delta>\<^sub>k' 0 (fterm D'\<langle>r \<cdot> \<tau>\<rangle>) C\<langle>State q'\<rangle>" by (auto simp: \<Delta>\<^sub>k'_def intro: in_k.intros)
        with tail [unfolded v q'] have *: "\<Delta>\<^sub>k' m' (fterm D'\<langle>r \<cdot> \<tau>\<rangle>) ?t"
          using in_k_trans unfolding \<Delta>\<^sub>k'_def by fastforce
        from in_kE2 [OF this [unfolded \<Delta>\<^sub>k'_def]]
          have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, ?t) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>*" by (auto)
        from move_mono'_rtrancl [OF this tt et] have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, ?t) \<in> (move C\<^sub>R)\<^sup>*" .
        from fterm_State_funas_imp_ta_fground [OF this funas_ta_cr] have "fground F D'\<langle>r \<cdot> \<tau>\<rangle>" .
        then have "funas_term D'\<langle>r \<cdot> \<tau>\<rangle> \<subseteq> F" by blast
        from IH\<^sub>2 [OF * Suc(4) this] have "D'\<langle>r \<cdot> \<tau>\<rangle> \<in> ancestors F R (ground_instances F t)" .
        then obtain w where "(D'\<langle>r \<cdot> \<tau>\<rangle>, w) \<in> (rstep R)\<^sup>*" and "w \<in> ground_instances F t"
          and "fground F D'\<langle>r \<cdot> \<tau>\<rangle>" unfolding ancestors_def by blast
        moreover then have "(s, w) \<in> (rstep R)\<^sup>*" using r2 by simp
        ultimately show ?case using fgs ancestors_def by blast
      qed
    qed
  qed
qed

lemma in_ancestors_of_final_State:
  assumes "(fterm s, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>+"
    and "q\<^sub>f \<in> ta_final C\<^sub>R" and "funas_term s \<subseteq> F"
  shows "s \<in> ancestors F R (lang A)"
proof -
  obtain k where "(fterm s, State q\<^sub>f) \<in> (move (ta_union cr_ta (TA ({r. it_k k r}))))\<^sup>+"
    using cr_moves_imp_it_k_moves [OF assms(1)] by blast
  then show ?thesis using assms(2,3)
  proof (induct k arbitrary: s q\<^sub>f)
    case (0)
    then have "(fterm s, State q\<^sub>f) \<in> (move cr_ta)\<^sup>+" by (auto simp: move_mono'_trancl)
    then have 1: "(fterm s, State q\<^sub>f) \<in> (move cr_ta)\<^sup>*" by fast
    have 3: "fground F s" using fterm_State_funas_imp_ta_fground [OF 1 funas_ta_cr_ta] by simp
    from move_reachable_states [THEN iffD1, OF 1]
      have "q\<^sub>f \<in> reachable_states cr_ta (fterm s)" .
    then have "s \<in> lang cr_ta" using 0(2) 3 by (auto simp: lang_def C\<^sub>R_def)
    then have "s \<in> lang A" using lang_cr_ta_eq_lang_A by auto
    from ancestors_refl [OF this 3, of R] show ?case .
  next
    case (Suc k)
    note IH\<^sub>0 = this(1)
    have *: "{r. it_k (Suc k) r} = ({r. it_k (Suc k) r} - {r. it_k k r}) \<union> {r. it_k k r}"
      (is "?\<Gamma>\<^sub>1 = ?\<Gamma>\<^sub>2 \<union> ?\<Gamma>\<^sub>3") by (auto simp: it_k_Suc_mono)
    let ?A\<^sub>1 = "ta_union cr_ta (TA ?\<Gamma>\<^sub>1)" and ?A\<^sub>2 = "ta_union cr_ta (ta_union (TA ?\<Gamma>\<^sub>2) (TA ?\<Gamma>\<^sub>3))"
    have "ta_rules ?A\<^sub>1 \<subseteq> ta_rules ?A\<^sub>2" by (auto)
    moreover have "ta_eps ?A\<^sub>1 \<subseteq> ta_eps ?A\<^sub>2" by (auto)
    ultimately have 3: "(fterm s, State q\<^sub>f) \<in> (move ?A\<^sub>2)\<^sup>+" using move_mono'_trancl Suc(2) by blast
    have "finite ?\<Gamma>\<^sub>2" using it_k_imp_\<Delta>\<^sub>C by (intro finite_subset [OF _ finite_\<Delta>\<^sub>C]) auto
    moreover have "?\<Gamma>\<^sub>2 \<subseteq> ?\<Gamma>\<^sub>2" ..
    ultimately show ?case using 3 Suc(3,4)
    proof (induct arbitrary: s q\<^sub>f rule: finite_subset_induct')
      case (empty)
      let ?D = "ta_union cr_ta (TA {r. it_k k r})"
      from move_mono'_trancl [OF empty(1)] have "(fterm s, State q\<^sub>f) \<in> (move ?D)\<^sup>+" by (auto)
      then show ?case using IH\<^sub>0 empty by blast
    next
      case (insert r' \<Gamma>)
      note IH\<^sub>1 = this(5)
      let ?D = "TA (insert r' \<Gamma>)" and ?D\<^sub>1 = "TA \<Gamma>" and ?D\<^sub>2 = "TA {r'}" and ?G = "TA ?\<Gamma>\<^sub>3"
      let ?E = "ta_union cr_ta (ta_union ?D\<^sub>1 ?G)"
      have "?\<Gamma>\<^sub>1 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>2 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>3 \<subseteq> \<Delta>\<^sub>C" using it_k_imp_\<Delta>\<^sub>C by blast+
      then have "\<Gamma> \<subseteq> \<Delta>\<^sub>C" and "r' \<in> \<Delta>\<^sub>C" using insert(2,3) by blast+
      have t1: "ta_rules ?E \<subseteq> ta_rules C\<^sub>R" and t2: "ta_rules (ta_union ?E ?D\<^sub>2) \<subseteq> ta_rules C\<^sub>R"
        by (auto simp: \<Delta>\<^sub>C.base \<open>r' \<in> \<Delta>\<^sub>C\<close> \<open>\<Gamma> \<subseteq> \<Delta>\<^sub>C\<close> it_k_imp_\<Delta>\<^sub>C C\<^sub>R_def)
      have e1: "ta_eps ?E \<subseteq> ta_eps C\<^sub>R" and e2: "ta_eps (ta_union ?E ?D\<^sub>2) \<subseteq> ta_eps C\<^sub>R"
        by (auto simp: C\<^sub>R_def)
      have *: "ta_union cr_ta (ta_union ?D ?G) = ta_union ?E ?D\<^sub>2" by simp
      have 1: "r' \<in> {r. it_k (Suc k) r}" and 2: "r' \<notin> ta_rules cr_ta"
        using Diff_iff [THEN iffD1, OF insert(3)] it_k_0 it_k_0_subseteq_it_k by auto
      obtain f ls r \<theta> q' \<Delta>' where rule: "r' = TA_rule f (map (qi r \<theta>) ls) q'"
      and r'': "(Fun f ls, r) \<in> R"
      and range: "range \<theta> \<subseteq> states cr_ta"
      and "q' \<in> states cr_ta"
      and \<Delta>': "\<forall>r \<in> \<Delta>'. (\<exists>n \<le> k. it_k n r)"
      and seq: "(fterm r \<cdot> ((State \<circ> \<theta>) :: 'v \<Rightarrow> ('f, 'q, 'v) crterm), State q')
           \<in> (move (cr_ta\<lparr>ta_rules := \<Delta>'\<rparr>))\<^sup>*" using 1 2 by (auto elim: it_k.cases)
      let ?\<theta> = "(State \<circ> \<theta>) :: 'v \<Rightarrow> ('f, 'q, 'v) crterm"
      def \<Delta>\<^sub>k' \<equiv> "(in_k ?E r') :: nat \<Rightarrow> ('f, 'q, 'v) crterm \<Rightarrow> ('f, 'q, 'v) crterm \<Rightarrow> bool"
      from insert(6) have "(fterm s, State q\<^sub>f) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>+" by (simp add: *)
      from move_trancl_in_k_ex [OF this]
        obtain m where "\<Delta>\<^sub>k' m (fterm s) (State q\<^sub>f)" unfolding \<Delta>\<^sub>k'_def by blast
      then show ?case using insert(2,7,8)
      proof (induction m arbitrary: s q\<^sub>f)
        case (0)
        then have "(fterm s, State q\<^sub>f) \<in> (move ?E)\<^sup>*"
          unfolding \<Delta>\<^sub>k'_def using in_k.simps by blast
        moreover have "fterm s \<noteq> State q\<^sub>f" using fterm_no_State by fast
        ultimately have "(fterm s, State q\<^sub>f) \<in> (move ?E)\<^sup>+"
          by (simp add: \<open>fterm s \<noteq> State q\<^sub>f\<close> rtrancl_eq_or_trancl)
        from IH\<^sub>1 [OF this 0(3,4)] show ?case .
      next
        case (Suc m')
        note IH\<^sub>2 = this(1) [OF _ Suc(3)]
        let ?s = "fterm s" and ?q\<^sub>f = "State q\<^sub>f"
        obtain qs q where r': "r' = TA_rule f qs q" using rule by auto
        from in_k_SucE [OF Suc(2) [unfolded \<Delta>\<^sub>k'_def]] obtain u v where 1: "\<Delta>\<^sub>k' 0 ?s u"
          and uv: "(u, v) \<in> move (TA {r'})"
          and ne': "u \<noteq> v"
          and tail: "\<Delta>\<^sub>k' m' v ?q\<^sub>f" unfolding \<Delta>\<^sub>k'_def by blast
        obtain C where u: "u = C\<langle>fqterm f qs\<rangle>"
          and v: "v = C\<langle>State q\<rangle>"
          and "(C\<langle>fqterm f qs\<rangle>, C\<langle>State q\<rangle>) \<in> move (TA {r'})"
          using uv ne' rule apply (cases rule: move.cases) using uv and r' by auto
        from 1 have 2: "(?s, C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*" unfolding u \<Delta>\<^sub>k'_def by cases blast
        then obtain s' :: "('f, 'q, 'v) crterm" where s': "s' = ?s"
          and wf: "s' \<in> wf_terms" using fterm_in_wf_terms [of s] by blast
        have "(s', C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*" using 2 unfolding s' by blast
        from distr_moves'' [OF this wf] obtain D ss where s'': "s' = D\<langle>Fun (Inl f) ss\<rangle>"
          and l'': "length (map State qs) = length ss"
          and sqs: "\<forall>i < length ss. (ss ! i, map State qs ! i) \<in> (move ?E)\<^sup>*"
          and "(D\<langle>fqterm f qs\<rangle>, C\<langle>fqterm f qs\<rangle>) \<in> (move ?E)\<^sup>*"
          and mc: "(D, C) \<in> move_comp ?E" by force
        from move_comp_imp_moves [OF mc] have dc: "\<forall>t. (D\<langle>t\<rangle>, C\<langle>t\<rangle>) \<in> (move ?E)\<^sup>*" ..
        have len': "length ls = length ss" using l'' rule r' by auto
        have "fterm s = D\<langle>Fun (Inl f) ss\<rangle>" using s' s'' by auto
        then obtain ss' where 9: "ss = map fterm ss'"
          by (metis fterm_args_fterm map_funs_term_ctxt_decomp)
        have x: "?s = D\<langle>fterm (Fun f ss')\<rangle>" using s' s'' 9 by auto
        then obtain D' where 12: "D\<langle>fterm (Fun f ss')\<rangle> = (fctxt D')\<langle>fterm (Fun f ss')\<rangle>"
          and D': "D = fctxt D'" using map_funs_term_ctxt_decomp by blast
        then have "D\<langle>fterm (Fun f ss')\<rangle> = (fterm D'\<langle>Fun f ss'\<rangle>)" by simp
        then have y: "?s = (fterm D'\<langle>Fun f ss'\<rangle> :: ('f, 'q, 'v) crterm)" unfolding x [symmetric] .
        have "inj fterm" using map_funs_term_inj' [OF inj_Inl [of UNIV]] .
        from inj_onD [OF this y] have s: "s = D'\<langle>Fun f ss'\<rangle>" by auto
        { fix i
          assume i: "i < length ls"
          { assume "ls ! i \<in> Var ` vars_term r"
            moreover def \<tau> \<equiv> "(\<lambda>y. if Var y = ls ! i then ss' ! i else Var y)"
            ultimately have "(ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
              \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i)" by force
          }
          moreover
          { assume a: "ls ! i \<notin> Var ` vars_term r"
            have "qs ! i = qi r \<theta> (ls ! i)" using i rule r' nth_map by auto
            also have "\<dots> = Inr (star c (ls ! i))" using a by (induct r \<theta> "ls ! i" rule: qi.induct) auto
            finally have "qs ! i = Inr (star c (ls ! i))" .
            then have "(fterm (ss' ! i), State (Inr (star c (ls ! i)))) \<in> (move ?E)\<^sup>*"
              using sqs i l'' rule r' by (auto simp: 9)
            moreover have "fterm (ss' ! i) \<noteq> State (Inr (star c (ls ! i)))" using fterm_no_State by fast
            ultimately have b: "(fterm (ss' ! i), State (Inr (star c (ls ! i)))) \<in> (move ?E)\<^sup>+"
              by (auto simp: rtrancl_eq_or_trancl)
            have c: "ls ! i \<in> sr R" unfolding sr_def using i r'' nth_mem by blast
  
            have "i < length ss'" using i len' 9 by auto
            then have "(ss' ! i) \<unlhd> s" using s by (meson arg_subteq ctxt_supteq nth_mem supteq_trans) 
            then have d: "funas_term (ss' ! i) \<subseteq> F" using Suc by force
            from move_mono'_trancl [OF b t1 e1]
              have b': "(fterm (ss' ! i), State (Inr (star c (ls ! i)))) \<in> (move C\<^sub>R)\<^sup>+" .
            from in_ancestors_of_State [OF b' c d]
              have "ss' ! i \<in> ancestors F R (ground_instances F (ls ! i))" .
            then have "\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
              unfolding ground_instances_def ancestors_def by auto
          }
          ultimately have "\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
            \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i)" by blast
        }
        then have "\<forall>i < length ls. (\<exists>\<tau>. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*
          \<and> (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau> y = ss' ! i))" by blast
        then obtain \<tau>\<^sub>0 where 11: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> (\<tau>\<^sub>0 i)) \<in> (rstep R)\<^sup>*" 
          and 14: "\<forall>i < length ls. (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau>\<^sub>0 i y = ss' ! i)"
          by metis+
        have "is_partition (map vars_term ls)" using linear r'' by fastforce
        from subst_merge [OF this] obtain \<tau>\<^sub>l
          where 16: "\<forall>i < length ls. \<forall>x \<in> vars_term (ls ! i). \<tau>\<^sub>l x = \<tau>\<^sub>0 i x" by blast
        then have \<tau>\<^sub>l: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> \<tau>\<^sub>l) \<in> (rstep R)\<^sup>*"
          using 11 by (metis term_subst_eq)
        have 15: "\<forall>i < length ls. (\<forall>y. ls ! i \<in> Var ` vars_term r \<and> Var y = ls ! i \<longrightarrow> \<tau>\<^sub>l y = ss' ! i)"
          using 14 16 by auto
        { fix x
          assume "x \<in> vars_term r - vars_term (Fun f ls)"
          from range have "\<theta> x \<in> states cr_ta" by auto
          from cr_ta_all_states_accessible [OF this, THEN accessible_ground_move_rtrancl]
            have "\<exists>u. ground (fterm u) \<and>
              (fterm u, (State (\<theta> x)) :: ('f, 'q, 'v) crterm) \<in> (move cr_ta)\<^sup>*"
            by auto
        }
        then obtain u where ux: "\<forall>x \<in> vars_term r - vars_term (Fun f ls).
          ground (fterm (u x))
          \<and> (fterm (u x), (State (\<theta> x)) :: ('f, 'q, 'v) crterm) \<in> (move cr_ta)\<^sup>*"
          by metis
        def \<tau>' \<equiv> "\<lambda>y. if y \<in> vars_term r - vars_term (Fun f ls) then u y else Var y"
        def \<tau> \<equiv> "\<lambda>y. if y \<in> vars_term r - vars_term (Fun f ls) then \<tau>' y else \<tau>\<^sub>l y"
        have "\<forall>i < length ls. ls ! i \<cdot> \<tau> = ls ! i \<cdot> \<tau>\<^sub>l" by (auto simp: term_subst_eq_conv \<tau>_def)
        then have *: "\<forall>i < length ls. (ss' ! i, ls ! i \<cdot> \<tau>) \<in> (rstep R)\<^sup>*" using \<tau>\<^sub>l by auto
        have l'': "length ss' = length (map (\<lambda>x. x \<cdot> \<tau>) ls)" using 9 len' by auto
        then have "\<forall>i < length ss'. (ss' ! i, (map (\<lambda>x. x \<cdot> \<tau>) ls) ! i) \<in> (rstep R)\<^sup>*" using *
          by auto
        from args_steps_imp_steps [OF ctxt_closed_rstep l'' this, of f] have
          "(s, D'\<langle>Fun f ls \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" by (simp add: rsteps_closed_ctxt s)
        moreover have "(D'\<langle>Fun f ls \<cdot> \<tau>\<rangle>, D'\<langle>r \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" using r'' by blast
        ultimately have r2: "(s, D'\<langle>r \<cdot> \<tau>\<rangle>) \<in> (rstep R)\<^sup>*" by auto
        { fix x
          assume x: "x \<in> vars_term r"
          { assume x': "x \<in> vars_term (Fun f ls)"
            from vars_term_varposs_iff [THEN iffD1, OF this] obtain p where
              p: "p \<in> varposs (Fun f ls)" and p': "Var x = Fun f ls |_ p" by auto
            have p2: "p \<in> poss (Fun f ls)" using varposs_imp_poss [OF p] .
            have p'': "p \<noteq> \<epsilon>" using p' by fastforce
            from bspec [OF growing [unfolded growing_def] r''] have
              "\<forall>x \<in> vars_term r. \<forall>p \<in> varposs (Fun f ls). Var x = (Fun f ls) |_ p \<longrightarrow> size p \<le> 1"
              by auto
            from bspec [OF bspec [OF this x] p] p' have
              "size p = 1" using p'' by (simp add: Suc_leI le_antisym)
            from Var_Fun_pos_size_One [OF p' p2 this] obtain i where i: "i < length ls"
              and x'': "ls ! i = Var x" by auto
            then have "Var x \<cdot> \<tau> = ls ! i \<cdot> \<tau>\<^sub>l" using \<tau>_def x x' by auto
            also have "\<dots> = ss' ! i" using 15 i x x'' by auto
            finally have 14: "Var x \<cdot> \<tau> = ss' ! i" .
            have "State (qs ! i) = State (qi r \<theta> (ls ! i))" using i rule r' nth_map by auto
            also have "\<dots> = State (\<theta> x)"
              unfolding x'' using x by (induct r \<theta> "Var x :: ('f ,'v) term" rule: qi.induct) auto
            also have "\<dots> = Var x \<cdot> ?\<theta>" by simp
            finally have "State (qs ! i) = Var x \<cdot> ?\<theta>" .
            then have 13: "qs ! i = \<theta> x" by auto
            have "(fterm (ss' ! i), State (qs ! i)) \<in> (move ?E)\<^sup>*"
              using sqs len' rule r' i by (auto simp: 9)
            then have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*"
              unfolding 13 14 [symmetric] by auto
          }
          moreover
          { assume x': "x \<notin> vars_term (Fun f ls)"
            then have "\<tau>' x = \<tau> x" using \<tau>'_def \<tau>_def x by auto
            moreover then have *: "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move cr_ta)\<^sup>*"
              by auto (metis DiffI \<tau>'_def ux x x')
            ultimately have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*"
              using x x' move_mono'_rtrancl [OF *] by (auto simp: \<tau>'_def)
          }
          ultimately have "((fterm \<circ> \<tau>) x, ?\<theta> x) \<in> (move ?E)\<^sup>*" by auto
        }
        then have "(fterm r \<cdot> (fterm \<circ> \<tau>), fterm r \<cdot> ?\<theta>) \<in> (move ?E)\<^sup>*"
          by (intro vars_move_imp_move) auto
        then have "(C\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>, C\<langle>fterm r \<cdot> ?\<theta>\<rangle>) \<in> (move ?E)\<^sup>*"
          using move_ctxt_rtrancl by blast
        moreover have w: "D\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle> = fterm D'\<langle>r \<cdot> \<tau>\<rangle>" using D'
          by (metis fterm_subst_distrib map_funs_term_ctxt_distrib)
        moreover have "(D\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>, C\<langle>fterm r \<cdot> (fterm \<circ> \<tau>)\<rangle>) \<in> (move ?E)\<^sup>*"
          using dc by auto
        ultimately have x2: "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, C\<langle>fterm r \<cdot> ?\<theta>\<rangle>) \<in> (move ?E)\<^sup>*" by fastforce
        have q': "q = q'" using rule r' by simp
        have "?\<Gamma>\<^sub>1 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>2 \<subseteq> \<Delta>\<^sub>C" and "?\<Gamma>\<^sub>3 \<subseteq> \<Delta>\<^sub>C" using it_k_imp_\<Delta>\<^sub>C by blast+
        then have "\<Gamma> \<subseteq> \<Delta>\<^sub>C" and "r' \<in> \<Delta>\<^sub>C" using insert(2,3) by blast+
        then have tt: "ta_rules (ta_union ?E ?D\<^sub>2) \<subseteq> ta_rules C\<^sub>R"
          by (auto simp: \<Delta>\<^sub>C.base it_k_imp_\<Delta>\<^sub>C C\<^sub>R_def)
        have s2: "(?s, ?q\<^sub>f) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>*"
          using in_kE2 [OF Suc(2) [unfolded \<Delta>\<^sub>k'_def]] by blast
        have et: "ta_eps (ta_union ?E ?D\<^sub>2) \<subseteq> ta_eps C\<^sub>R" by (auto simp: C\<^sub>R_def)
        from move_mono'_rtrancl [OF s2 tt this] have "(?s, ?q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" .
        from fterm_State_funas_imp_ta_fground [OF this funas_ta_cr] have fgs: "fground F s" .
        have "\<Delta>' \<subseteq> ?\<Gamma>\<^sub>3" using \<Delta>' it_k_mono by blast
        from move_mono [OF this, of cr_ta] seq have
          "(fterm r \<cdot> ?\<theta>, State q') \<in> (move (cr_ta\<lparr>ta_rules := ?\<Gamma>\<^sub>3\<rparr>))\<^sup>*" by blast
        from move_mono'_rtrancl [OF this, of ?E] have
          "(fterm r \<cdot> ?\<theta>, State q') \<in> (move ?E)\<^sup>*" by (auto)
        then have "(C\<langle>fterm r \<cdot> ?\<theta>\<rangle>, C\<langle>State q'\<rangle>) \<in> (move ?E)\<^sup>*" using move_ctxt_rtrancl by blast
        with x2 have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, C\<langle>State q'\<rangle>) \<in> (move ?E)\<^sup>*" by fastforce
        then have "\<Delta>\<^sub>k' 0 (fterm D'\<langle>r \<cdot> \<tau>\<rangle>) C\<langle>State q'\<rangle>" by (auto simp: \<Delta>\<^sub>k'_def intro: in_k.intros)
        with tail [unfolded v q'] have *: "\<Delta>\<^sub>k' m' (fterm D'\<langle>r \<cdot> \<tau>\<rangle>) (State q\<^sub>f)"
          using in_k_trans unfolding \<Delta>\<^sub>k'_def w by fastforce
        from in_kE2 [OF this [unfolded \<Delta>\<^sub>k'_def]]
          have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, ?q\<^sub>f) \<in> (move (ta_union ?E ?D\<^sub>2))\<^sup>*" by (auto)
        from move_mono'_rtrancl [OF this tt et] have "(fterm D'\<langle>r \<cdot> \<tau>\<rangle>, ?q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" .
        from fterm_State_funas_imp_ta_fground [OF this funas_ta_cr] have "fground F D'\<langle>r \<cdot> \<tau>\<rangle>" .
        then have "funas_term D'\<langle>r \<cdot> \<tau>\<rangle> \<subseteq> F" by blast
        from IH\<^sub>2 [OF * Suc(4) this] have "D'\<langle>r \<cdot> \<tau>\<rangle> \<in> ancestors F R (lang A)" .
        then obtain w where "(D'\<langle>r \<cdot> \<tau>\<rangle>, w) \<in> (rstep R)\<^sup>*" and "w \<in> lang A"
          and "fground F D'\<langle>r \<cdot> \<tau>\<rangle>" unfolding ancestors_def by blast
        moreover then have "(s, w) \<in> (rstep R)\<^sup>*" using r2 by simp
        ultimately show ?case using fgs ancestors_def by blast
      qed
    qed
  qed
qed

lemma etac_ancestors:
  "lang C\<^sub>R = ancestors F R (lang A)" (is "?L = ?R")
proof (intro equalityI subsetI)
  fix s
  assume "s \<in> ?L"
  from langE [OF this] obtain q\<^sub>f where 1: "(fterm s, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>*" and 2: "q\<^sub>f \<in> ta_final C\<^sub>R"
    by blast
  have "fground F s" using fterm_State_funas_imp_ta_fground [OF 1 funas_ta_cr] by simp
  then have 3: "funas_term s \<subseteq> F" by simp
  have 4: "(fterm s, State q\<^sub>f) \<in> (move C\<^sub>R)\<^sup>+" using 1
    by (simp add: fterm_no_State rtrancl_eq_or_trancl) 
  from in_ancestors_of_final_State [OF 4 2 3] show "s \<in> ?R" .
next
  fix s
  assume "s \<in> ?R"
  then show "s \<in> ?L" using ancestors_subseteq_lang_cr by blast
qed

end

fun var2state :: "('f, 'v) term \<Rightarrow> ('f + 'v, 'w) term"
where
  "var2state (Var x) = Fun (Inr x) []"
| "var2state (Fun f ts) = Fun (Inl f) (map var2state ts)"

fun state2var :: "('f + 'q, 'v) term \<Rightarrow> ('f, 'q) term"
where
  "state2var (Fun (Inr q) []) = Var q"
| "state2var (Fun (Inl f) ts) = Fun f (map state2var ts)"
| "state2var _ = undefined"

lemma ta_res_reachable_states_eq:
  fixes t :: "('f, 'v) term"
  shows "ta_res A t = reachable_states A (var2state t)"
  (is "?L = ?R" is "_ = reachable_states A (?v t)")
proof (intro equalityI subsetI)
  fix q
  assume "q \<in> ?L"
  then show "q \<in> ?R"
  proof (induction t arbitrary: q)
    case (Var x)
    then show ?case by simp
  next
    case (Fun f ts)
    then have "q \<in> ta_res A (Fun f ts)" by simp
    then obtain q' qs where 1: "TA_rule f qs q' \<in> ta_rules A"
      and 2: "(q', q) \<in> (ta_eps A)\<^sup>*"
      and 3: "length qs = length ts"
      and 4: "\<forall>i<length ts. qs ! i \<in> map (ta_res A) ts ! i" by auto
    let ?ts = "map ?v ts"
    have 5: "length qs = length ?ts" using 3 by simp
    { fix i
      assume i: "i < length ts"
      then have "ts ! i \<in> set ts" by simp
      moreover from 4 have "qs ! i \<in> ta_res A (ts ! i)" using i by simp
      ultimately have "qs ! i \<in> reachable_states A (?v (ts ! i))" using Fun by simp
    }
    then have "\<forall>i<length ts. qs ! i \<in> reachable_states A (?v (ts ! i))" by blast
    then have "\<forall>i<length ?ts. qs ! i \<in> reachable_states A (map ?v ts ! i)" by simp
    with reachable_states_InlI [OF 1 2 5 this]
      have "q \<in> reachable_states A (Fun (Inl f) (map ?v ts))" by blast
    then show ?case unfolding Fun by force
  qed
next
  fix q
  assume "q \<in> ?R"
  then show "q \<in> ?L"
  proof (induct t arbitrary: q)
    case (Var x)
    then show ?case by simp
  next
    case (Fun f ts)
    let ?ts = "map ?v ts"
    from Fun have "q \<in> reachable_states A (?v (Fun f ts))" by simp
    then have "q \<in> reachable_states A (Fun (Inl f) (map ?v ts))" by auto
    then obtain q' qs where 1: "TA_rule f qs q' \<in> ta_rules A"
      and 2: "(q', q) \<in> (ta_eps A)\<^sup>*"
      and 3: "length qs = length ?ts"
      and 4: "\<forall>i<length ?ts. qs ! i \<in> reachable_states A (?ts ! i)" by auto
    have 5: "length qs = length ts" using 3 by simp
    { fix i
      assume i: "i < length ts"
      then have "ts ! i \<in> set ts" by simp
      moreover from 4 have "qs ! i \<in> reachable_states A (?v (ts ! i))" using i by simp
      ultimately have "qs ! i \<in> ta_res A (ts ! i)" using Fun(1) by simp
    }
    then have "\<forall>i<length ts. qs ! i \<in> ta_res A (ts ! i)" by simp
    then have "\<forall>i<length ts. qs ! i \<in> map (ta_res A) ts ! i" by simp
    then show ?case using 1 2 5 by auto
  qed
qed

lemma ground_fterm_var2_state [simp]:
  assumes "ground t"
  shows "fterm (adapt_vars t) = var2state t"
using assms by (induct t) (auto)

lemma ground_fterm_var2_state' [simp]:
  assumes "ground t"
  shows "fterm t = var2state t"
using assms by (induct t) (auto)

lemma ta_lang_lang_eq:
  "ta_lang A = lang A"
apply (auto simp: ta_lang_def lang_def ta_res_reachable_states_eq)
apply (subst (asm) ta_res_reachable_states_eq)
apply auto
by (metis Set.set_insert adapt_vars_adapt_vars disjoint_insert(2) ground_adapt_vars
    ground_fterm_var2_state ta_res_reachable_states_eq)

end
